#ifndef _Auxilary_h_
#define _Auxilary_h_

#include <Base.h>
#include <defines/Auxilary.d>

class Auxilary : public SymbolBase {
private:
public:
    Auxilary(uint32_t idx) : SymbolBase(idx) {}
    ~Auxilary() {}
    bool isAuxilary() { return true; }
    const char* getTypeName();

    AUXILARY_MACROS_BASIS("For the get_X field macros check the defines directory");
    virtual uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_No_Type; }

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);

};

class AuxilarySection : public Auxilary {
private:
    AUXENT entry;
public:
    AuxilarySection(uint32_t idx) : Auxilary(idx) {}
    ~AuxilarySection() {}
    uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_Section; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_scnlen,x_scn)    { return entry.x_scn.x_scnlen; }
    GET_FIELD_CLASS_A(uint16_t,x_nreloc,x_scn)    { return entry.x_scn.x_nreloc; }
    GET_FIELD_CLASS_A(uint16_t,x_nlinno,x_scn)     { return entry.x_scn.x_nlinno; }

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);
};

class AuxilaryException : public Auxilary {
private:
    AUXENT_64 entry;
public:
    AuxilaryException(uint32_t idx) : Auxilary(idx) {}
    ~AuxilaryException() {}
    uint8_t getAuxilaryType() { return entry.x_auxtype.x_auxtype; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint64_t,x_exptr,x_except) { return entry.x_except.x_exptr; }
    GET_FIELD_CLASS_A(uint32_t,x_fsize,x_except) { return entry.x_except.x_fsize; }
    GET_FIELD_CLASS_A(uint32_t,x_endndx,x_except){ return entry.x_except.x_endndx; }

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);

    void changeExptrCopy(uint64_t exptr,char* buff);
};

class AuxilaryFile : public Auxilary {
public:
    AuxilaryFile(uint32_t idx) : Auxilary(idx) {}
    ~AuxilaryFile() {}

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);
};

class AuxilaryFile32 : public AuxilaryFile {
private:
    AUXENT entry;
public:
    AuxilaryFile32(uint32_t idx) : AuxilaryFile(idx) {}
    ~AuxilaryFile32() {}
    uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_File; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(char*,x_fname,x_file)     { return entry.x_file.x_fname; }
    GET_FIELD_CLASS_A(uint32_t,x_zeroes,x_file) { return entry.x_file._x.x_zeroes; }
    GET_FIELD_CLASS_A(uint32_t,x_offset,x_file) { return entry.x_file._x.x_offset; }
    GET_FIELD_CLASS_A(uint8_t,x_ftype,x_file)     { return entry.x_file._x.x_ftype; }
};

class AuxilaryFile64 : public AuxilaryFile {
private:
    AUXENT_64 entry;
public:
    AuxilaryFile64(uint32_t idx) : AuxilaryFile(idx) {}
    ~AuxilaryFile64() {}
    uint8_t getAuxilaryType() { return entry.x_auxtype.x_auxtype; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(char*,x_fname,x_file)     { return entry.x_file.x_fname; }
    GET_FIELD_CLASS_A(uint32_t,x_zeroes,x_file) { return entry.x_file._x.x_zeroes; }
    GET_FIELD_CLASS_A(uint32_t,x_offset,x_file) { return entry.x_file._x.x_offset; }
    GET_FIELD_CLASS_A(uint8_t,x_ftype,x_file)     { return entry.x_file._x.x_ftype; }
};

class AuxilaryBlock : public Auxilary {
public:
    AuxilaryBlock(uint32_t idx) : Auxilary(idx) {}
    ~AuxilaryBlock() {}

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);
};

class AuxilaryBlock32 : public AuxilaryBlock {
private:
    AUXENT entry;
public:
    AuxilaryBlock32(uint32_t idx) : AuxilaryBlock(idx) {}
    ~AuxilaryBlock32() {}
    uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_Block; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_lnno,x_misc) { return entry.x_sym.x_misc.x_lnsz.x_lnno; }
};

class AuxilaryBlock64 : public AuxilaryBlock {
private:
    AUXENT_64 entry;
public:
    AuxilaryBlock64(uint32_t idx) : AuxilaryBlock(idx) {}
    ~AuxilaryBlock64() {}
    uint8_t getAuxilaryType() { return entry.x_auxtype.x_auxtype; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_lnno,x_misc) { return entry.x_sym.x_misc.x_lnsz.x_lnno; }
};

class AuxilaryFunction : public Auxilary {
public:
    AuxilaryFunction(uint32_t idx) : Auxilary(idx) {}
    ~AuxilaryFunction() {}

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);
    virtual void changeExptrLnnoptrCopy(uint32_t exptr,uint64_t lnnoptr,char* buff) { __SHOULD_NOT_ARRIVE; }
};

class AuxilaryFunction32 : public AuxilaryFunction {
private:
    AUXENT entry;
public:
    AuxilaryFunction32(uint32_t idx) : AuxilaryFunction(idx) {}
    ~AuxilaryFunction32() {}
    uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_Function; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_exptr,x_fcn)    { return entry.x_sym.x_tagndx; }
    GET_FIELD_CLASS_A(uint32_t,x_fsize,x_fcn)    { return entry.x_sym.x_misc.x_fsize; }
    GET_FIELD_CLASS_A(uint64_t,x_lnnoptr,x_fcn)    { return entry.x_sym.x_fcnary.x_fcn.x_lnnoptr; }
    GET_FIELD_CLASS_A(uint32_t,x_endndx,x_fcn)    { return entry.x_sym.x_fcnary.x_fcn.x_endndx; }
    void changeExptrLnnoptrCopy(uint32_t exptr,uint64_t lnnoptr,char* buff);
};

class AuxilaryFunction64 : public AuxilaryFunction {
private:
    AUXENT_64 entry;
public:
    AuxilaryFunction64(uint32_t idx) : AuxilaryFunction(idx) {}
    ~AuxilaryFunction64() {}
    uint8_t getAuxilaryType() { return entry.x_auxtype.x_auxtype; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_exptr,x_fcn)    { return 0; }
    GET_FIELD_CLASS_A(uint32_t,x_fsize,x_fcn)    { return entry.x_fcn.x_fsize; }
    GET_FIELD_CLASS_A(uint64_t,x_lnnoptr,x_fcn)    { return entry.x_fcn.x_lnnoptr; }
    GET_FIELD_CLASS_A(uint32_t,x_endndx,x_fcn)    { return entry.x_fcn.x_endndx; }
    void changeExptrLnnoptrCopy(uint32_t exptr,uint64_t lnnoptr,char* buff);
};

class AuxilaryCSect : public Auxilary {
public:
    AuxilaryCSect(uint32_t idx) : Auxilary(idx) {}
    ~AuxilaryCSect() {}

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);
    virtual uint64_t getLength() { __SHOULD_NOT_ARRIVE; return 0; }

};

class AuxilaryCSect32 : public AuxilaryCSect {
private:
    AUXENT entry;
public:
    AuxilaryCSect32(uint32_t idx) : AuxilaryCSect(idx) {}
    ~AuxilaryCSect32() {}
    uint8_t getAuxilaryType() { return Type__Auxilary_Symbol_CSect; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_scnlen,x_csect)    { return entry.x_csect.x_scnlen; }
    GET_FIELD_CLASS_A(uint32_t,x_parmhash,x_csect)    { return entry.x_csect.x_parmhash; }
    GET_FIELD_CLASS_A(uint16_t,x_snhash,x_csect)    { return entry.x_csect.x_snhash; }
    GET_FIELD_CLASS_A(uint8_t,x_smtyp,x_csect)        { return entry.x_csect.x_smtyp; }
    GET_FIELD_CLASS_A(uint8_t,x_smclas,x_csect)        { return entry.x_csect.x_smclas; }
    GET_FIELD_CLASS_A(uint32_t,x_stab,x_csect)        { return entry.x_csect.x_stab; }
    GET_FIELD_CLASS_A(uint16_t,x_snstab,x_csect)    { return entry.x_csect.x_snstab; }

    uint64_t getLength();
};

class AuxilaryCSect64 : public AuxilaryCSect {
private:
    AUXENT_64 entry;
public:
    AuxilaryCSect64(uint32_t idx) : AuxilaryCSect(idx) {}
    ~AuxilaryCSect64() {}
    uint8_t getAuxilaryType() { return entry.x_auxtype.x_auxtype; }

    char* charStream() { return (char*)&entry; }
    GET_FIELD_CLASS_A(uint32_t,x_scnlen_lo,x_csect)    { return entry.x_csect.x_scnlen_lo; }
    GET_FIELD_CLASS_A(uint32_t,x_parmhash,x_csect)    { return entry.x_csect.x_parmhash; }
    GET_FIELD_CLASS_A(uint16_t,x_snhash,x_csect)    { return entry.x_csect.x_snhash; }
    GET_FIELD_CLASS_A(uint8_t,x_smtyp,x_csect)        { return entry.x_csect.x_smtyp; }
    GET_FIELD_CLASS_A(uint8_t,x_smclas,x_csect)        { return entry.x_csect.x_smclas; }
    GET_FIELD_CLASS_A(uint32_t,x_scnlen_hi,x_csect)    { return entry.x_csect.x_scnlen_hi; }

    uint64_t getLength();
};

#endif
