#define _S_(name) name ## _64

/*
 * Defines for the flags to identify characteristics of a collection
 * of object files link-edited together
 *    Bit-mapped field of the o_cpuflag field above
 */
#define    TOBJ_SAME    0x80    /* All object files have same cputype field. */
#define    TOBJ_COM     0x40    /* All object files are in a valid cputype */
                                /*     subset. */
#define    TOBJ_EMULATE 0x20    /* One or more object files are of a */
                                /* cputype that would require emulated */
                                /* instructions on the resultant cputype. */
                                /* E.g., PWR objects promoted to PPC. */
#define    TOBJ_COMPAT  0x10    /* One or more object files are TCPU_ANY. */
                                /* They have been ignored in determining */
                                /* the resultant o_cputype value. */
/*
 * Defines for the POWER and PowerPC CPU Types 
 *    The o_cputype field above
 *    The n_cputype field of C_FILE symbol table entry
 */
#define TCPU_INVALID  0    /* Invalid id - assumes POWER for old objects */
#define TCPU_PPC      1    /* PowerPC common architecture 32 bit mode */
#define TCPU_PPC64    2    /* PowerPC common architecture 64 bit mode */
#define TCPU_COM      3    /* POWER and PowerPC architecture common */
#define TCPU_PWR      4    /* POWER common architecture objects */
#define TCPU_ANY      5    /* Mixture of any incompatable POWER */
                           /* and PowerPC architecture implementations */
#define TCPU_601      6    /* 601 implementation of PowerPC architecture */
#define TCPU_603      7    /* 603 implementation of PowerPC architecture */
#define TCPU_604      8    /* 604 implementation of PowerPC architecture */
#define TCPU_PWR1    10    /* RS1 implementation of POWER architecture */
#define TCPU_620     16    /* 620 implementation of PowerPC 64-bit architecture */
#define TCPU_A35     17    /* A35 implementation of PowerPC 64-bit architecture */
#define TCPU_PWRX   224    /* RS2 implementation of POWER architecture */

/* Define the auxiliary header of an XCOFF file.  In general, the auxiliary
   header can be any size, and the length of the auxiliary in any particular
   object file is specified in the f_opthdr field of the XCOFF header.

   Some AIX programs generate auxiliary headers for 32-bit object files that
   end after the data_start field.  The macro _AOUTHSZ_SHORT defines the
   length of such a header.  No auxiliary header is required for an object
   file that is not an executable.

   Executables, however, must have auxiliary headers that include the
   full structure definitions.  The macro _AOUTHSZ_EXEC defines the length
   the auxiliary header must have for an executable.
*/
typedef struct aouthdr {
    uint16_t    o_mflag;        /* flags - how to execute */
                                /* 0x010B: text & data may be paged */
    uint16_t    o_vstamp;       /* version stamp */
                                /* 1 is only valid value */
    uint32_t    o_tsize;        /* text size in bytes */
    uint32_t    o_dsize;        /* data size in bytes */
    uint32_t    o_bsize;        /* bss size in bytes */
    uint32_t    o_entry;        /* entry point descriptor address */
    uint32_t    o_text_start;   /* virtual address of beginning of text */
    uint32_t    o_data_start;   /* virtual address of beginning of data */
                                /* A short 32-bit auxiliary header can end here. */

    uint32_t    o_toc;          /* address of TOC */
    uint16_t    o_snentry;      /* section number for entry point */
    uint16_t    o_sntext;       /* section number for text */
    uint16_t    o_sndata;       /* section number for data */
    uint16_t    o_sntoc;        /* section number for toc */
    uint16_t    o_snloader;     /* section number for loader */
    uint16_t    o_snbss;        /* section number for bss */
    uint16_t    o_algntext;     /* log (base 2) of max text alignment */
    uint16_t    o_algndata;     /* log (base 2) of max data alignment */
    uint8_t     o_modtype[2];   /* Module type field, 1L, RE, RO */
    uint8_t     o_cpuflag;      /* bit flags - cputypes of objects */
    uint8_t     o_cputype;      /* executable cpu type identification */
    uint32_t    o_maxstack;     /* max stack size allowed (bytes) */
    uint32_t    o_maxdata;      /* max data size allowed (bytes) */
    uint32_t    o_debugger;     /* Reserved for use by debugger if it */
                                /* needs a location for a breakpoint. */
    uint32_t    o_resv2[2];     /* reserved fields */
} AOUTHDR;

typedef struct aouthdr_64 {
    uint16_t    o_mflag;      /* flags - how to execute */
                              /* 0x010B: text & data may be paged */
    uint16_t    o_vstamp;     /* version stamp */
                              /* 1 is only valid value */
    uint32_t    o_debugger;   /* Reserved for use by debugger if it */
                              /* needs a location for a breakpoint. */
    uint64_t    o_text_start; /* virtual address of text */
    uint64_t    o_data_start; /* virtual address of data */
    uint64_t    o_toc;        /* TOC address */
    uint16_t    o_snentry;    /* section number for entry point */
    uint16_t    o_sntext;     /* section number for text */
    uint16_t    o_sndata;     /* section number for data */
    uint16_t    o_sntoc;      /* section number for toc */
    uint16_t    o_snloader;   /* section number for loader */
    uint16_t    o_snbss;      /* section number for bss */
    uint16_t    o_algntext;   /* log (base 2) of max text alignment */
    uint16_t    o_algndata;   /* log (base 2) of max data alignment */
    uint8_t     o_modtype[2]; /* Module type field, 1L, RE, RO */
    uint8_t     o_cpuflag;    /* bit flags - cputypes of objects */
    uint8_t     o_cputype;    /* executable cpu type identification */
    uint32_t    o_resv2[1];
    uint64_t    o_tsize;      /* text size in bytes */
    uint64_t    o_dsize;      /* data size in bytes */
    uint64_t    o_bsize;      /* bss size in bytes */
    uint64_t    o_entry;      /* entry point address */
    uint64_t    o_maxstack;   /* max stack size allowed */
    uint64_t    o_maxdata;    /* max data size allowed */
    uint32_t    o_resv3[4];   /* Reserved */
} AOUTHDR_64;

#define _AOUTHSZ_SHORT      (size_t)&(((AOUTHDR *)0)->o_toc)
#define _AOUTHSZ_EXEC       sizeof(AOUTHDR)
#define _AOUTHSZ_EXEC_64    sizeof(AOUTHDR_64)

/*
 *      Bits for f_flags: (Most were defined for original COFF)
 *
 *      F_RELFLG    relocation info stripped from file
 *      F_EXEC      file is executable (i.e., it has a loader section)
 *      F_LNNO      line numbers stripped from file
 *      F_LSYMS     local symbols stripped from file
 *      F_FDPR_PROF File was profiled with FDPR
 *      F_FDPR_OPTI File was reordered with FDPR
 *      F_DSA       file uses Dynamic Segment Allocation (32-bit only)
 *      F_LPTEXT    executable requires large pages for text
 *      F_LPDATA    executable requires large pages for data
 *      F_DYNLOAD   file is dynamically loadable and executable
 *                  (equivalent to F_EXEC on AIX)
 *      F_SHROBJ    file is a shared object
 *      F_LOADONLY  file can be loaded by the system loader, but it is
 *                  ignored by the linker if it is a member of an archive.
 *
 *      F_AR16WR    this file has the byte ordering of an AR16WR machine
 *                  (e.g. 11/70)
 *      F_AR32WR    this file has the byte ordering of an AR32WR machine
 *                  (e.g. vax)
 *      F_AR32W     this file has the byte ordering of an AR32W machine
 *                  (e.g. S370,POWER,3b,maxi)
 *
 */

#define F_RELFLG    0x0001
#define F_EXEC      0x0002
#define F_LNNO      0x0004
#define F_LSYMS     0x0008
#define F_FDPR_PROF     0x0010
#define F_FDPR_OPTI     0x0020
#define F_DSA       0x0040
#define F_LPTEXT    0x0400
#define F_LPDATA    0x0800
#define F_DYNLOAD       0x1000
#define F_SHROBJ    0x2000
#define F_LOADONLY      0x4000

/* Unused f_flags values */
#define F_AR16WR    0x0080
#define F_AR32WR    0x0100
#define F_AR32W     0x0200

/*
 *      Magic Numbers
 */

/* IBM POWER and PowerPC */
#define U802TOCMAGIC    0737    /* readonly text segments and TOC */
#define U803TOCMAGIC    0747    /* Obsolete 64-bit XCOFF--not used */
#define U803XTOCMAGIC   0757    /* Discontinued AIX 64-bit XCOFF */
#define U64_TOCMAGIC    0767    /* AIX 64-bit XCOFF */

/* Other magic numbers--not used on AIX */
#define X386MAGIC       0514
#define U802WRMAGIC     0730    /* writeable text segments */
#define U802ROMAGIC     0735    /* readonly sharable text segments */

/* IBM RT */
#define U800TOCMAGIC    0637    /* readonly text segments and TOC */
#define U800WRMAGIC     0630    /* writeable text segments */
#define U800ROMAGIC     0635    /* readonly sharable text segments */

typedef struct filehdr {
    uint16_t    f_magic;    /* Target machine on which the object file is executable */
    uint16_t    f_nscns;    /* number of sections */
    uint32_t    f_timdat;   /* time & date stamp:
                               A value of 0 indicates no timestamp.
                               Negative values are reserved for
                               future use and should be treated
                               as 0. */
    uint32_t    f_symptr;   /* File offset to symbol table. */
    uint32_t    f_nsyms;    /* Number of symbol table entries:
                               Negative values are reserved for
                               future use and should be treated
                               as 0. */
    uint16_t    f_opthdr;   /* Size of the auxiliary header */
    uint16_t    f_flags;    /* flags */
} FILHDR;

typedef struct _S_(filehdr) {
    uint16_t    f_magic;    /* Target machine on which the object file is executable */
    uint16_t    f_nscns;    /* number of sections */
    uint32_t    f_timdat;   /* time & date stamp:
                               A value of 0 indicates no timestamp.
                               Negative values are reserved for
                               future use and should be treated
                               as 0. */
    uint64_t    f_symptr;   /* File offset to symbol table. */
    uint16_t    f_opthdr;   /* Size of the auxiliary header */
    uint16_t    f_flags;    /* flags */
    uint32_t    f_nsyms;    /* Number of symbol table entries:
                               Negative values are reserved for
                               future use and should be treated
                               as 0. */
} _S_(FILHDR);

#define FILHSZ     sizeof(FILHDR)
#define FILHSZ_64  sizeof(FILHDR_64)

/* There is one line number entry for every
 * source file line that has associated code.
 * Line numbers are grouped by function.
 * The first entry in a function grouping
 * will have l_lnno == 0 and in place of physical
 * address will be the symbol table index of
 * the symbol for the function.
*/
typedef struct lineno {
    union {
        uint32_t    l_symndx;   /* symbol table index of function if l_lnno == 0 */
        uint32_t    l_paddr;    /* address of instruction corresponding to line number */
    } l_addr;
    uint16_t    l_lnno;         /* line number */
} LINENO;

typedef struct _S_(lineno) {
    union  {
        uint32_t    l_symndx;   /* symbol table index of function if l_lnno == 0 */
        uint64_t    l_paddr;    /* address of instruction corresponding to line number */
    } l_addr;
    uint32_t    l_lnno;         /* line number */
} _S_(LINENO);

#define LINESZ       (6)        /* Do not use sizeof(LINENO) */
#define LINESZ_64   (12)        /* Do not use sizeof(LINENO_64) */


#define SYMNMLEN     8  /* Number of characters in a symbol name */
#define FILNMLEN    14  /* Number of characters in a file name */
#define DIMNUM       4  /* Number of array dimensions in auxiliary entry */

#define LDHDRSZ     sizeof(LDHDR)
#define LDHDRSZ_64  sizeof(LDHDR_64)

#define LDSYMSZ     sizeof(LDSYM)
#define LDSYMSZ_64  sizeof(LDSYM_64)

#define LDRELSZ     sizeof(LDREL)
#define LDRELSZ_64  sizeof(LDREL_64)

typedef struct ldhdr {
#define _CURRENT_LDR_VERSION    (1)
    uint32_t    l_version; /* Loader section version number */
    uint32_t    l_nsyms;   /* Qty of loader Symbol table entries */
    uint32_t    l_nreloc;  /* Qty of loader relocation table entries */
    uint32_t    l_istlen;  /* Length of loader import file id strings */
    uint32_t    l_nimpid;  /* Qty of loader import file ids. */
    uint32_t    l_impoff;  /* Offset to start of loader import file id strings */
    uint32_t    l_stlen;   /* Length of loader string table */
    uint32_t    l_stoff;   /* Offset to start of loader string table */
} LDHDR;

typedef struct _S_(ldhdr) {
#define _CURRENT_LDR_VERSION_64    (2)
    uint32_t    l_version;  /* Loader section version number */
    uint32_t    l_nsyms;    /* Qty of loader Symbol table entries */
    uint32_t    l_nreloc;   /* Qty of loader relocation table entries */
    uint32_t    l_istlen;   /* Length of loader import file id strings */
    uint32_t    l_nimpid;   /* Qty of loader import file ids. */
    uint32_t    l_stlen;    /* Length of loader string table */
    uint64_t    l_impoff;   /* Offset to start of loader import file id strings */
    uint64_t    l_stoff;    /* Offset to start of loader string table */
    uint64_t    l_symoff;   /* Offset to start of loader symbol table */
    uint64_t    l_rldoff;   /* Offset to start of loader rlds */
} _S_(LDHDR);

#define    l_name      _l._l_name
#define    l_zeroes    _l._l_l._l_zeroes
#define    l_offset    _l._l_l._l_offset
#define    l_nptr      _l._l_nptr

typedef struct ldsym {
    union {
        char          _l_name[SYMNMLEN];    /* Symbol name */
        struct {
            uint32_t  _l_zeroes;            /* offset if 0 */
            uint32_t  _l_offset;            /* offset into loader string */
        } _l_l;
        uint32_t       _l_nptr[2];          /* allows for overlaying */
    } _l;
    uint32_t          l_value;              /* Address field */
    uint16_t          l_scnum;              /* Section number */
    uint8_t           l_smtype;             /* type and imp/exp/eps */
                                            /* 0    Unused */
                                            /* 1    Import */
                                            /* 2    Entry point */
                                            /* 3    Export */
                                            /* 4    Unused */
                                            /* 5-7    Symbol type */
    uint8_t           l_smclas;            /* storage class */
    uint32_t          l_ifile;             /* import file id  string offset */
    uint32_t          l_parm;              /* type check offset  into loader string */
} LDSYM;

typedef struct _S_(ldsym) {
    uint64_t          l_value;    /* Address field */
    union {
        struct {
            uint32_t  _l_offset;  /* offset into loader string */
        } _l_l;
        uint32_t       _l_nptr;   /* Allows for overlaying. */
    } _l;
    uint16_t          l_scnum;    /* Section number */
    uint8_t           l_smtype;   /* type and imp/exp/eps */
                                  /* 0    Unused */
                                  /* 1    Import */
                                  /* 2    Entry point */
                                  /* 3    Export */
                                  /* 4    Unused */
                                  /* 5-7    Symbol type */
    uint8_t           l_smclas;   /* storage class */
    uint32_t          l_ifile;    /* import file id string offset */
    uint32_t          l_parm;     /* type check offset into loader string */
} _S_(LDSYM);


#define    L_WEAK         0x08
#define    L_EXPORT       0x10
#define    L_ENTRY        0x20
#define    L_IMPORT       0x40

#define    LDR_WEAK(x)    ((x).l_smtype & L_WEAK)
#define    LDR_EXPORT(x)  ((x).l_smtype & L_EXPORT)
#define    LDR_ENTRY(x)   ((x).l_smtype & L_ENTRY)
#define    LDR_IMPORT(x)  ((x).l_smtype & L_IMPORT)
#define    LDR_TYPE(x)    ((x).l_smtype & 0x07)


typedef struct ldrel
{
    uint32_t    l_vaddr;    /* Address field */
    uint32_t    l_symndx;   /* Loader symbol table index of */
                            /* reloc value to apply. This field */
                            /* is zero-based where 0,1,2 are */
                            /* text,data,bss and 3 is the first */
                            /* symbol entry from above */
    uint16_t    l_rtype;    /* relocation type */
    uint16_t    l_rsecnm;   /* section number being relocated */
                            /* one-based index in scnhdr table */
} LDREL;

typedef struct _S_(ldrel)
{
    uint64_t    l_vaddr;    /* Address field */
    uint16_t    l_rtype;    /* relocation type */
    uint16_t    l_rsecnm;   /* section number being relocated */
                            /* one-based index in scnhdr table */
    uint32_t    l_symndx;   /* Loader symbol table index of */
                            /* reloc value to apply. This field */
                            /* is zero-based where 0,1,2 are */
                            /* text,data,bss and 3 is the first */
                            /* symbol entry from above */
} _S_(LDREL);


/* When generated, there is one exception table entry for each
 * trap instruction in the executable code.
 * Exception entries are grouped by function.
 * The first entry in a function grouping
 * will have e_reason = 0 and in place of physical
 * address will be the symbol table index of
 * the symbol for the function.
*/
typedef struct exceptab {
    union {
        uint32_t   e_symndx;   /* sym. table index of function if e_reason == 0 */
        uint32_t   e_paddr;    /* address of trap inst */
    } e_addr;
    uint8_t        e_lang;     /* compiler language id code */
    uint8_t        e_reason;   /* exception reason code */
} EXCEPTAB;

typedef struct _S_(exceptab) {
    union {
        uint32_t   e_symndx;   /* sym. table index of function if e_reason == 0 */
        uint64_t   e_paddr;    /* address of trap inst */
    } e_addr;
    uint8_t        e_lang;     /* compiler language id code */
    uint8_t        e_reason;   /* exception reason code */
} _S_(EXCEPTAB);

#define    EXCEPTSZ     (6)    /* Do not use sizeof(EXCEPTAB) */
#define    EXCEPTSZ_64 (10)    /* Do not use sizeof(EXCEPTAB_64) */

#define    RELSZ       (10)    /* Do not use sizeof(RELOC) */
#define    RELSZ_64    (14)    /* Do not use sizeof(RELOC_64) */

typedef struct reloc {
    uint32_t        r_vaddr;   /* (virtual) address of reference */
    uint32_t        r_symndx;  /* index into symbol table */
    union {
        uint16_t    _r_type;   /* old style coff relocation type */
        struct {
            uint8_t _r_rsize;  /* sign and reloc bit len */
            uint8_t _r_rtype;  /* toc relocation type */
        } _r_r;
    } _r;
} RELOC;

typedef struct _S_(reloc) {
    uint64_t        r_vaddr;   /* (virtual) address of reference */
    uint32_t        r_symndx;  /* index into symbol table */
    union {
        uint16_t    _r_type;   /* old style coff relocation type */
        struct {
            uint8_t _r_rsize;  /* sign and reloc bit len */
            uint8_t _r_rtype;  /* toc relocation type */
        } _r_r;
    } _r;
} _S_(RELOC);

#define r_type     _r._r_type        /* old style field name */
#define r_rsize    _r._r_r._r_rsize  /* extract sign and bit len */
#define r_rtype    _r._r_r._r_rtype  /* extract toc relocation type */

#define R_LEN      0x3F   /* extract bit-length field */
#define R_SIGN     0x80   /* extract sign of relocation */
#define R_FIXUP    0x40   /* extract code-fixup bit */

#define RELOC_RLEN(x)    ((x)._r._r_r._r_rsize & R_LEN)
#define RELOC_RSIGN(x)   ((x)._r._r_r._r_rsize & R_SIGN)
#define RELOC_RFIXUP(x)  ((x)._r._r_r._r_rsize & R_FIXUP)
#define RELOC_RTYPE(x)   ((x)._r._r_r._r_rtype)

/*
 *    POWER and PowerPC - relocation types
 */
#define R_POS    0x00    /* A(sym) Positive Relocation */
#define R_NEG    0x01    /* -A(sym) Negative Relocation */
#define R_REL    0x02    /* A(sym-*) Relative to self */
#define R_TOC    0x03    /* A(sym-TOC) Relative to TOC */
#define R_TRL    0x12    /* A(sym-TOC) TOC Relative indirect load. */
#define R_TRLA   0x13    /* A(sym-TOC) TOC Rel load address. modifiable inst */
#define R_GL     0x05    /* A(external TOC of sym) Global Linkage */
#define R_TCL    0x06    /* A(local TOC of sym) Local object TOC address */
#define R_RL     0x0C    /* A(sym) Pos indirect load. modifiable instruction */
#define R_RLA    0x0D    /* A(sym) Pos Load Address. modifiable instruction */
#define R_REF    0x0F    /* AL0(sym) Non relocating ref. No garbage collect */
#define R_BA     0x08    /* A(sym) Branch absolute. Cannot modify instruction */
#define R_RBA    0x18    /* A(sym) Branch absolute. modifiable instruction */
#define R_BR     0x0A    /* A(sym-*) Branch rel to self. non modifiable */
#define R_RBR    0x1A    /* A(sym-*) Branch rel to self. modifiable instr */

/* UNUSED */
#define R_RBAC   0x19    /* A(sym) Branch absolute constant. modifiable instr */
#define R_RBRC   0x1B    /* A(sym-*) Branch absolute const. */
#define R_RTB    0x04    /* A((sym-*)/2) RT IAR Rel Branch. non modifiable */
#define R_RRTBI  0x14    /* A((sym-*)/2) RT IAR Rel Br. modifiable to R_RRTBA */
#define R_RRTBA  0x15    /* A((sym-*)/2) RT absolute br. modifiable to R_RRTBI */

/*
 *    original style - relocation types from coff
 *    UNUSED: The #defines retained for compatability
 */
#define R_ABS      0
#define R_OFF8     07
#define R_OFF16    010
#define R_SEG12    011
#define R_AUX      013
#define R_DIR16    01
#define R_REL16    02
#define R_IND16    03
#define R_DIR24    04
#define R_REL24    05
#define R_OPT16    014
#define R_IND24    015
#define R_IND32    016
#define R_DIR10    025
#define R_REL10    026
#define R_REL32    027
#define R_DIR32S   012
#define R_RELBYTE  017
#define R_RELWORD  020
#define R_RELLONG  021
#define R_PCRBYTE  022
#define R_PCRWORD  023
#define R_PCRLONG  024

#define    SCNHSZ        sizeof(SCNHDR)
#define    SCNHSZ_64     sizeof(SCNHDR_64)

/* If both s_nreloc and s_nlnno fields are set to _OVERFLOW_MARKER,
   an overflow header contains the real values. */

#define _OVERFLOW_MARKER       (65535)
#define _OVERFLOW_MARKER_64    (4294967295)

typedef struct scnhdr {
    char        s_name[8];  /* section name */
    uint32_t    s_paddr;    /* physical address */
    uint32_t    s_vaddr;    /* virtual address */
    uint32_t    s_size;     /* section size */
    uint32_t    s_scnptr;   /* file ptr to raw data for section */
    uint32_t    s_relptr;   /* file ptr to relocation for section */
    uint32_t    s_lnnoptr;  /* file ptr to line numbers for sect. */
    uint16_t    s_nreloc;   /* number of relocation entries */
    uint16_t    s_nlnno;    /* number of line number entries */
    uint32_t    s_flags;    /* flags */
} SCNHDR;

typedef struct _S_(scnhdr) {
    char        s_name[8];  /* section name */
    uint64_t    s_paddr;    /* physical address */
    uint64_t    s_vaddr;    /* virtual address */
    uint64_t    s_size;     /* section size */
    uint64_t    s_scnptr;   /* file ptr to raw data */
    uint64_t    s_relptr;   /* file ptr to relocation */
    uint64_t    s_lnnoptr;  /* file ptr to line numbers */
    uint32_t    s_nreloc;   /* # of relocation entries */
    uint32_t    s_nlnno;    /* # of line number entries */
    uint32_t    s_flags;    /* flags */
} _S_(SCNHDR);

/*
 * Define constants for names of "special" sections
 */

#define _TEXT    ".text"
#define _DATA    ".data"
#define _BSS     ".bss"
#define _PAD     ".pad"
#define _INFO    ".info"
#define _LOADER  ".loader"
#define _TYPCHK  ".typchk"
#define _DEBUG   ".debug"
#define _EXCEPT  ".except"
#define _OVRFLO  ".ovrflo"

/*
 * The low order 16 bits of s_flags is used as section "type"
 */

#define STYP_REG     0x00    /* "regular" section */
#define STYP_PAD     0x08    /* "padding" section */
#define STYP_TEXT    0x20    /* section contains text only */
#define STYP_DATA    0x40    /* section contains data only */
#define STYP_BSS     0x80    /* section contains bss only */
#define STYP_EXCEPT  0x0100  /* Exception section */
#define STYP_INFO    0x0200  /* Comment section */
#define STYP_LOADER  0x1000  /* Loader section */
#define STYP_DEBUG   0x2000  /* Debug section */
#define STYP_TYPCHK  0x4000  /* Type check section */
#define STYP_OVRFLO  0x8000  /* handling relocation and line number overflows */

#define    SYMESZ    (18)    /* Do not use sizeof(SYMENT) */
#define    SYMESZ_64 (18)    /* Do not use sizeof(SYMENT_64) */

#define    SYMNMLEN    8     /* Number of characters in a symbol name */
#define    FILNMLEN   14     /* Number of characters in a file name */
#define    DIMNUM      4     /* Number of array dimensions in auxiliary entry */

/*
 Note 2: For an XCOFF64 file, the n_type field may only have the values
    0 (if the symbol is not a function) or 0x0020 (if the symbol
    is a function).

 Note 3: All the fields containing "x_auxtype" refer to the same byte
    of the auxiliary symbol table entry.

 Note 4: The XCOFF64 equivalent for x_csect.x_scnlen is the concatenation
    of x_csect.x_scnlen_hi and x_csect.x_scnlen_lo.

 Note 5: When _AUX_EXCEPT and  _AUX_FCN    auxiliary symbol table entries 
    exist for the same primary symbol table entry, the x_fsize and
    x_endndx fields of the two entries should have the same values.

 Note 6: Field names not listed in the table above are obsolete and are
    not used in the current release of AIX.
*/

typedef struct syment {
    union {
        char          _n_name[SYMNMLEN]; /* Symbol name (if 8 characters or less) */
        struct {
            uint32_t  _n_zeroes;       /* new == 0 */
            uint32_t  _n_offset;       /* offset into string table */
        } _n_n;
        uint32_t      _n_nptr[2];      /* allows for overlaying */
    } _n;
    uint32_t          n_value;         /* value of symbol */
    int16_t          n_scnum;         /* section number */
    union {
        uint16_t      _n_type;         /* type and derived type */
        struct {
            uint8_t   _n_lang;         /* source language id */
            uint8_t   _n_cpu;          /* cputype id */
        }_n_lc;
    } _n_tylc;
    uint8_t           n_sclass;        /* storage class */
    uint8_t           n_numaux;        /* number of aux. entries */
} SYMENT;

#define n_name        _n._n_name
#define n_zeroes      _n._n_n._n_zeroes
#define n_nptr        _n._n_nptr[1]

typedef struct _S_(syment) {
    uint64_t          n_value;    /* value of symbol */
    union {
        struct {
            uint32_t  _n_offset;  /* offset into string table */
        } _n_n;
        uint32_t      _n_nptr;    /* Allows for overlaying */
    } _n;
    int16_t          n_scnum;    /* section number */
    union {
        uint16_t      _n_type;    /* type and derived type */
        struct {
            uint8_t   _n_lang;    /* source language id */
            uint8_t   _n_cpu;     /* cputype id */
        }_n_lc;
    } _n_tylc;

    uint8_t           n_sclass;    /* storage class */
    uint8_t           n_numaux;    /* number of aux. entries */
} _S_(SYMENT);

#define n_offset      _n._n_n._n_offset

#ifndef n_type
#define n_type        _n_tylc._n_type
#endif  

#define n_cputype     _n_tylc._n_lc._n_cpu
#define n_lang        _n_tylc._n_lc._n_lang

/* Structure size is the same for both XCOFF32 and XCOFF64 */
/*
 * Relocatable symbols have a section number of the
 * section in which they are defined. Otherwise, section
 * numbers have the following meanings:
 */
#define    N_UNDEF     0     /* undefined symbol */
#define    N_ABS      -1     /* value of symbol is absolute */
#define    N_DEBUG    -2     /* special debugging symbol */

/* Obsolete definitions--not defined for XCOFF64. */
/*
 * The fundamental type of a symbol packed into the low
 * 4 bits of the word.
 * Unused in xcoff.
 */
#define    T_NULL     0
#define    T_ARG      1    /* function argument (only used by compiler) */
#define    T_CHAR     2    /* character */
#define    T_SHORT    3    /* short integer */
#define    T_INT      4    /* integer */
#define    T_LONG     5    /* long integer */
#define    T_FLOAT    6    /* floating point */
#define    T_DOUBLE   7    /* double word */
#define    T_STRUCT   8    /* structure */
#define    T_UNION    9    /* union */
#define    T_ENUM    10    /* enumeration */
#define    T_MOE     11    /* member of enumeration */
#define    T_UCHAR   12    /* unsigned character */
#define    T_USHORT  13    /* unsigned short */
#define    T_UINT    14    /* unsigned integer */
#define    T_ULONG   15    /* unsigned long */

/*
 * derived types are:
 */
#define DT_NON    0    /* no derived type */
#define DT_PTR    1    /* pointer */
#define DT_FCN    2    /* function */
#define DT_ARY    3    /* array */

/*
 *    type packing constants
 */
#define N_BTMASK     017
#define N_TMASK      060
#define N_TMASK1    0300
#define N_TMASK2    0360
#define N_BTSHFT       4
#define N_TSHIFT       2

#ifndef _COMPILER
/*
 *    MACROS
 */
#define BTYPE(x)  ((x) & N_BTMASK)                                /* Basic Type of x */
#define ISPTR(x)  (((x) & N_TMASK) == (DT_PTR << N_BTSHFT))       /* Is x a pointer? */
#define ISFCN(x)  (((x) & N_TMASK) == (DT_FCN << N_BTSHFT))       /* Is x a function? */
#define ISARY(x)  (((x) & N_TMASK) == (DT_ARY << N_BTSHFT))       /* Is x an array? */
#define ISTAG(x)  ((x)==C_STRTAG || (x)==C_UNTAG || (x)==C_ENTAG) /* Is x a structure, union, or enumeration TAG? */
#define INCREF(x) ((((x)&~N_BTMASK)<<N_TSHIFT)|(DT_PTR<<N_BTSHFT)|(x&N_BTMASK))
#define DECREF(x) ((((x)>>N_TSHIFT)&~N_BTMASK)|((x)&N_BTMASK))
#endif 


/*************************************************************************
 *
 *    AUXILIARY ENTRY FORMAT
 *
 *************************************************************************/

/* Structure name and size are the same for both XCOFF32 and XCOFF64. */

#define    AUXESZ      (18)    /* Do not use sizeof(AUXENT) */
#define    AUXESZ_64   (18)    /* Do not use sizeof(AUXENT_64) */

typedef union auxent {
 struct {
        uint32_t    x_tagndx;    /* str, un, or enum tag indx */
                                 /* exception table offset */
#define x_exptr     x_tagndx
        union {
            struct {
                uint16_t    x_lnno;    /* declaration line number */
                uint16_t    x_size;    /* str, union, array size */
            } x_lnsz;
            uint32_t        x_fsize;   /* size of function */
        } x_misc;
        union {
            struct {                   /* if ISFCN, tag, or .bb */
                uint32_t    x_lnnoptr; /* ptr to fcn line # */
                uint32_t    x_endndx;  /* entry ndx past block end */
            } x_fcn;
            struct {                   /* if ISARY, up to 4 dimen. */
                uint16_t    x_dimen[DIMNUM];
            } x_ary;
        } x_fcnary;
        uint16_t            x_tvndx;   /* tv index */
    } x_sym;

    union {
        char            x_fname[FILNMLEN];
        struct {
            uint32_t    x_zeroes;
            uint32_t    x_offset;
            char        x_pad[FILNMLEN-8];
            uint8_t     x_ftype;
        } _x;
    } x_file;

    struct {
        uint32_t    x_scnlen; /* section length */
        uint16_t    x_nreloc; /* number of relocation entries */
        uint16_t    x_nlinno; /* number of line numbers */
    } x_scn;

    /* CSECT auxiliary entry */
    struct {
        uint32_t    x_scnlen;    /* csect length */
        uint32_t    x_parmhash;  /* parm type hash index */
        uint16_t    x_snhash;    /* sect num with parm hash */
        uint8_t     x_smtyp;     /* symbol align and type */
                                 /* 0-4 - Log 2 of alignment */
                                 /* 5-7 - symbol type */
        uint8_t     x_smclas;           /* storage mapping class */
        uint32_t    x_stab;      /* dbx stab info index */
        uint16_t    x_snstab;    /* sect num with dbx stab */
    } x_csect;
} AUXENT;

typedef union _S_(auxent) {
    struct {
        struct {
            struct {
                uint32_t  x_lnno;       /* declaration line number */
            } x_lnsz;
        } x_misc;
        char              pad[13];
        uint8_t           x_auxtype;    /* Auxiliary symbol type */
    } x_sym;

    union {
        char              x_fname[FILNMLEN];
        struct {
            uint32_t      x_zeroes;
            uint32_t      x_offset;
            char          x_pad[FILNMLEN-8];
            uint8_t       x_ftype;
            char          pad[2];
            uint8_t       x_auxtype; /* Auxiliary symbol type */
        } _x;
    } x_file;

    /* CSECT auxiliary entry */
    struct {
        uint32_t    x_scnlen_lo; /* csect length */
        uint32_t    x_parmhash;  /* parm type hash index */
        uint16_t    x_snhash;    /* sect num with parm hash */
        uint8_t     x_smtyp;     /* symbol align and type */
                                 /* 0-4 - Log 2 of alignment */
                                 /* 5-7 - symbol type */
        uint8_t     x_smclas;    /* storage mapping class */
        uint32_t    x_scnlen_hi; 
        char        pad;
        uint8_t     x_auxtype;   /* Auxiliary symbol type */
    } x_csect;
    struct {
        uint64_t    x_exptr;     /* exception table offset */
        uint32_t    x_fsize;     /* size of function */
        uint32_t    x_endndx;    /* entry ndx past block end */
        char        pad;
        uint8_t     x_auxtype;   /* Auxiliary symbol type */
    } x_except;
    struct {
        uint64_t    x_lnnoptr;   /* line number entry offset */
        uint32_t    x_fsize;     /* size of function */
        uint32_t    x_endndx;    /* entry ndx past block end */
        char        pad;
        uint8_t     x_auxtype;   /* Auxiliary symbol type */
    } x_fcn;
    struct {
        char        x_pad[AUXESZ-1];
        uint8_t     x_auxtype;
    } x_auxtype;
} _S_(AUXENT);


#define _AUX_EXCEPT    255
#define _AUX_FCN       254
#define _AUX_SYM       253
#define _AUX_FILE      252
#define _AUX_CSECT     251


/* Defines for File auxiliary definitions: x_ftype field of x_file */
#define    XFT_FN    0    /* Source File Name */
#define    XFT_CT    1    /* Compile Time Stamp */
#define    XFT_CV    2    /* Compiler Version Number */
#define    XFT_CD  128    /* Compiler Defined Information */

/* Defines for CSECT auxiliary definitions */
/* Symbol Type (5-7 of x_smtyp field of x_csect) */
#define XTY_ER    0    /* External Reference */
#define XTY_SD    1    /* CSECT Section Definition */
#define XTY_LD    2    /* Entry Point - Label Definition */
#define XTY_CM    3    /* Common (BSS) */

/* Following are unused but retained for source compatability when compiling for 32-bit . */
#define XTY_HL    6    /* Hidden Label Definition */
#define XTY_US    5    /* Unset */
#define XTY_EM    4    /* Error Message - Linkedit usage */

/* Storage Mapping Class definitions: x_smclas field of x_csect */
/* READ ONLY CLASSES */
#define XMC_PR      0    /* Program Code */
#define XMC_RO      1    /* Read Only Constant */
#define XMC_DB      2    /* Debug Dictionary Table */
#define XMC_GL      6    /* Global Linkage (Interfile Interface Code) */
#define XMC_XO      7    /* Extended Operation (Pseudo Machine Instruction */
#define XMC_SV      8    /* Supervisor Call (32-bit process only) */
#define XMC_SV64   17    /* Supervisor Call for 64-bit process */
#define XMC_SV3264 18    /* Supervisor Call for both 32- and 64-bit processes. */
#define XMC_TI     12    /* Traceback Index csect */
#define XMC_TB     13    /* Traceback table csect */
/* READ WRITE CLASSES */
#define XMC_RW      5    /* Read Write Data */
#define XMC_TC0    15    /* TOC Anchor for TOC Addressability */
#define XMC_TC      3    /* General TOC item */
#define XMC_TD     16    /* Scalar data item in the TOC */ 
#define XMC_DS     10    /* Descriptor csect */
#define XMC_UA      4    /* Unclassified - Treated as Read Write */
#define XMC_BS      9    /* BSS class (uninitialized static internal) */
#define XMC_UC     11    /* Un-named Fortran Common */

/* Defines for "special" symbols */

#define _ETEXT  "etext"
#define _EDATA  "edata"
#define _END    "end"
#define _EF     ".ef"
#define _START  "__start"

/*
 *    STORAGE CLASSES - n_sclass field of syment
 */

#define C_EFCN      255    /* physical end of function */
#define C_NULL        0    /* logically deleted symbol table entry */
#define C_AUTO        1    /* automatic variable */
#define C_EXT         2    /* external symbol */
#define C_STAT        3    /* static */
#define C_REG         4    /* register variable */
#define C_EXTDEF      5    /* external definition */
#define C_LABEL       6    /* label */
#define C_ULABEL      7    /* undefined label */
#define C_MOS         8    /* member of structure */
#define C_ARG         9    /* function argument */
#define C_STRTAG     10    /* structure tag */
#define C_MOU        11    /* member of union */
#define C_UNTAG      12    /* union tag */
#define C_TPDEF      13    /* type definition */
#define C_USTATIC    14    /* undefined static */
#define C_ENTAG      15    /* enumeration tag */
#define C_MOE        16    /* member of enumeration */
#define C_REGPARM    17    /* register parameter */
#define C_FIELD      18    /* bit field */
#define C_BLOCK     100    /* ".bb" or ".eb" */
#define C_FCN       101    /* ".bf" or ".ef" */
#define C_EOS       102    /* end of structure */
#define C_FILE      103    /* file name */
#define C_LINE      104
#define C_ALIAS     105    /* duplicate tag */
#define C_HIDDEN    106    /* special storage class for external */
/* symbols in dmert public libraries */
#define C_HIDEXT    107    /* Un-named external symbol */
#define C_BINCL     108    /* Marks beginning of include file */
#define C_EINCL     109    /* Marks ending of include file */
#define C_INFO      110    /* Comment string in .info section */
#define C_WEAKEXT   111    /* weak external symbol */


/* Only valid n_value for a C_NULL (deletable) symbol table entry */
#define C_NULL_VALUE    0x00DE1E00

/* Source Language Identifiers */

#ifndef TB_C
/* Following are also defined in <sys/debug.h> */
#define TB_C              0    /* C */
#define TB_FORTRAN        1    /* Fortran */
#define TB_PASCAL         2    /* Pascal */
#define TB_ADA            3    /* Ada */
#define TB_PLI            4    /* PL/I */
#define TB_BASIC          5    /* BASIC */
#define TB_LISP           6    /* LISP */
#define TB_COBOL          7    /* COBOL */
#define TB_MODULA2        8    /* Modula2 */
#define TB_CPLUSPLUS      9    /* C++ */
#define TB_RPG           10    /* RPG */
#define TB_PL8           11    /* PL8, PLIX */
#define TB_ASM           12    /* Assembly */
#define TB_HPJ           13    /* Java */
#define TB_RESERVED_1   251
#define TB_RESERVED_2   252
#define TB_RESERVED_3   253
#define TB_RESERVED_4   254
#define TB_RESERVED_5   255
#endif

#define TB_OBJECT       248    /* Object File */
#define TB_FRONT        249    /* entries collected to start of symbol table */
#define TB_BACK         250    /* entries collected to end of symbol table */

/*
 *    XCOFF STORAGE CLASSES AND STABSTRINGS DESIGNED SPECIFICALLY FOR DBX
 */
#define DBXMASK      0x80

#define C_GSYM       0x80
#define C_LSYM       0x81
#define C_PSYM       0x82
#define C_RSYM       0x83
#define C_RPSYM      0x84
#define C_STSYM      0x85
#define C_TCSYM      0x86
#define C_BCOMM      0x87
#define C_ECOML      0x88
#define C_ECOMM      0x89
#define C_DECL       0x8c
#define C_ENTRY      0x8d
#define C_FUN        0x8e
#define C_BSTAT      0x8f
#define C_ESTAT      0x90

#define TP_ARRAY {\
    "int:t-1=r-1;-2147483648;2147483647",\
    "char:t-2=@s8;r-2;0;255",\
    "short:t-3=@s16;r-3;-32768;32767",\
    "long:t-4=-1",\
    "unsigned char:t-5=@s8;r-5;0;255",\
    "signed char:t-6=@s8;r-6;-128;127",\
    "unsigned short:t-7=@s16;r-7;0;65535",\
    "unsigned int:t-8=r-8;0;4294967295",\
    "unsigned:t-9=-8",\
    "unsigned long:t-10=-8",\
    "void:t-11=r-11;0;0",\
    "float:t-12=g-12;4",\
    "double:t-13=g-12;8",\
    "long double:t-14=g-12;16",\
    "integer:t-15=-1",\
    "boolean:t-16=efalse:0,true:1,;",\
    "shortreal:t-17=g-12;4",\
    "real:t-18=g-12;8",\
    "stringptr:t-19=N-19",\
    "character:t-20=@s8;r-20;0;255",\
    "logical*1:t-21=@s8;r-21;0;255",\
    "logical*2:t-22=@s16;r-22;0;65535",\
    "logical*4:t-23=r-23;0;4294967295",\
    "logical:t-24=-23",\
    "complex:t-25=c-25;8",\
    "double complex:t-26=c-25;16",\
    "integer*1:t-27=-6",\
    "integer*2:t-28=-3",\
    "integer*4:t-29=-1",\
    "wchar:t-30=@s16;r-30;0;65535", \
    "long long:t-31=r-31;-9223372036854775808;9223372036854775807", \
    "unsigned long long:t-32=r-32;0;18446744073709551615", \
    "logical*8:t-33=r-33;0;18446744073709551615",\
    "integer*8:t-34=-31",\
    "long:t-35=r-35;-9223372036854775808;9223372036854775807", \
    "unsigned long:t-36=r-36;0;18446744073709551615", \
    }

#define TP_INT         (-1)
#define TP_CHAR        (-2)
#define TP_SHORT       (-3)
#define TP_LONG        (-4)
#define TP_UCHAR       (-5)
#define TP_SCHAR       (-6)
#define TP_USHORT      (-7)
#define TP_UINT        (-8)
#define TP_UNSIGNED    (-9)
#define TP_ULONG      (-10)
#define TP_VOID       (-11)
#define TP_FLOAT      (-12)
#define TP_DOUBLE     (-13)
#define TP_LDOUBLE    (-14)
#define TP_PASINT     (-15)
#define TP_BOOL       (-16)
#define TP_SHRTREAL   (-17)
#define TP_REAL       (-18)
#define TP_STRNGPTR   (-19)
#define TP_FCHAR      (-20)
#define TP_LOGICAL1   (-21)
#define TP_LOGICAL2   (-22)
#define TP_LOGICAL4   (-23)
#define TP_LOGICAL    (-24)
#define TP_COMPLEX    (-25)
#define TP_DCOMPLEX   (-26)
#define TP_INTEGER1   (-27)
#define TP_INTEGER2   (-28)
#define TP_INTEGER4   (-29)
#define TP_WCHAR      (-30)
#define TP_LLONG      (-31)
#define TP_ULLONG     (-32)
#define TP_LOGICAL8   (-33)
#define TP_INTEGER8   (-34)
#define TP_LONG_64    (-35)
#define TP_ULONG_64   (-36)
#define TP_NTYPES     36
