#ifndef _XCoffFileGen_h_
#define _XCoffFileGen_h_

#include <Base.h>
#include <BinaryFile.h>
#include <Instruction.h>

class FileHeader;
class AOutHeader;
class SectHeader;
class RawSection;
class RelocationTable;
class LineInfoTable;
class SymbolTable;
class StringTable;
class XCoffFileGen;
class TextSection;
class DataSection;

class BasicBlock;

class FileHeaderGen : public BaseGen {
protected:
    FileHeader* fileHeader;
public:
    FileHeaderGen(FileHeader* fh) : BaseGen(XCoffClassTypes_file_header),fileHeader(fh) {}
    virtual ~FileHeaderGen() {}
    Base* getParsedBase() { return (Base*)fileHeader; }
};

class AOutHeaderGen : public BaseGen {
protected:
    AOutHeader* aOutheader;
public:
    AOutHeaderGen(AOutHeader* ah) : BaseGen(XCoffClassTypes_aout_header),aOutheader(ah) {}
    virtual ~AOutHeaderGen() {}
    Base* getParsedBase() { return (Base*)aOutheader; }
};

class SectHeaderGen : public BaseGen {
protected:
    SectHeader* sectHeader;
public:
    SectHeaderGen(SectHeader* sh) : BaseGen(XCoffClassTypes_sect_header),sectHeader(sh) {}
    virtual ~SectHeaderGen() {}
    Base* getParsedBase() { return (Base*)sectHeader; }
};

class RawSectionGen : public BaseGen {
protected:
    RawSection* rawSection;
public:
    RawSectionGen(RawSection* rs) : BaseGen(XCoffClassTypes_sect_rawdata),rawSection(rs) {}
    virtual ~RawSectionGen() {}
    uint32_t setFileOffset(uint32_t offset);
    Base* getParsedBase() { return (Base*)rawSection; }
};

class RelocationTableGen : public BaseGen {
protected:
    RelocationTable* relocationTable;
public:
    RelocationTableGen(RelocationTable* rt) : BaseGen(XCoffClassTypes_relocation),relocationTable(rt) {}
    virtual ~RelocationTableGen() {}
    Base* getParsedBase() { return (Base*)relocationTable; }
};

class LineInfoTableGen : public BaseGen {
protected:
    LineInfoTable* lineInfoTable;
public:
    LineInfoTableGen(LineInfoTable* li) : BaseGen(XCoffClassTypes_line_info),lineInfoTable(li) {}
    virtual ~LineInfoTableGen() {}
    Base* getParsedBase() { return (Base*)lineInfoTable; }
};

class SymbolTableGen : public BaseGen {
protected:
    SymbolTable* symbolTable;
public:
    SymbolTableGen(SymbolTable* st) : BaseGen(XCoffClassTypes_symbol_table),symbolTable(st) {}
    virtual ~SymbolTableGen() {}
    Base* getParsedBase() { return (Base*)symbolTable; }
};

class StringTableGen : public BaseGen {
protected:
    StringTable* stringTable;
public:
    StringTableGen(StringTable* st) : BaseGen(XCoffClassTypes_string_table),stringTable(st) {}
    virtual ~StringTableGen() {}
    Base* getParsedBase() { return (Base*)stringTable; }
};

class XCoffFileGen {
protected:

    XCoffFile*              xcoffFile;

    char*                   inputFileName;
    char*                   outputFileName;
    BinaryOutputFile        binaryOutputFile;
    char*                   fileExtension;
    uint32_t                phaseIndex;

    FileHeaderGen*          fileHeaderGen;
    AOutHeaderGen*          aOutHeaderGen;

    uint32_t                numberOfSections;
    SectHeaderGen**         sectHeadersGen;
    RawSectionGen**         rawSectionsGen;
    RelocationTableGen**    relocationTablesGen;
    LineInfoTableGen**      lineInfoTablesGen;

    SymbolTableGen*         symbolTableGen;
    StringTableGen*         stringTableGen;
    
    uint32_t                instrumentedFileSize;
    uint32_t                numberOfInstPoints;

    uint32_t                extendedDataSize;
    uint32_t                dataBufferEntryCount;

    char*                   pathToInstLibraries;

    void inst_step1_allocateBuffers();
    void inst_step2_setFileOffsets();
    void inst_step3_instrumentInBuffer();
    void inst_step4_updateBSSCSectionEntries();
    
    XCoffFileGen(XCoffFile* xcoff,char* extension,uint32_t phaseNo,char* blockFile);

public:

    virtual ~XCoffFileGen() { }

    char*    getFileExtension()        { return fileExtension; }
    uint32_t getInstrumentedFileSize() { return instrumentedFileSize; }
    uint32_t getPhaseIndex()           { return phaseIndex; }

    uint32_t getNumberOfAllFunctions();
    uint32_t getNumberOfAllBlocks();
    uint32_t getNumberOfAllMemoryOps();
    uint32_t getNumberOfAllFloatPOps();
    uint32_t getNumberOfAllObjects();
    uint32_t getAllObjects(BaseGen** arr,uint32_t s);

    uint32_t getSymbolTableOffsetForInst();
    uint32_t getRawSectionSizeForInst(uint32_t idx);
    uint32_t getRawSectionOffsetForInst(uint32_t idx);
    uint32_t getRelocOffsetForInst(uint32_t idx);
    uint32_t getLineInfoOffsetForInst(uint32_t idx);
    uint64_t getNewBSSSectionVAddressForInst();

    virtual void selectInstrumentationPoints(char* blockFile) { __SHOULD_NOT_ARRIVE; }
    virtual void printInstrumentationPoints() { __SHOULD_NOT_ARRIVE; }
    virtual uint32_t getNumberOfInstPoints(TextSection* textSect) { return numberOfInstPoints; }

    uint32_t        getExtendedDataSize()                  { return extendedDataSize; }
    uint32_t        getDataBufferEntryCount()              { return dataBufferEntryCount; }
    virtual int32_t getDataOffsetForInst(uint32_t idx) {  __SHOULD_NOT_ARRIVE; return 0; }
    virtual void    reserveDataForInstrumentation()        { __SHOULD_NOT_ARRIVE; }

    virtual void verify(uint32_t written) { __SHOULD_NOT_ARRIVE; }

    virtual uint32_t getNumOfSharedLibFuncs()    { __SHOULD_NOT_ARRIVE; return 0; }
    virtual uint32_t byteCountForSharedLibFuncWrappers() { __SHOULD_NOT_ARRIVE; return 0; }
    virtual uint32_t generateSharedLibFuncWrapper(uint32_t libFuncIdx,uint64_t funcCallAddr,uint32_t genBufferOffset,BaseGen* gen) {
        __SHOULD_NOT_ARRIVE; return 0;
    }
    virtual void     setAddrOfSharedLibFuncWrapper(uint32_t i,uint64_t addr){ __SHOULD_NOT_ARRIVE; }
    uint32_t         generateStubForAllLibraryCalls(uint32_t genBufferOffset,BaseGen* gen);
    virtual char*    getSharedLibFuncName(uint32_t i) { __SHOULD_NOT_ARRIVE; return NULL; }
    virtual uint64_t getSharedLibFuncAddrLocation(uint32_t i) { __SHOULD_NOT_ARRIVE; return 0; }
    virtual void     getSharedLibraryPathAndObj(char** path,char** name,char** obj){ __SHOULD_NOT_ARRIVE; }

    virtual void initializeReservedData(DataSection* dataSect,BaseGen* gen) { __SHOULD_NOT_ARRIVE; }

    virtual uint32_t maxByteCountPerInst() { __SHOULD_NOT_ARRIVE; return 0; }
    virtual uint32_t byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect) { 
        __SHOULD_NOT_ARRIVE; return 0; 
    }
    virtual uint32_t generateCodeForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect,
                                         BaseGen* gen,uint32_t genBufferOffset) {
        __SHOULD_NOT_ARRIVE; return 0; 
    }

    void         instrument();
    uint32_t     dump();

    virtual char* getGenClassName() { __SHOULD_NOT_ARRIVE; return NULL; }

    char* getPathToInstLib() { return pathToInstLibraries; }
    void setPathToInstLib(char* libPathTop);
};

#endif
