#ifndef _Iterator_h_
#define _Iterator_h_

#include <Base.h>

class AddressIterator {
protected:
    uint64_t baseAddress;
    uint64_t sizeInBytes;
    uint64_t currAddress;
    uint8_t  unitInBytes;

    inline AddressIterator(uint64_t base,uint64_t size,uint8_t unit) : 
            baseAddress(base),sizeInBytes(size),currAddress(base),unitInBytes(unit)
            {}

public:
    virtual ~AddressIterator() {}
    bool hasMore() { 
        return ((baseAddress <= currAddress) && (currAddress < (baseAddress + sizeInBytes))); 
    }
    bool hasPrev() {
        return ((baseAddress < currAddress) && (currAddress <= (baseAddress + sizeInBytes))); 
    }
    inline void reset() { currAddress = baseAddress; }
    inline uint8_t unitSize() { return unitInBytes; }
    inline uint64_t operator*() { return currAddress; }
    uint64_t getOffset(){ return (currAddress-baseAddress); }

    uint64_t operator++();
    uint64_t operator--();
    uint64_t operator++(int n);
    uint64_t operator--(int n);

    void skip(uint64_t s) { currAddress += s; }
    void skipTo(uint64_t s) { currAddress = s; }

    static AddressIterator newAddressIteratorWord(uint64_t base,uint64_t size){
        return AddressIterator(base,size,sizeof(uint32_t));
    }

    static AddressIterator newAddressIteratorDouble(uint64_t base,uint64_t size){
        return AddressIterator(base,size,sizeof(uint64_t));
    }

    static AddressIterator invalidIterator(){
        return AddressIterator(0,0,0);
    }

    bool isInvalid() { return !unitInBytes; }

    uint64_t readBytes(char* ptr);
    void print() { PRINT_INFOR("AI : %#18llx - %6lld - %#18llx",baseAddress,sizeInBytes,currAddress); }

    uint8_t unit() { return unitInBytes; }

    bool isWord() { return (unitInBytes == sizeof(uint32_t)); }
    bool isDouble() { return (unitInBytes == sizeof(uint64_t)); }
};

#endif
