#ifndef _LineInfoFinder_h_
#define _LineInfoFinder_h_

#include <Base.h>
#include <LineInfoTable.h>
#include <SymbolTable.h>

#define C_FCN_BEGIN_NAME ".bf"
#define C_FCN_END_NAME ".ef"
#define CPP_FILE_FORWARD_OFFSET_TO_NAME 1
#define CPP_GENERIC_FILE_SYM_NAME ".file"
#define MAX_LINE_TEST 100
#define MAX_LINE_VALUE 0x7fffffff
#define FILE_TYPE_SOURCE 0
#define FILE_TYPE_INCLUDE 1

#define GET_IS_INCL_FILE_BIT(i) ((isIncludeFile[i / (sizeof(uint16_t))] >> (i % sizeof(uint16_t))) & 0x0001)
#define SET_IS_INCL_FILE_BIT(i) (isIncludeFile[i / (sizeof(uint16_t))] = isIncludeFile[i / (sizeof(uint16_t))] | (1 << (i % sizeof(uint16_t))))
#define GET_IS_INCL_ADDR_BIT(i) ((isAddressFromIncludeFile[i / (sizeof(uint16_t))] >> (i % sizeof(uint16_t))) & 0x0001)
#define SET_IS_INCL_ADDR_BIT(i) (isAddressFromIncludeFile[i / (sizeof(uint16_t))] = isAddressFromIncludeFile[i / (sizeof(uint16_t))] | (1 << (i % sizeof(uint16_t))))


typedef struct idx_symndx {
    uint32_t index;
    uint32_t symndx;
};

typedef struct idx_addr {
    uint32_t index;
    uint64_t address;
};

int32_t compare_idx_addr(const void* a, const void* b);

class FileFinder;

class FunctionFinder : public Base {
protected:
    char* functionName;
    uint32_t beginLineInfoIndex;
    uint32_t endLineInfoIndex;
    uint32_t firstLine;
    bool fromIncludeFile;

    FileFinder* parentFile;

    StringTable* stringTable;
    LineInfoTable* lineInfoTable;
    SymbolTable* symbolTable;
public:
    FunctionFinder(SymbolTable* symTable, LineInfoTable* linTable, StringTable* strTable, uint32_t idx, 
        struct idx_symndx* lineInfoFunctions, uint32_t numberOfLineInfoFunctions);
    ~FunctionFinder();
    FileFinder* setParent(FileFinder** files, uint32_t numberOfFiles);
    char* getName() { return functionName; }
    bool isFromIncludeFile() { return fromIncludeFile; }
    uint32_t getBeginLineInfoIndex() {return beginLineInfoIndex; }
    uint32_t getEndLineInfoIndex() { return endLineInfoIndex; }
    uint32_t getFirstLine() {return firstLine; }
    bool containsLineNumber(uint32_t lineno);
    FileFinder* getParentFile() { return parentFile; }
    uint64_t getAddressByLineNumber(uint32_t lineno);
    uint64_t getAddressByLineNumber(uint32_t lineno, uint64_t addr);
    void print();
};


class FileFinder : public Base {
protected:
    char* fileName;
    uint32_t fileType;
    uint32_t symbolIndex;
    FunctionFinder** memberFunctions;
    uint32_t numberOfFunctions;
    bool includeFile;

    StringTable* stringTable;
    LineInfoTable* lineInfoTable;
    SymbolTable* symbolTable;
public:
    ~FileFinder() { delete memberFunctions; }
    char* getName() { return fileName; }
    uint32_t getSymbolIndex() { return symbolIndex;}
    uint32_t getFileType() { return fileType; }
    bool isIncludeFile() { return includeFile; }
    bool isSourceFile() { return !includeFile; }
    uint32_t setNumberOfFunctions(uint32_t numberFunctions);
    uint32_t addFunction(FunctionFinder* child, uint32_t idx);

    virtual void print() { __SHOULD_NOT_ARRIVE; }
    virtual uint32_t getNumberOfFunctions() { __SHOULD_NOT_ARRIVE; return 0; }
};

class IncludeFileFinder : public FileFinder {
protected:
    FileFinder* parentFile;
    uint32_t lineInfoIndex;
    uint32_t endLineInfoIndex;

public:
    IncludeFileFinder(SymbolTable* symTable, LineInfoTable* linTable, StringTable* strTable, uint32_t idx, uint32_t linIndex, uint32_t endLinIndex);
    ~IncludeFileFinder();
    uint32_t getLineInfoIndex() { return lineInfoIndex; }
    uint32_t getEndLineInfoIndex() { return endLineInfoIndex; }
    FileFinder* getParentFile() { return parentFile; }
    bool containsAddress(uint64_t addr);
    void print();
};


class SourceFileFinder : public FileFinder {
protected:
    uint32_t nameIndex;
    bool cppNameScheme;

public:
    SourceFileFinder(SymbolTable* symTable, LineInfoTable* linTable, StringTable* strTable, uint32_t idx);
    ~SourceFileFinder();
    bool containsLineNumber(uint32_t lineno);
    char* getName() { return fileName; }
    uint32_t getSymbolIndex() { return symbolIndex; }
    uint32_t getNameIndex() { return nameIndex; }
    bool isCppNameScheme() { return cppNameScheme; }
    uint32_t getNumberOfFunctions() { return numberOfFunctions; }
    void print();
};


class LineInfoFinder : public Base {
protected:
    SymbolTable* symbolTable;
    StringTable* stringTable;
    uint64_t lineInfoPointer;
    LineInfoTable* lineInfoTable;
    bool is64Bit;

        /* don't really need this stuff -- helps with efficiency */
    LineInfo** lineInfos;
    uint32_t numberOfLineInfos;
    uint32_t* lineInfoLineNumbers;
    uint64_t* lineInfoAddresses;
    uint16_t* isIncludeFile;
    bool* isIncludeFile2;
    uint16_t* isAddressFromIncludeFile;
    bool* isAddressFromIncludeFile2;
    struct idx_addr* sortedAddresses;
    uint32_t numberOfLineAddresses;
    struct idx_symndx* lineInfoFunctions;
    uint32_t numberOfLineInfoFunctions;

    uint32_t numberOfFiles;    
    uint32_t numberOfFunctions;
    FileFinder** files;
    FunctionFinder** functions;

public:
    LineInfoFinder(uint32_t idx, XCoffFile* xCoffFile);
    ~LineInfoFinder();

    uint32_t getLineInfoIndex(uint64_t fileOffset);
    uint32_t getLineInfoIndexOfAddr(uint64_t addr);
    uint32_t getLineInfoIndexOfPreviousAddr(uint64_t addr);

    uint32_t getNumberOfFiles() { return numberOfFiles; }
    uint32_t getNumberOfFunctions() { return numberOfFunctions; }

    uint32_t getLineNumberInFunction(uint64_t addr);
    uint32_t getLineNumberInFile(uint64_t addr);

    bool isAddressInLineInfoTable(uint64_t addr);
    bool isFileNameInSymbolTable(char* fileName);
    bool isFunctionNameInSymbolTable(char* fcnName);
    bool functionContainsAddress(uint32_t idx, uint64_t addr);

    char* getFileName(uint64_t addr);
    char* getFunctionName(uint64_t addr);

    uint64_t getAddressByFile(uint32_t lineno, char* fileName);
    uint64_t getAddressByFunction(uint32_t lineno, char* fcnName);
    uint64_t getAddressByFile(uint32_t lineno, char* fileName, uint64_t addr);
    uint64_t getAddressByFunction(uint32_t lineno, char* fcnName, uint64_t addr);

    bool isLineInFile(uint32_t lineno, char* fileName);
    bool isLineInFunction(uint32_t lineno, char* fcnName);

    char* getFileNameOfFunction(char* fcnName);

// these functions have side effects -- namely that they print
// stuff to stdout. These are to help debug and test this class.
    void printFileSymbols();
    void printFiles();
    void printFunctions();
    void printSymbolTable();
    void printLineInfoTable();
    void printFunctionSymbols();
    bool testLineInfoFinder();
    void commandLineTest();
};

#endif // _LineInfoFinder_h_
