#ifndef _PowerPCInstruct_h_
#define _PowerPCInstruct_h_

#include <Base.h>
#include <defines/PowerPCInstruct.d>

typedef struct {
    uint32_t opc    : 6;
    uint32_t frt    : 5;
    uint32_t fra    : 5;
    uint32_t frb    : 5;
    uint32_t frc    : 5;
    uint32_t xo     : 5;
    uint32_t rc     : 1;
} AForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t bo     : 5;
    uint32_t bi     : 5;
    int32_t  bd     :14;
    uint32_t aa     : 1;
    uint32_t lk     : 1;
} BForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rt_s_bf: 5;
    uint32_t ra     : 5;
    int32_t  d_ui_si:16;
} DForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rt_s   : 5;
    uint32_t ra     : 5;
    int32_t  ds     :14;
    uint32_t xo     : 2;
} DsForm;

typedef struct {
    uint32_t opc    : 6;
    int32_t  li     :24;
    uint32_t aa     : 1;
    uint32_t lk     : 1;
} IForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rs     : 5;
    uint32_t ra     : 5;
    uint32_t rb_sh  : 5;
    uint32_t mb     : 5;
    uint32_t me     : 5;
    uint32_t rc     : 1;
} MForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rs     : 5;
    uint32_t ra     : 5;
    uint32_t sh     : 5;
    uint32_t mb_me  : 6;
    uint32_t xo     : 3;
    uint32_t sh2    : 1;
    uint32_t rc     : 1;
} MdForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rs     : 5;
    uint32_t ra     : 5;
    uint32_t rb     : 5;
    uint32_t mb_me  : 6;
    uint32_t xo     : 4;
    uint32_t rc     : 1;
} MdsForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t rt_s_bf: 5;
    uint32_t ra_sr  : 5;
    uint32_t rb_sh  : 5;
    uint32_t xo     :10;
    uint32_t rc     : 1;
} XForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t bt_bo  : 5;
    uint32_t ba_bi  : 5;
    uint32_t bb     : 5;
    uint32_t xo     :10;
    uint32_t lk     : 1;
} XlForm;

typedef struct {
    uint32_t opc    : 6;
    uint32_t res1   : 1;
    uint32_t flm    : 8;
    uint32_t res2   : 1;
    uint32_t frb    : 5;
    uint32_t xo     :10;
    uint32_t rc     : 1;
} XflForm;

typedef struct {
    uint32_t opc     : 6;
    uint32_t rt_rs   : 5;
    uint32_t spr_tbr :10;
    uint32_t xo      :10;
    uint32_t rc      : 1;
} XfxForm;

typedef struct {
    uint32_t opc  : 6;
    uint32_t rt   : 5;
    uint32_t ra   : 5;
    uint32_t rb   : 5;
    uint32_t oe   : 1;
    uint32_t xo   : 9;
    uint32_t rc   : 1;
} XoForm;

typedef struct {
    uint32_t opc  : 6;
    uint32_t rs   : 5;
    uint32_t ra   : 5;
    uint32_t sh1  : 5;
    uint32_t xo   : 9;
    uint32_t sh2  : 1;
    uint32_t rc   : 1;
} XsForm;

typedef struct {
    uint32_t opc  : 6;
    uint32_t rest :26;
} OcForm;

typedef struct {
    uint32_t opc  : 6; 
    uint32_t rest : 24; 
    uint32_t aa   : 1;
    uint32_t lk   : 1;
} LkForm;

typedef union {
    AForm        a;
    BForm        b;
    DForm        d;
    DsForm       ds;
    IForm        i;
    MForm        m;
    MdForm       md;
    MdsForm      mds;
    XForm        x;
    XflForm      xfl;
    XfxForm      xfx;
    XlForm       xl;
    XoForm       xo;
    XsForm       xs;

    OcForm       opc;
    LkForm       lk;
    uint32_t     bits;

} PowerPCInstruction;

#endif

