#ifndef _SectHeader_h_
#define _SectHeader_h_

#include <Base.h>
#include <defines/SectHeader.d>

class RelocationTable;
class LineInfoTable;
class XCoffFile;

class SectHeader : public Base {
protected:
    char* rawDataPtr;
    char* relocationPtr;
    char* lineInfoPointer;
    uint32_t numOfRelocations;
    uint32_t numOfLineInfo;
    RelocationTable* relocationTable;
    LineInfoTable* lineInfoTable;

    SectHeader* overFlowSection;

    uint16_t index;

protected:
    SectHeader() : Base(XCoffClassTypes_sect_header),
                   rawDataPtr(NULL), relocationPtr(NULL), lineInfoPointer(NULL),
                   numOfRelocations(0),numOfLineInfo(0),
                   relocationTable(NULL),lineInfoTable(NULL),overFlowSection(NULL),index(0) {}
    ~SectHeader() {}
    bool verify();

public:

    SECTHEADER_MACROS_BASIS("For the get_X field macros check the defines directory");

    void print();
    void initFilePointers(BinaryInputFile* b);

    void setOverFlowSection(SectHeader* sh);
    SectHeader* getOverFlowSection() { return overFlowSection; }

    const char* getTypeName();

    char* getRawDataPtr() { return rawDataPtr; }
    uint64_t getRawDataSize() { return GET(s_size); }

    RelocationTable* readRelocTable(BinaryInputFile* binaryInputFile,XCoffFile* xcoff);
    LineInfoTable* readLineInfoTable(BinaryInputFile* binaryInputFile,XCoffFile* xcoff);

    RelocationTable* getRelocationTable() { return relocationTable; }
    LineInfoTable* getLineInfoTable() { return lineInfoTable; }

    uint16_t getIndex() { return index; }

    bool inRange(uint64_t address);

    const char* briefName() { return "SectHeader"; }
};

class SectHeader32 : public SectHeader {
protected:
    SCNHDR entry;

public:

    SECTHEADER_MACROS_CLASS("For the get_X field macros check the defines directory");

    SectHeader32(uint16_t idx) { sizeInBytes = Size__32_bit_Section_Header; index = idx; }
    ~SectHeader32() {}
    uint32_t read(BinaryInputFile* b);
    uint32_t instrument(char* buffer,XCoffFileGen* xCoffGen,BaseGen* gen);
};

class SectHeader64 : public SectHeader {
protected:
    SCNHDR_64 entry;

public:

    SECTHEADER_MACROS_CLASS("For the get_X field macros check the defines directory");

    SectHeader64(uint16_t idx) { sizeInBytes = Size__64_bit_Section_Header; index = idx; }
    ~SectHeader64() {}
    uint32_t read(BinaryInputFile* b);
    uint32_t instrument(char* buffer,XCoffFileGen* xCoffGen,BaseGen* gen);
};

#endif /* _SectHeader_h_ */
