#ifndef _SymbolTable_h_
#define _SymbolTable_h_

#include <Base.h>
#include <defines/SymbolTable.d>

class StringTable;
class DebugSection;
class Symbol;
class SectHeader;
class XCoffFile;

class SymbolBase {
private:
    SymbolBase* next;    
    SymbolBase* prev;
    uint32_t index;
protected:
public:
    SymbolBase(uint32_t idx) : next(NULL),prev(NULL),index(idx) {}
    virtual ~SymbolBase() {}
    SymbolBase* getNext() { return next; }
    SymbolBase* getPrev() { return prev; }
    void setPrev(SymbolBase* p) { prev = p; }
    void setNext(SymbolBase* n) { next = n; }

    virtual bool isAuxilary() { __SHOULD_NOT_ARRIVE; return false; }
    virtual char* charStream() { __SHOULD_NOT_ARRIVE; return NULL; }
    virtual void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux)
                        { __SHOULD_NOT_ARRIVE; }
    virtual char* getName(StringTable* stringTable,DebugSection* debugRawSect)
                        { __SHOULD_NOT_ARRIVE; return NULL; }
    virtual const char* getTypeName() {  __SHOULD_NOT_ARRIVE; return NULL; }

    uint32_t getIndex() { return index; }

};

class Symbol : public SymbolBase {
public:

    Symbol(uint32_t idx) : SymbolBase(idx) {}
    ~Symbol() {}
    bool isAuxilary() { return false; }

    SYMBOL_MACROS_BASIS("For the get_X field macros check the defines directory");

    bool unknownSymbol();
    bool stringInDebugSection();
    const char* getTypeName();
    char* getName(StringTable* stringTable,DebugSection* debugRawSect);

    void print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux);

    virtual uint32_t getNameOffset() { __SHOULD_NOT_ARRIVE; return 0; }
    virtual char* getNameStringTable(StringTable* stringTable) { __SHOULD_NOT_ARRIVE; return NULL; }
    virtual char* getNameDebugSection(DebugSection* rawDebugSect) { __SHOULD_NOT_ARRIVE; return NULL; }

    static bool isSorted(Symbol** symbols,uint32_t symbolCount) { return true; }
    static bool builtinSaveRestore(char* ptr);
    static Symbol* findSymbol(Symbol** symbols,uint32_t symbolCount,uint64_t value);

    virtual void changeValueCopy(uint64_t value,char* buff) { __SHOULD_NOT_ARRIVE; }
};

class Symbol32 : public Symbol {
private:
    SYMENT entry;
protected:
public:
    Symbol32(uint32_t idx) : Symbol(idx){}
    ~Symbol32() {}
    char* charStream() { return (char*)&entry; }

    SYMBOL_MACROS_CLASS("For the get_X field macros check the defines directory");
    GET_FIELD_CLASS(char*,n_name); 
    GET_FIELD_CLASS(uint32_t,n_zeroes);

    uint32_t getNameOffset();
    char* getNameStringTable(StringTable* stringTable);
    char* getNameDebugSection(DebugSection* rawDebugSect);

    void changeValueCopy(uint64_t value,char* buff);
};

class Symbol64 : public Symbol {
private:
    SYMENT_64 entry;
protected:
public:
    Symbol64(uint32_t idx) : Symbol(idx) {}
    ~Symbol64() {}
    char* charStream() { return (char*)&entry; }

    SYMBOL_MACROS_CLASS("For the get_X field macros check the defines directory");

    uint32_t getNameOffset();
    char* getNameStringTable(StringTable* stringTable);
    char* getNameDebugSection(DebugSection* rawDebugSect);

    void changeValueCopy(uint64_t value,char* buff);
};

class SymbolTable : public Base {
protected:
    char* symbolTablePtr;
    uint32_t numberOfSymbols;
    SymbolBase** symbols;
    StringTable* stringTable;
    DebugSection* debugRawSect;

    XCoffFile* xCoffFile;

public:

    SymbolTable(char* ptr,uint32_t nsyms,XCoffFile* xcoff);
    ~SymbolTable();

    char* getSymbolTablePtr() { return symbolTablePtr; }
    uint32_t getNumberOfSymbols() { return numberOfSymbols; }

    void print();
    void printSymbol(uint32_t index);
    void printSymbol(SymbolBase* sym);
    uint32_t read(BinaryInputFile* b);

    void setStringTable(StringTable* st) { stringTable = st; }
    void setDebugSection(DebugSection* ds) { debugRawSect = ds; }

    SymbolBase* getSymbol(uint32_t index) { ASSERT(index < numberOfSymbols); return symbols[index]; }
    char* getSymbolName(SymbolBase* sym) { return sym->getName(stringTable,debugRawSect); }
    char* getSymbolName(uint32_t index) { return getSymbolName(getSymbol(index)); }

    static SymbolBase* newSymbol(SymbolBase* lastSymbol,
                                 uint8_t remaningAux,BinaryInputFile* binaryInputFile,uint32_t index,
                                 XCoffFile* xcoff);

    uint64_t getSymbolLength(SymbolBase* symBase);
    uint8_t getStorageMapping(SymbolBase* symBase);

    uint32_t filterSortAddressSymbols(Symbol** symbolArray,uint32_t arrayLength);
    uint32_t filterSortFuncSymbols(Symbol** symbolArray,uint32_t arrayLength,SectHeader* header);
    uint32_t filterSortBSSSymbols(Symbol** symbolArray,uint32_t arrayLength);

    XCoffFile* getXCoffFile() { return xCoffFile; }

    const char* briefName() { return "SymbolTable"; }

    uint32_t instrument(char* buffer,XCoffFileGen* xCoffGen,BaseGen* gen);

};

extern int searchSymbolValue(const void* arg1,const void* arg2);

#endif /* _SymbolTable_h_ */
