
/*********************************************************************
D    2    _    _    60            tdi                    Trap Doubleword Immediate
D    3    _    _    60            twi                    Trap Word Immediate
D    7    _    _    54            mulli                Multiply Low Immediate
D    8    _    SR    51            subfic                Subtract From Immediate Carrying
D    10    _    _    59            cmpli                Compare Logical Immediate
D    11    _    _    58            cmpi                Compare Immediate
D    12    _    SR    50            addic                Add Immediate Carrying
D    13    _    SR    50            addic.                Add Immediate Carrying and Record
D    14    _    _    49            addi                Add Immediate
D    15    _    _    49            addis                Add Immediate Shifted
B    16    _    CT    23            bc[l][a]            Branch Conditional
Sc    17    _    _    25            sc                    System Call
I    18    _    _    23            b[l][a]                Branch
Xl    19    0    _    28            mcrf                Move Condition Register Field
Xl    19    16    CT    24            bclr[l]                Branch Conditional to Link Register
Xl    19    18    _    III            rfid                Return from Interrupt Doubleword
Xl    19    33    _    27            crnor                Condition Register NOR
Xl    19    129    _    27            crandc                Condition Register AND with Complement
Xl    19    150    _    II            isync                Instruction Synchronize
Xl    19    193    _    26            crxor                Condition Register XOR
Xl    19    225    _    26            crnand                Condition Register NAND
Xl    19    257    _    26            crand                Condition Register AND
Xl    19    289    _    27            creqv                Condition Register Equivalent
Xl    19    417    _    27            crorc                Condition Register OR with Complement
Xl    19    449    _    26            cror                Condition Register OR
Xl    19    528    CT    24            bcctr[l]            Branch Conditional to Count Register
M    20    _    SR    73            rlwimi[.]            Rotate Left Word Immediate then Mask Insert
M    21    _    SR    70            rlwinm[.]            Rotate Left Word Immediate then AND with Mask
M    23    _    SR    72            rlwnm[.]            Rotate Left Word then AND with Mask
D    24    _    _    63            ori                    OR Immediate
D    25    _    _    63            oris                OR Immediate Shifted
D    26    _    _    63            xori                XOR Immediate
D    27    _    _    63            xoris                XOR Immediate Shifted
D    28    _    SR    62            andi.                AND Immediate
D    29    _    SR    62            andis.                AND Immediate Shifted
Md    30    0    SR    69            rldicl[.]            Rotate Left Doubleword Immediate then Clear Left
Md    30    1    SR    69            rldicr[.]            Rotate Left Doubleword Immediate then Clear Right
Md    30    2    SR    70            rldic[.]            Rotate Left Doubleword Immediate then Clear
Md    30    3    SR    73            rldimi[.]            Rotate Left Doubleword Immediate then Mask Insert
Mds    30    8    SR    71            rldcl[.]            Rotate Left Doubleword then Clear Left
Mds    30    9    SR    72            rldcr[.]            Rotate Left Doubleword then Clear Right
X    31    0    _    58            cmp                    Compare
X    31    4    _    61            tw                    Trap Word
Xo    31    8    SR    51            subfc[o][.]            Subtract From Carrying
Xo    31    9    SR    55            mulhdu[.]            Multiply High Doubleword Unsigned
Xo    31    10    SR    51            addc[o][.]            Add Carrying
Xo    31    11    SR    55            mulhwu[.]            Multiply High Word Unsigned
Xfx    31    19    _    80            mfcr                Move From Condition Register
Xfx    31    19    _    118            mfocrf                Move From One Condition Register Field
X    31    20    _    II            lwarx                Load Word And Reserve Indexed
X    31    21    _    37            ldx                    Load Doubleword Indexed
X    31    23    _    35            lwzx                Load Word and Zero Indexed
X    31    24    SR    74            slw[.]                Shift Left Word
X    31    26    SR    67            cntlzw[.]            Count Leading Zeros Word
X    31    27    SR    74            sld[.]                Shift Left Doubleword
X    31    28    SR    64            and[.]                AND
X    31    32    _    59            cmpl                Compare Logical
Xo    31    40    SR    50            subf[o][.]            Subtract From
X    31    53    _    37            ldux                Load Doubleword with Update Indexed
X    31    54    _    II            dcbst                Data Cache Block Store
X    31    55    _    35            lwzux                Load Word and Zero with Update Indexed
X    31    58    SR    67            cntlzd[.]            Count Leading Zeros Doubleword
X    31    60    SR    65            andc[.]                AND with Complement
X    31    68    _    61            td                    Trap Doubleword
Xo    31    73    SR    55            mulhd[.]            Multiply High Doubleword
Xo    31    75    SR    55            mulhw[.]            Multiply High Word
X    31    83    _    III            mfmsr                Move From Machine State Register
X    31    84    _    II            ldarx                Load Doubleword And Reserve Indexed
X    31    86    _    II            dcbf                Data Cache Block Flush
X    31    87    _    32            lbzx                Load Byte and Zero Indexed
Xo    31    104    SR    53            neg[o][.]            Negate
X    31    119    _    32            lbzux                Load Byte and Zero with Update Indexed
X    31    124    SR    65            nor[.]                NOR
Xo    31    136    SR    52            subfe[o][.]            Subtract From Extended
Xo    31    138    SR    52            adde[o][.]            Add Extended
Xfx    31    144    _    80            mtcrf                Move To Condition Register Fields
Xfx    31    144    _    118            mtocrf                Move To One Condition Register Field
X    31    146    _    III            mtmsr                Move To Machine State Register
X    31    149    _    41            stdx                Store Doubleword Indexed
X    31    150    _    II            stwcx.                Store Word Conditional Indexed
X    31    151    _    40            stwx                Store Word Indexed
X    31    178    _    III            mtmsrd                Move To Machine State Register Doubleword
X    31    181    _    41            stdux                Store Doubleword with Update Indexed
X    31    183    _    40            stwux                Store Word with Update Indexed
Xo    31    200    SR    53            subfze[o][.]        Subtract From Zero Extended
Xo    31    202    SR    53            addze[o][.]            Add to Zero Extended
X    31    210    32    III            mtsr                Move To Segment Register
X    31    214    _    II            stdcx.                Store Doubleword Conditional Indexed
X    31    215    _    38            stbx                Store Byte Indexed
Xo    31    232    SR    52            subfme[o][.]        Subtract From Minus One Extended
Xo    31    233    SR    54            mulld[o][.]            Multiply Low Doubleword
Xo    31    234    SR    52            addme[o][.]            Add to Minus One Extended
Xo    31    235    SR    54            mullw[o][.]            Multiply Low Word
X    31    242    32    III            mtsrin                Move To Segment Register Indirect
X    31    246    _    II            dcbtst                Data Cache Block Touch for Store
X    31    247    _    38            stbux                Store Byte with Update Indexed
Xo    31    266    SR    50            add[o][.]            Add
X    31    278    _    II            dcbt                Data Cache Block Touch
X    31    279    _    33            lhzx                Load Halfword and Zero Indexed
X    31    284    SR    65            eqv[.]                Equivalent
X    31    306    64    III            tlbie                TLB Invalidate Entry
X    31    310    _    II            eciwx                External Control In Word Indexed
X    31    311    _    33            lhzux                Load Halfword and Zero with Update Indexed
X    31    316    SR    64            xor[.]                XOR
Xfx    31    339    _    79            mfspr                Move From Special Purpose Register
X    31    341    _    36            lwax                Load Word Algebraic Indexed
X    31    343    _    34            lhax                Load Halfword Algebraic Indexed
X    31    370    _    III            tlbia                TLB Invalidate All
Xfx    31    371    _    II            mftb                Move From Time Base
X    31    373    _    36            lwaux                Load Word Algebraic with Update Indexed
X    31    375    _    34            lhaux                Load Halfword Algebraic with Update Indexed
X    31    402    _    III            slbmte                SLB Move To Entry
X    31    407    _    39            sthx                Store Halfword Indexed
X    31    412    SR    65            orc[.]                OR with Complement
Xs    31    413    SR    76            sradi[.]            Shift Right Algebraic Doubleword Immediate
X    31    434    _    III            slbie                SLB Invalidate Entry
X    31    438    _    II            ecowx                External Control Out Word Indexed
X    31    439    _    39            sthux                Store Halfword with Update Indexed
X    31    444    SR    64            or[.]                OR
Xo    31    457    SR    57            divdu[o][.]            Divide Doubleword Unsigned
Xo    31    459    SR    57            divwu[o][.]            Divide Word Unsigned
Xfx    31    467    _    78            mtspr                Move To Special Purpose Register
X    31    476    SR    64            nand[.]                NAND
Xo    31    489    SR    56            divd[o][.]            Divide Doubleword
Xo    31    491    SR    56            divw[o][.]            Divide Word
X    31    498    _    III            slbia                SLB Invalidate All
X    31    512    _    131            mcrxr                Move to Condition Register from XER
X    31    533    _    46            lswx                Load String Word Indexed
X    31    534    _    42            lwbrx                Load Word Byte-Reverse Indexed
X    31    535    _    98            lfsx                Load Floating-Point Single Indexed
X    31    536    SR    75            srw[.]                Shift Right Word
X    31    539    SR    75            srd[.]                Shift Right Doubleword
X    31    566    _    III            tlbsync                TLB Synchronize
X    31    567    _    98            lfsux                Load Floating-Point Single with Update Indexed
X    31    595    32    III            mfsr                 Move From Segment Register
X    31    597    _    46            lswi                Load String Word Immediate
X    31    598    _    II            sync                Synchronize
X    31    599    _    99            lfdx                Load Floating-Point Double Indexed
X    31    631    _    99            lfdux                Load Floating-Point Double with Update Indexed
X    31    659    32    III            mfsrin                Move From Segment Register Indirect
X    31    661    _    47            stswx                Store String Word Indexed
X    31    662    _    43            stwbrx                Store Word Byte-Reverse Indexed
X    31    663    _    101            stfsx                Store Floating-Point Single Indexed
X    31    695    _    101            stfsux                Store Floating-Point Single with Update Indexed
X    31    725    _    47            stswi                Store String Word Immediate
X    31    727    _    102            stfdx                Store Floating-Point Double Indexed
X    31    759    _    102            stfdux                Store Floating-Point Double with Update Indexed
X    31    790    _    42            lhbrx                Load Halfword Byte-Reverse Indexed
X    31    792    SR    77            sraw[.]                Shift Right Algebraic Word
X    31    794    SR    77            srad[.]                Shift Right Algebraic Doubleword
X    31    824    SR    76            srawi[.]            Shift Right Algebraic Word Immediate
X    31    851    _    III            slbmfev                SLB Move From Entry VSID
X    31    854    _    II            eieio                Enforce In-order Execution of I/O
X    31    915    _    III            slbmfee                SLB Move From Entry ESID
X    31    918    _    43            sthbrx                Store Halfword Byte-Reverse Indexed
X    31    922    SR    66            extsh[.]            Extend Sign Halfword
X    31    954    SR    66            extsb[.]            Extend Sign Byte
X    31    982    _    II            icbi                Instruction Cache Block Invalidate
X    31    983    _    103            stfiwx                Store Floating-Point as Integer Word Indexed
X    31    986    SR    66            extsw[.]            Extend Sign Word
X    31    1014    _    II        dcbz                Data Cache Block set to Zero
D    32    _    _    35            lwz                    Load Word and Zero
D    33    _    _    35            lwzu                Load Word and Zero with Update
D    34    _    _    32            lbz                    Load Byte and Zero
D    35    _    _    32            lbzu                Load Byte and Zero with Update
D    36    _    _    40            stw                    Store Word
D    37    _    _    40            stwu                Store Word with Update
D    38    _    _    38            stb                    Store Byte
D    39    _    _    38            stbu                Store Byte with Update
D    40    _    _    33            lhz                    Load Halfword and Zero
D    41    _    _    33            lhzu                Load Halfword and Zero with Update
D    42    _    _    34            lha                    Load Halfword Algebraic
D    43    _    _    34            lhau                Load Halfword Algebraic with Update
D    44    _    _    39            sth                    Store Halfword
D    45    _    _    39            sthu                Store Halfword with Update
D    46    _    _    44            lmw                    Load Multiple Word
D    47    _    _    44            stmw                Store Multiple Word
D    48    _    _    98            lfs                    Load Floating-Point Single
D    49    _    _    98            lfsu                Load Floating-Point Single with Update
D    50    _    _    99            lfd                    Load Floating-Point Double
D    51    _    _    99            lfdu                Load Floating-Point Double with Update
D    52    _    _    101            stfs                Store Floating-Point Single
D    53    _    _    101            stfsu                Store Floating-Point Single with Update
D    54    _    _    102            stfd                Store Floating-Point Double
D    55    _    _    102            stfdu                Store Floating-Point Double with Update
Ds    58    0    _    37            ld                    Load Doubleword
Ds    58    1    _    37            ldu                    Load Doubleword with Update
Ds    58    2    _    36            lwa                    Load Word Algebraic
A    59    18    _    106            fdivs[.]            Floating Divide Single
A    59    20    _    105            fsubs[.]            Floating Subtract Single
A    59    21    _    105            fadds[.]            Floating Add Single
A    59    22    _    120            fsqrts[.]            Floating Square Root Single
A    59    24    _    120            fres[.]                Floating Reciprocal Estimate Single
A    59    25    _    106            fmuls[.]            Floating Multiply Single
A    59    28    _    107            fmsubs[.]            Floating Multiply-Subtract Single
A    59    29    _    107            fmadds[.]            Floating Multiply-Add Single
A    59    30    _    108            fnmsubs[.]            Floating Negative Multiply-Subtract Single
A    59    31    _    108            fnmadds[.]            Floating Negative Multiply-Add Single
Ds    62    0    _    41            std                    Store Doubleword
Ds    62    1    _    41            stdu                Store Doubleword with Update
X    63    0    _    113            fcmpu                Floating Compare Unordered
X    63    12    _    109            frsp[.]                Floating Round to Single-Precision
X    63    14    _    111            fctiw[.]            Floating Convert To Integer Word
X    63    15    _    111            fctiwz[.]            Floating Convert To Integer Word with round toward Zero
A    63    18    _    106            fdiv[.]                Floating Divide
A    63    20    _    105            fsub[.]                Floating Subtract
A    63    21    _    105            fadd[.]                Floating Add
A    63    22    _    120            fsqrt[.]            Floating Square Root
A    63    23    _    121            fsel[.]                Floating Select
A    63    25    _    106            fmul[.]                Floating Multiply
A    63    26    _    121            frsqrte[.]            Floating Reciprocal Square Root Estimate
A    63    28    _    107            fmsub[.]            Floating Multiply-Subtract
A    63    29    _    107            fmadd[.]            Floating Multiply-Add
A    63    30    _    108            fnmsub[.]            Floating Negative Multiply-Subtract
A    63    31    _    108            fnmadd[.]            Floating Negative Multiply-Add
X    63    32    _    113            fcmpo                Floating Compare Ordered
X    63    38    _    116            mtfsb1[.]            Move To FPSCR Bit 1
X    63    40    _    104            fneg[.]                Floating Negate
X    63    64    _    114            mcrfs                Move to Condition Register from FPSCR
X    63    70    _    116            mtfsb0[.]            Move To FPSCR Bit 0
X    63    72    _    104            fmr[.]                Floating Move Register
X    63    134    _    115            mtfsfi[.]            Move To FPSCR Field Immediate
X    63    136    _    104            fnabs[.]            Floating Negative Absolute Value
X    63    264    _    104            fabs[.]                Floating Absolute Value
X    63    583    _    114            mffs[.]                Move From FPSCR
Xfl    63    711    _    115            mtfsf[.]            Move To FPSCR Fields
X    63    814    _    110            fctid[.]            Floating Convert To Integer Doubleword
X    63    815    _    110            fctidz[.]            Floating Convert To Integer Doubleword with round toward
X    63    846    _    112            fcfid[.]            Floating Convert From Integer Doubleword
*********************************************************************/


#define    opcode_Tdi       2
#define    opcode_Twi       3
#define    opcode_Mulli     7
#define    opcode_Subfic    8
#define    opcode_Cmpli     10
#define    opcode_Cmpi      11
#define    opcode_Addic     12
#define    opcode_Addicc    13
#define    opcode_Addi      14
#define    opcode_Addis     15
#define    opcode_BcLA      16
#define    opcode_Sc        17
#define    opcode_BLA       18
#define    opcode_Mcrf      19
#define    opcode_BclrL     19
#define    opcode_Rfid      19
#define    opcode_Crnor     19
#define    opcode_Crandc    19
#define    opcode_Isync     19
#define    opcode_Crxor     19
#define    opcode_Crnand    19
#define    opcode_Crand     19
#define    opcode_Creqv     19
#define    opcode_Crorc     19
#define    opcode_Cror      19
#define    opcode_BcctrL    19
#define    opcode_RlwimiC   20
#define    opcode_RlwinmC   21
#define    opcode_RlwnmC    23
#define    opcode_Ori       24
#define    opcode_Oris      25
#define    opcode_Xori      26
#define    opcode_Xoris     27
#define    opcode_Andic     28
#define    opcode_Andisc    29
#define    opcode_RldiclC   30
#define    opcode_RldicrC   30
#define    opcode_RldicC    30
#define    opcode_RldimiC   30
#define    opcode_RldclC    30
#define    opcode_RldcrC    30
#define    opcode_Cmp       31
#define    opcode_Tw        31
#define    opcode_SubfcOC   31
#define    opcode_MulhduC   31
#define    opcode_AddcOC    31
#define    opcode_MulhwuC   31
#define    opcode_Mfcr      31
#define    opcode_Mfocrf    31
#define    opcode_Lwarx     31
#define    opcode_Ldx       31
#define    opcode_Lwzx      31
#define    opcode_SlwC      31
#define    opcode_CntlzwC   31
#define    opcode_SldC      31
#define    opcode_AndC      31
#define    opcode_Cmpl      31
#define    opcode_SubfOC    31
#define    opcode_Ldux      31
#define    opcode_Dcbst     31
#define    opcode_Lwzux     31
#define    opcode_CntlzdC   31
#define    opcode_AndcC     31
#define    opcode_Td        31
#define    opcode_MulhdC    31
#define    opcode_MulhwC    31
#define    opcode_Mfmsr     31
#define    opcode_Ldarx     31
#define    opcode_Dcbf      31
#define    opcode_Lbzx      31
#define    opcode_NegOC     31
#define    opcode_Lbzux     31
#define    opcode_NorC      31
#define    opcode_SubfeOC   31
#define    opcode_AddeOC    31
#define    opcode_Mtcrf     31
#define    opcode_Mtocrf    31
#define    opcode_Mtmsr     31
#define    opcode_Stdx      31
#define    opcode_Stwcxc    31
#define    opcode_Stwx      31
#define    opcode_Mtmsrd    31
#define    opcode_Stdux     31
#define    opcode_Stwux     31
#define    opcode_SubfzeOC  31
#define    opcode_AddzeOC   31
#define    opcode_Mtsr      31
#define    opcode_Stdcxc    31
#define    opcode_Stbx      31
#define    opcode_SubfmeOC  31
#define    opcode_MulldOC   31
#define    opcode_AddmeOC   31
#define    opcode_MullwOC   31
#define    opcode_Mtsrin    31
#define    opcode_Dcbtst    31
#define    opcode_Stbux     31
#define    opcode_AddOC     31
#define    opcode_Dcbt      31
#define    opcode_Lhzx      31
#define    opcode_EqvC      31
#define    opcode_Tlbie     31
#define    opcode_Eciwx     31
#define    opcode_Lhzux     31
#define    opcode_XorC      31
#define    opcode_Mfspr     31
#define    opcode_Lwax      31
#define    opcode_Lhax      31
#define    opcode_Tlbia     31
#define    opcode_Mftb      31
#define    opcode_Lwaux     31
#define    opcode_Lhaux     31
#define    opcode_Slbmte    31
#define    opcode_Sthx      31
#define    opcode_OrcC      31
#define    opcode_SradiC    31
#define    opcode_Slbie     31
#define    opcode_Ecowx     31
#define    opcode_Sthux     31
#define    opcode_OrC       31
#define    opcode_DivduOC   31
#define    opcode_DivwuOC   31
#define    opcode_Mtspr     31
#define    opcode_NandC     31
#define    opcode_DivdOC    31
#define    opcode_DivwOC    31
#define    opcode_Slbia     31
#define    opcode_Mcrxr     31
#define    opcode_Lswx      31
#define    opcode_Lwbrx     31
#define    opcode_Lfsx      31
#define    opcode_SrwC      31
#define    opcode_SrdC      31
#define    opcode_Tlbsync   31
#define    opcode_Lfsux     31
#define    opcode_Mfsr      31
#define    opcode_Lswi      31
#define    opcode_Sync      31
#define    opcode_Lfdx      31
#define    opcode_Lfdux     31
#define    opcode_Mfsrin    31
#define    opcode_Stswx     31
#define    opcode_Stwbrx    31
#define    opcode_Stfsx     31
#define    opcode_Stfsux    31
#define    opcode_Stswi     31
#define    opcode_Stfdx     31
#define    opcode_Stfdux    31
#define    opcode_Lhbrx     31
#define    opcode_SrawC     31
#define    opcode_SradC     31
#define    opcode_SrawiC    31
#define    opcode_Slbmfev   31
#define    opcode_Eieio     31
#define    opcode_Slbmfee   31
#define    opcode_Sthbrx    31
#define    opcode_ExtshC    31
#define    opcode_ExtsbC    31
#define    opcode_Icbi      31
#define    opcode_Stfiwx    31
#define    opcode_ExtswC    31
#define    opcode_Dcbz      31
#define    opcode_Lwz       32
#define    opcode_Lwzu      33
#define    opcode_Lbz       34
#define    opcode_Lbzu      35
#define    opcode_Stw       36
#define    opcode_Stwu      37
#define    opcode_Stb       38
#define    opcode_Stbu      39
#define    opcode_Lhz       40
#define    opcode_Lhzu      41
#define    opcode_Lha       42
#define    opcode_Lhau      43
#define    opcode_Sth       44
#define    opcode_Sthu      45
#define    opcode_Lmw       46
#define    opcode_Stmw      47
#define    opcode_Lfs       48
#define    opcode_Lfsu      49
#define    opcode_Lfd       50
#define    opcode_Lfdu      51
#define    opcode_Stfs      52
#define    opcode_Stfsu     53
#define    opcode_Stfd      54
#define    opcode_Stfdu     55
#define    opcode_Ld        58
#define    opcode_Ldu       58
#define    opcode_Lwa       58
#define    opcode_FdivsC    59
#define    opcode_FsubsC    59
#define    opcode_FaddsC    59
#define    opcode_FsqrtsC   59
#define    opcode_FresC     59
#define    opcode_FmulsC    59
#define    opcode_FmsubsC   59
#define    opcode_FmaddsC   59
#define    opcode_FnmsubsC  59
#define    opcode_FnmaddsC  59
#define    opcode_Std       62
#define    opcode_Stdu      62
#define    opcode_Fcmpu     63
#define    opcode_FrspC     63
#define    opcode_FctiwC    63
#define    opcode_FctiwzC   63
#define    opcode_FdivC     63
#define    opcode_FsubC     63
#define    opcode_FaddC     63
#define    opcode_FsqrtC    63
#define    opcode_FselC     63
#define    opcode_FmulC     63
#define    opcode_FrsqrteC  63
#define    opcode_FmsubC    63
#define    opcode_FmaddC    63
#define    opcode_FnmsubC   63
#define    opcode_FnmaddC   63
#define    opcode_Fcmpo     63
#define    opcode_Mtfsb1    63
#define    opcode_FnegC     63
#define    opcode_Mcrfs     63
#define    opcode_Mtfsb0    63
#define    opcode_FmrC      63
#define    opcode_MtfsfiC   63
#define    opcode_FnabsC    63
#define    opcode_FabsC     63
#define    opcode_MffsC     63
#define    opcode_MtfsfC    63
#define    opcode_FctidC    63
#define    opcode_FctidzC   63
#define    opcode_FcfidC    63


#define    xcode_Mcrf      0
#define    xcode_BclrL     16
#define    xcode_Rfid      18
#define    xcode_Crnor     33
#define    xcode_Crandc    129
#define    xcode_Isync     150
#define    xcode_Crxor     193
#define    xcode_Crnand    225
#define    xcode_Crand     257
#define    xcode_Creqv     289
#define    xcode_Crorc     417
#define    xcode_Cror      449
#define    xcode_BcctrL    528
#define    xcode_Ld        0
#define    xcode_Ldu       1
#define    xcode_Lwa       2
#define    xcode_FdivsC    18
#define    xcode_FsubsC    20
#define    xcode_FaddsC    21
#define    xcode_FsqrtsC   22
#define    xcode_FresC     24
#define    xcode_FmulsC    25
#define    xcode_FmsubsC   28
#define    xcode_FmaddsC   29
#define    xcode_FnmsubsC  30
#define    xcode_FnmaddsC  31
#define    xcode_Std       0
#define    xcode_Stdu      1
#define    xcode_Fcmpu     0
#define    xcode_FrspC     12
#define    xcode_FctiwC    14
#define    xcode_FctiwzC   15
#define    xcode_FdivC     18
#define    xcode_FsubC     20
#define    xcode_FaddC     21
#define    xcode_FsqrtC    22
#define    xcode_FselC     23
#define    xcode_FmulC     25
#define    xcode_FrsqrteC  26
#define    xcode_FmsubC    28
#define    xcode_FmaddC    29
#define    xcode_FnmsubC   30
#define    xcode_FnmaddC   31
#define    xcode_Fcmpo     32
#define    xcode_Mtfsb1    38
#define    xcode_FnegC     40
#define    xcode_Mcrfs     64
#define    xcode_Mtfsb0    70
#define    xcode_FmrC      72
#define    xcode_MtfsfiC   134
#define    xcode_FnabsC    136
#define    xcode_FabsC     264
#define    xcode_MffsC     583
#define    xcode_MtfsfC    711
#define    xcode_FctidC    814
#define    xcode_FctidzC   815
#define    xcode_FcfidC    846
#define    xcode_RldiclC   0
#define    xcode_RldicrC   1
#define    xcode_RldicC    2
#define    xcode_RldimiC   3
#define    xcode_RldclC    8
#define    xcode_RldcrC    9
#define    xcode_Cmp       0
#define    xcode_Tw        4
#define    xcode_SubfcOC   8
#define    xcode_MulhduC   9
#define    xcode_AddcOC    10
#define    xcode_MulhwuC   11
#define    xcode_Mfcr      19
#define    xcode_Mfocrf    19
#define    xcode_Lwarx     20
#define    xcode_Ldx       21
#define    xcode_Lwzx      23
#define    xcode_SlwC      24
#define    xcode_CntlzwC   26
#define    xcode_SldC      27
#define    xcode_AndC      28
#define    xcode_Cmpl      32
#define    xcode_SubfOC    40
#define    xcode_Ldux      53
#define    xcode_Dcbst     54
#define    xcode_Lwzux     55
#define    xcode_CntlzdC   58
#define    xcode_AndcC     60
#define    xcode_Td        68
#define    xcode_MulhdC    73
#define    xcode_MulhwC    75
#define    xcode_Mfmsr     83
#define    xcode_Ldarx     84
#define    xcode_Dcbf      86
#define    xcode_Lbzx      87
#define    xcode_NegOC     104
#define    xcode_Lbzux     119
#define    xcode_NorC      124
#define    xcode_SubfeOC   136
#define    xcode_AddeOC    138
#define    xcode_Mtcrf     144
#define    xcode_Mtocrf    144
#define    xcode_Mtmsr     146
#define    xcode_Stdx      149
#define    xcode_Stwcxc    150
#define    xcode_Stwx      151
#define    xcode_Mtmsrd    178
#define    xcode_Stdux     181
#define    xcode_Stwux     183
#define    xcode_SubfzeOC  200
#define    xcode_AddzeOC   202
#define    xcode_Mtsr      210
#define    xcode_Stdcxc    214
#define    xcode_Stbx      215
#define    xcode_SubfmeOC  232
#define    xcode_MulldOC   233
#define    xcode_AddmeOC   234
#define    xcode_MullwOC   235
#define    xcode_Mtsrin    242
#define    xcode_Dcbtst    246
#define    xcode_Stbux     247
#define    xcode_AddOC     266
#define    xcode_LscbxC    277
#define    xcode_Dcbt      278
#define    xcode_Lhzx      279
#define    xcode_EqvC      284
#define    xcode_Tlbie     306
#define    xcode_Eciwx     310
#define    xcode_Lhzux     311
#define    xcode_XorC      316
#define    xcode_Mfspr     339
#define    xcode_Lwax      341
#define    xcode_Lhax      343
#define    xcode_Tlbia     370
#define    xcode_Mftb      371
#define    xcode_Lwaux     373
#define    xcode_Lhaux     375
#define    xcode_Slbmte    402
#define    xcode_Sthx      407
#define    xcode_OrcC      412
#define    xcode_SradiC    413
#define    xcode_Slbie     434
#define    xcode_Ecowx     438
#define    xcode_Sthux     439
#define    xcode_OrC       444
#define    xcode_DivduOC   457
#define    xcode_DivwuOC   459
#define    xcode_Mtspr     467
#define    xcode_NandC     476
#define    xcode_DivdOC    489
#define    xcode_DivwOC    491
#define    xcode_Slbia     498
#define    xcode_Mcrxr     512
#define    xcode_Lswx      533
#define    xcode_Lwbrx     534
#define    xcode_Lfsx      535
#define    xcode_SrwC      536
#define    xcode_SrdC      539
#define    xcode_Tlbsync   566
#define    xcode_Lfsux     567
#define    xcode_Mfsr      595
#define    xcode_Lswi      597
#define    xcode_Sync      598
#define    xcode_Lfdx      599
#define    xcode_Lfdux     631
#define    xcode_Mfsrin    659
#define    xcode_Stswx     661
#define    xcode_Stwbrx    662
#define    xcode_Stfsx     663
#define    xcode_Stfsux    695
#define    xcode_Stswi     725
#define    xcode_Stfdx     727
#define    xcode_Stfdux    759
#define    xcode_Lhbrx     790
#define    xcode_SrawC     792
#define    xcode_SradC     794
#define    xcode_SrawiC    824
#define    xcode_Slbmfev   851
#define    xcode_Eieio     854
#define    xcode_Slbmfee   915
#define    xcode_Sthbrx    918
#define    xcode_ExtshC    922
#define    xcode_ExtsbC    954
#define    xcode_Icbi      982
#define    xcode_Stfiwx    983
#define    xcode_ExtswC    986
#define    xcode_Dcbz      1014


#define BRANCH_ALWAYS(__ho)  (((__ho) & 0x14) == 0x14)
#define IS_B(__i)            ((__i).opc.opc == opcode_BLA)
#define IS_BC(__i)           ((__i).opc.opc == opcode_BcLA)
#define IS_BCLR(__i)         (((__i).opc.opc == opcode_BclrL) && ((__i).xl.xo == xcode_BclrL))
#define IS_BCCTR(__i)        (((__i).opc.opc == opcode_BcctrL) && ((__i).xl.xo == xcode_BcctrL))
#define IS_CMPLI(__i)        ((__i).opc.opc == opcode_Cmpli)
#define IS_LWZ(__i)          ((__i).opc.opc == opcode_Lwz)
#define IS_LD(__i)           (((__i).opc.opc == opcode_Ld) && ((__i).ds.xo == xcode_Ld))
#define IS_ADD(__i)          (((__i).opc.opc == opcode_AddOC) && ((__i).xo.xo == xcode_AddOC))
#define IS_LDX(__i)          (((__i).opc.opc == opcode_Ldx) && ((__i).x.xo == xcode_Ldx))
#define IS_LWZX(__i)         (((__i).opc.opc == opcode_Lwzx) && ((__i).x.xo == xcode_Lwzx))
#define IS_ADDI(__i)         ((__i).opc.opc == opcode_Addi)
#define IS_LSWI(__i)         (((__i).opc.opc == opcode_Lswi) && ((__i).x.xo == xcode_Lswi))
#define IS_STSWI(__i)        (((__i).opc.opc == opcode_Stswi) && ((__i).x.xo == xcode_Stswi))

#define GET_BH(__i)          (((__i).xl.bb) & 0x3)
#define REG_TOC              2
#define REG_SP               1
#define INVALID_REGISTER     0xffffffff
#define INVALID_XOP          0x400

#define MEMORY_MASK          0x1c
#define MEMORY_MATCH         0x14
#define FOP_AFORM_MASK       0x10
#define FOP_AFORM_MATCH      0x10
#define FOP_CONDREG_MASK     0x1e
#define FOP_CONDREG_MATCH    0x06

#define NUM_OF_GPR_REGS      0x20
#define NUM_OF_FPR_REGS      0x20
#define LAST_VOLATILE_REG    0xd

#define REG_XER              0x020
#define REG_LNK              0x100
#define REG_CTR              0x120

#define SAFE_FRAME_LOC       64

#define COMPARE_LT           0
#define COMPARE_GT           1
#define COMPARE_EQ           2
#define COMPARE_TRUE         1
#define COMPARE_FALSE        0
