/*
##################################################################
XCOFF32     XCOFF64     Name     Description
Offset     Length     Offset     Length
0     8     0      4   n_name     Symbol name (occupies the same 8 bytes as n_zeroes and n_offset)
0     4     0   4      n_zeroes     Zero, indicating name in string table or .debug section (overlays first 4 bytes of n_name)
4     4     8     4     n_offset+     Offset of the name in string table or .debug section (In XCOFF32: overlays last 4 bytes of n_name)
8     4     0     8     n_value+     Symbol value; storage class-dependent
12     2     12     2     n_scnum     Section number of symbol
14     2     14     2     n_type     Basic and derived type specification
14     1     14     1     n_lang     Source language ID (overlays first byte of n_type)
15     1     15     1     n_cpu     CPU Type ID (overlays second byte of n_type)
16     1     16     1     n_sclass     Storage class of symbol
17     1     17     1     n_numaux     Number of auxiliary entries
+Use "32" or "64" suffix when __XCOFF_HYBRID__ is defined.
##################################################################
*/

#define SYMB_IS_OF_TYPE_DECL(__str) inline bool is ## __str() { return (GET(n_sclass) == C_## __str); }
#define IS_SYMB_TYPE(__str) is ## __str()

#define SYMBOL_MACROS_BASIS(__str) /** __str **/ \
        GET_FIELD_BASIS(char*,n_name); \
        GET_FIELD_BASIS(uint32_t,n_zeroes); \
        GET_FIELD_BASIS(uint32_t,n_offset); \
        GET_FIELD_BASIS(uint64_t,n_value); \
        GET_FIELD_BASIS(int16_t,n_scnum); \
        GET_FIELD_BASIS(uint16_t,n_type); \
        GET_FIELD_BASIS(uint8_t,n_sclass); \
        GET_FIELD_BASIS(uint8_t,n_numaux); \
        \
        SYMB_IS_OF_TYPE_DECL(ALIAS); \
        SYMB_IS_OF_TYPE_DECL(ARG); \
        SYMB_IS_OF_TYPE_DECL(AUTO); \
        SYMB_IS_OF_TYPE_DECL(EFCN); \
        SYMB_IS_OF_TYPE_DECL(ENTAG); \
        SYMB_IS_OF_TYPE_DECL(EOS); \
        SYMB_IS_OF_TYPE_DECL(EXTDEF); \
        SYMB_IS_OF_TYPE_DECL(FIELD); \
        SYMB_IS_OF_TYPE_DECL(HIDDEN); \
        SYMB_IS_OF_TYPE_DECL(LABEL); \
        SYMB_IS_OF_TYPE_DECL(LINE); \
        SYMB_IS_OF_TYPE_DECL(MOE); \
        SYMB_IS_OF_TYPE_DECL(MOS); \
        SYMB_IS_OF_TYPE_DECL(MOU); \
        SYMB_IS_OF_TYPE_DECL(REG); \
        SYMB_IS_OF_TYPE_DECL(REGPARM); \
        SYMB_IS_OF_TYPE_DECL(STRTAG); \
        SYMB_IS_OF_TYPE_DECL(TPDEF); \
        SYMB_IS_OF_TYPE_DECL(ULABEL); \
        SYMB_IS_OF_TYPE_DECL(UNTAG); \
        SYMB_IS_OF_TYPE_DECL(USTATIC); \
        \
        SYMB_IS_OF_TYPE_DECL(BCOMM); \
        SYMB_IS_OF_TYPE_DECL(BINCL); \
        SYMB_IS_OF_TYPE_DECL(BLOCK); \
        SYMB_IS_OF_TYPE_DECL(BSTAT); \
        SYMB_IS_OF_TYPE_DECL(DECL); \
        SYMB_IS_OF_TYPE_DECL(ECOML); \
        SYMB_IS_OF_TYPE_DECL(ECOMM); \
        SYMB_IS_OF_TYPE_DECL(EINCL); \
        SYMB_IS_OF_TYPE_DECL(ENTRY); \
        SYMB_IS_OF_TYPE_DECL(ESTAT); \
        SYMB_IS_OF_TYPE_DECL(EXT); \
        SYMB_IS_OF_TYPE_DECL(FCN); \
        SYMB_IS_OF_TYPE_DECL(FILE); \
        SYMB_IS_OF_TYPE_DECL(FUN); \
        SYMB_IS_OF_TYPE_DECL(GSYM); \
        SYMB_IS_OF_TYPE_DECL(HIDEXT); \
        SYMB_IS_OF_TYPE_DECL(INFO); \
        SYMB_IS_OF_TYPE_DECL(LSYM); \
        SYMB_IS_OF_TYPE_DECL(NULL); \
        SYMB_IS_OF_TYPE_DECL(PSYM); \
        SYMB_IS_OF_TYPE_DECL(RPSYM); \
        SYMB_IS_OF_TYPE_DECL(RSYM); \
        SYMB_IS_OF_TYPE_DECL(STAT); \
        SYMB_IS_OF_TYPE_DECL(STSYM); \
        SYMB_IS_OF_TYPE_DECL(TCSYM); \
        SYMB_IS_OF_TYPE_DECL(WEAKEXT); \
/** END of definitions **/

#define SYMBOL_MACROS_CLASS(__str) /** __str **/ \
        GET_FIELD_CLASS(uint32_t,n_offset); \
        GET_FIELD_CLASS(uint64_t,n_value); \
        GET_FIELD_CLASS(int16_t,n_scnum); \
        GET_FIELD_CLASS(uint16_t,n_type); \
        GET_FIELD_CLASS(uint8_t,n_sclass); \
        GET_FIELD_CLASS(uint8_t,n_numaux); \
/** END of definitions **/
