#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>

#include <CacheSimulationCommon.h>

#ifdef USE_SAFE_STD_FUNCTIONS

#define __MAX_BLOCK_AREA  (0x1000 * (sizeof(BasicBlockInfo)/sizeof(Address_t) + 1))
#define __MAX_MALLOC_SIZE __MAX_CACHE_AREA + __MAX_BLOCK_AREA
Address_t allocation_area[__MAX_MALLOC_SIZE];
Address_t* alloc_ptr = allocation_area;
void* local_malloc(uint32_t n){
    Address_t* end_ptr = alloc_ptr + __MAX_MALLOC_SIZE;
    Address_t* ret = alloc_ptr;
    uint32_t wcount = (n + (n % sizeof(Address_t) ? sizeof(Address_t) : 0))/sizeof(Address_t);
    alloc_ptr += wcount;
    if(alloc_ptr >= end_ptr){
    printf("FATAL Error FATAL Error in Cache Simulation\n");
    exit(-1);
    }
    return ret;
}

int getRandomNumber();
uint32_t local_random() {
    uint32_t dummy = 0;
    uint32_t first_base = (uint32_t)&dummy;
    first_base >>= 4;
    first_base &= 0xffff;
    uint32_t second_base = getRandomNumber();
    return (first_base + second_base);
}

#define bzero(__ptr,__n) memset((__ptr),0,(__n))
#define malloc local_malloc
#define random local_random

#endif

#define STATUS_IDX(__sys,__lvl,__sta) ((__sys)*(__MAX_LEVEL*Total_AccessStatus) + (__lvl) * Total_AccessStatus + (__sta))

Counter_t       lastSaturationAccess = 0;
Counter_t       saturatedBlockCount = 0;
Counter_t       totalNumberOfAccesses = 0;
Counter_t       totalNumberOfSamples = 0;

uint32_t        blockCount = 0;
BasicBlockInfo* blocks = NULL;

static unsigned int SEED = 93186752;
int getRandomNumber()
{
    double x;
    static unsigned a = 1588635695, m = 4294967291U, q = 2, r = 1117695901;
    SEED = a*(SEED % q) - r*(SEED / q);
    x = (double)SEED / (double)m; /* return the next random number x: 0 <= x < 1 */

    return ( (int)128*x +1);
}

FILE* openOutputFile(const char* comment){

    uint32_t rankId = 0;

    char strBuffer[__MAX_STRING_SIZE];

    uint32_t processId = getpid();
    uint32_t hostId = gethostid();

    FILE* fp = fopen("RankPid","r");
    if(fp){
        uint32_t entryFound = 0;
        while(fgets(strBuffer,1024,fp)){
            strBuffer[1023] = '\0';
            uint32_t corresPid = 0,corresHid = 0;
            sscanf(strBuffer,"%u %u %u",&rankId,&corresPid,&corresHid);
            if((corresPid == processId) && (corresHid == hostId)){
                entryFound = 1;
                break;
            }
        }
        fclose(fp);
        if(!entryFound){
            rankId = processId;
        }
    } else {
        rankId = processId;
    }

    char appName[__MAX_STRING_SIZE];
    uint32_t phaseId = 0;
    char extension[__MAX_STRING_SIZE];

    sscanf(comment,"%s %u %s",appName,&phaseId,extension);

    if(phaseId > 0){
        sprintf(strBuffer,"%s.phase.%d.meta_%04d.%s",appName,phaseId,rankId,extension);
    } else {
        sprintf(strBuffer,"%s.meta_%04d.%s",appName,rankId,extension);
    }
    
    fp = fopen(strBuffer,"w");
    if(!fp){
        fprintf(stderr,"Error : Can not open %s to write results\n",strBuffer);
        return NULL;
    }

    fprintf(fp,"# appname   = %s\n",appName);
    fprintf(fp,"# extension = %s\n",extension);
    fprintf(fp,"# phase     = %d\n",phaseId);
    fprintf(fp,"# rank      = %d\n",rankId);

    return fp;
}

#define __HASH_FUNC(__x,__y,__z) (((__x) / __y) % (__z))

HashEntry* initHash(Attribute_t setCount,Attribute_t assocCount){
    uint32_t setIdx;
    uint32_t linIdx;
    HashEntry* hashLines = NULL;

    if(assocCount >= __MAX_LINEAR_SEARCH_ASSOC){
        Attribute_t entryCount = setCount * assocCount * 2;
        hashLines = (HashEntry*) malloc (sizeof(HashEntry)*entryCount);
        bzero(hashLines,sizeof(HashEntry)*entryCount);

        for(setIdx=0;setIdx<setCount;setIdx++){
            HashEntry* nextFreePtr = hashLines + (setIdx * assocCount * 2) + assocCount;
            for(linIdx=0;linIdx<assocCount;linIdx++){
                nextFreePtr[linIdx].next = (linIdx+1) % assocCount;
            }
        }
    }

    return hashLines;
}

#ifdef DEBUG_RUN_2
uint64_t howManyHashSearches = 0;
uint64_t howManyHashTraverses = 0;
#endif

uint16_t findInHash(Address_t lineIdx,Attribute_t setCount,Attribute_t assocCount,
                    HashEntry* cacheHash,uint32_t setIdx)
{
#ifdef DEBUG_RUN_2
    howManyHashSearches++;
#endif
    uint16_t ret = assocCount;
    if(lineIdx && cacheHash){
        HashEntry* hashLines = cacheHash + (setIdx * assocCount * 2);
        HashEntry* nextFreePtr = hashLines + assocCount;
        uint16_t hashCode = __HASH_FUNC(lineIdx,setCount,assocCount);
        HashEntry* curr = hashLines + hashCode;
        while(curr){
#ifdef DEBUG_RUN_2
            howManyHashTraverses++;
#endif
            if(curr->key == lineIdx){
                ret = curr->value;
                break;
            }
            curr = ( curr->next ? (nextFreePtr + curr->next) : NULL );
        }
    }
    return ret;
}

void deleteFromHash(Address_t lineIdx,Attribute_t setCount,Attribute_t assocCount,
                    HashEntry* cacheHash,uint32_t setIdx)
{
    if(lineIdx && cacheHash){
        HashEntry* hashLines = cacheHash + (setIdx * assocCount * 2);
        HashEntry* nextFreePtr = hashLines + assocCount;
        uint16_t hashCode = __HASH_FUNC(lineIdx,setCount,assocCount);
        HashEntry* curr = hashLines + hashCode;
        if(curr->key == lineIdx){
            curr->key = 0;
            curr->value = assocCount;
        } else {
            HashEntry* prev = curr;
            while(prev){
                uint16_t currIdx = prev->next;
                curr = ( currIdx ? (nextFreePtr + currIdx) : NULL );
                if(curr && (curr->key == lineIdx)){
                    curr->key = 0;
                    curr->value = assocCount;
                    prev->next = curr->next;
                    curr->next = nextFreePtr->next;
                    nextFreePtr->next = currIdx;
                    break;
                }
                prev = curr;
            }
        }
    }
}

void insertInHash(Address_t lineIdx,Attribute_t setCount,Attribute_t assocCount,
                  HashEntry* cacheHash,uint32_t setIdx,uint16_t assocIdx)
{
    if(lineIdx && cacheHash){
        HashEntry* hashLines = cacheHash + (setIdx * assocCount * 2);
        HashEntry* nextFreePtr = hashLines + assocCount;
        uint16_t hashCode = __HASH_FUNC(lineIdx,setCount,assocCount);
        HashEntry* curr = hashLines + hashCode;
        if(!curr->key){
            curr->key = lineIdx;
            curr->value = assocIdx;
        } else {
            HashEntry* prev = curr;

            uint16_t currIdx = nextFreePtr->next; 
            curr = (currIdx ? (nextFreePtr + currIdx) : NULL);

#ifdef DEBUG_RUN_2
            if(!curr){
                printf("FATAL Error FATAL Error in Cache Simulation\n");
                exit(-1);
            }
#endif
            curr->key = lineIdx;
            curr->value = assocIdx;

            nextFreePtr->next = curr->next;
            curr->next = prev->next;
            prev->next = currIdx;
        }
    }
}

void initCaches(){

    register uint32_t systemIdx;
    register uint32_t level;
    register uint32_t setIdx;

    for(systemIdx=0;systemIdx<systemCount;systemIdx++){
        register MemoryHierarchy* memoryHierarchy = (systems + systemIdx);
        register uint32_t totalLineCount = 0;
        for(level=0;level<memoryHierarchy->levelCount;level++){

            register Attribute_t setCount = memoryHierarchy->levels[level].attributes[number_of_sets];
            register Attribute_t linePerSet = memoryHierarchy->levels[level].attributes[set_associativity];
            register Attribute_t sizeInLines = setCount*linePerSet;
            totalLineCount += sizeInLines;
        }

        Address_t* allLines = (Address_t*)malloc(sizeof(Address_t)*totalLineCount);
        bzero(allLines,sizeof(Address_t)*totalLineCount);

        for(level=0;level<memoryHierarchy->levelCount;level++){

            register Attribute_t setCount = memoryHierarchy->levels[level].attributes[number_of_sets];
            register Attribute_t linePerSet = memoryHierarchy->levels[level].attributes[set_associativity];
            register Attribute_t sizeInLines = setCount*linePerSet;

            memoryHierarchy->levels[level].hitMissCounters[cache_hit] = 0;
            memoryHierarchy->levels[level].hitMissCounters[cache_miss] = 0;
            memoryHierarchy->levels[level].content = allLines;
#ifdef PER_SET_RECENT
            memoryHierarchy->levels[level].mostRecent = (uint16_t*)malloc(sizeof(uint16_t)*setCount);
            for(setIdx=0;setIdx<setCount;setIdx++){
                memoryHierarchy->levels[level].mostRecent[setIdx] = (linePerSet-1);
            }
#else
            memoryHierarchy->levels[level].mostRecent = (linePerSet-1);
#endif
            allLines += sizeInLines;

            memoryHierarchy->levels[level].highAssocHash = NULL;
            
            if(linePerSet >= __MAX_LINEAR_SEARCH_ASSOC){
                memoryHierarchy->levels[level].highAssocHash = initHash(setCount,linePerSet);
            }
        }
    }
}

#ifdef STRIDE_VS_RANDOM
void initLastAccesses(uint16_t* memopInfo) {

    register uint32_t i;

    register uint32_t total = 0;
    for(i=0;i<blockCount;i++){
        blocks[i].memopCount = memopInfo[i];
        total += blocks[i].memopCount;
    }

    register Attribute_t* allLines = (Attribute_t*)malloc(total*sizeof(Attribute_t));
    bzero(allLines,total*sizeof(Attribute_t));

    for(i=0;i<blockCount;i++){
        blocks[i].lastCacheLines = allLines;
        allLines += blocks[i].memopCount;
    }
}
#endif

void processSamples(BufferEntry* entries,Attribute_t startIndex,Attribute_t lastIndex){

#ifdef DEBUG_RUN_2
    static Attribute_t howManyProcessSamplesCall = 0;
    printf("processSamples is called for %dth\n",howManyProcessSamplesCall++);
    fflush(stdout);
/*
    if (howManyProcessSamplesCall > 1){
        exit(0);
        return;
    }
*/

#endif
    register uint32_t systemIdx;
    register uint32_t accessIdx;
    register uint32_t level;

    for(systemIdx=0;systemIdx<systemCount;systemIdx++){
        register MemoryHierarchy* memoryHierarchy = (systems + systemIdx);

#ifdef VICTIM_CACHE

            // Find out if this memory hierarchy is an AMD w/ victim cache. This is
            // done by making sure it has the right number of cache levels, that all
            // but the highest level has replacement policy repl_lru_vc, and that
            // the replacement policy of the highest level is repl_lru.

            // We may want to add an attribute to the memory hierarchy for this. But
            // for now we search for it each time since it is the least intrusive way
            // to do it in the current framework.

        uint8_t isVictimCacheHierarchy = 1;

        if (memoryHierarchy->levelCount == __VICTIM_CACHE_LEVELS){
            uint32_t i;
            for (i = 0; i < __VICTIM_CACHE_LEVELS; i++){
                Cache* cache = &(memoryHierarchy->levels[i]);
                if (i < __VICTIM_CACHE_LEVELS - 1){
                    if (cache->attributes[replacement_policy] != repl_lru_vc){
                        isVictimCacheHierarchy = 0;
                    } 
                } else { // if i == __VICTIM_CACHE_LEVELS - 1
                    if (cache->attributes[replacement_policy] != repl_lru){
                        isVictimCacheHierarchy = 0;
                    }
                }
            }
        } else {
            isVictimCacheHierarchy = 0;
        }


            // Make sure a cache that isn't a VC doesn't try to use the VC policy for one of its caches
            // This shouldn't happen since generateCaches.pl won't let you set up such a cache, but it is bst to make sure.
        if (!isVictimCacheHierarchy){
            uint32_t i;
            for (i = 0; i < memoryHierarchy->levelCount; i++){
                Cache* cache = &(memoryHierarchy->levels[i]);
                if (cache->attributes[replacement_policy] == repl_lru_vc){
                    fprintf(stderr, "***** fatal error in instrumentation lib: cannot use replacement policy repl_lru_vc outside of a Victim Cache\n");
                    exit(-1);
                }
            }
        }
            
#endif // VICTIM_CACHE

        for(accessIdx=entries[startIndex-1].memOpId;accessIdx<=lastIndex;){
            register BufferEntry* currentEntry = (entries + accessIdx);
            register Attribute_t currentMemOp = currentEntry->memOpId;
            register Address_t currentAddress = currentEntry->address;

                // currentMemOp is set in such a way that values will be skipped by this code due to sampling
                // this is a bit of a hack. Ideally, we would use something like 0xffffffffffffffff as the "bad"
                // address value since addresses can be 0 sometimes.
            if(!currentAddress){
                if(currentMemOp > accessIdx){
                    accessIdx = currentMemOp;
                    continue;
                }
            }

            register BasicBlockInfo* currentBlock = (blocks + currentEntry->blockId);
            accessIdx++;

            register uint32_t isStrideCheckSystem = (memoryHierarchy->index == __STRIDE_TARGET_SYSTEM);

            register uint8_t levelCount = memoryHierarchy->levelCount;

#ifdef VICTIM_CACHE
            if (isVictimCacheHierarchy){

                register uint8_t hitLevel = __VICTIM_CACHE_LEVELS;
                register uint16_t leastRecent = 0;
                register uint16_t hitLine = 0;
                register Address_t victim = 0;

                Cache* cache[__VICTIM_CACHE_LEVELS];
                Attribute_t sizeInBits[__VICTIM_CACHE_LEVELS];
                Attribute_t setCount[__VICTIM_CACHE_LEVELS];
                Attribute_t assocCount[__VICTIM_CACHE_LEVELS];
                Address_t clineIdx[__VICTIM_CACHE_LEVELS];
                uint32_t setIdx[__VICTIM_CACHE_LEVELS];
                Address_t* content[__VICTIM_CACHE_LEVELS];


                    // Find which level (or main memory) the desired line resides in and which
                    // line is being replaced in that level.

                    // Search the caches starting at the lowest level. If we find the target,
                    // store the current level (hitLevel) and which line (leastRecent) we found the target
                for (level = 0; level < levelCount; level++){

                    cache[level] = &(memoryHierarchy->levels[level]);
                    sizeInBits[level] = cache[level]->attributes[line_size_in_bits];
                    setCount[level]  = cache[level]->attributes[number_of_sets];
                    assocCount[level] = cache[level]->attributes[set_associativity];
                    clineIdx[level] = currentAddress >> sizeInBits[level];
                    setIdx[level] = (clineIdx[level] % setCount[level]);
                    content[level] = cache[level]->content + (setIdx[level] * assocCount[level]);
                }
                for (level = 0; level < levelCount; level++){

                    register uint32_t lineInSet = 0;
                    register AccessStatus status = cache_miss;

                    if (assocCount[level] >= __MAX_LINEAR_SEARCH_ASSOC){
                        lineInSet = findInHash(clineIdx[level],setCount[level],assocCount[level],cache[level]->highAssocHash,setIdx[level]);
                        if(lineInSet < assocCount[level]){
                            status = cache_hit;
                            hitLine = lineInSet;
                            hitLevel = level;
                        }
                    } else {
                        for (lineInSet = 0; lineInSet < assocCount[level]; lineInSet++){ 
                            if(content[level][lineInSet] == clineIdx[level]){ 
                                status = cache_hit;  
                                hitLine = lineInSet;  
                                hitLevel = level;  
                                break;  
                            }
                        }
                    }
                    if (status == cache_hit){
                            // record that we hit in this level
                        cache[level]->hitMissCounters[cache_hit]++;
                        currentBlock->hitMissCounters[STATUS_IDX(systemIdx,level,cache_hit)]++;

                        if (level != __L1_CACHE_LEVEL){

                                // get the most recently used line. If the one we are removing is MRU, set
                                // the MRU to something else
                            leastRecent = (cache[level]->MOSTRECENT(setIdx[level]) + 1) % assocCount[level];


                            if (leastRecent == hitLine){
                                cache[level]->MOSTRECENT(setIdx[level]) = (cache[level]->MOSTRECENT(setIdx[level]) + (assocCount[level]-1)) % assocCount[level];
                            }

                                // move the found line out of the old cache
                            if(assocCount[level] >= __MAX_LINEAR_SEARCH_ASSOC){
                                deleteFromHash(content[level][hitLine],setCount[level],assocCount[level],cache[level]->highAssocHash,setIdx[level]);
                            }
                            content[level][leastRecent] = 0;
                        }
                        break;

                    } else { // status == cache_miss
                        cache[level]->hitMissCounters[cache_miss]++;
                        currentBlock->hitMissCounters[STATUS_IDX(systemIdx,level,cache_miss)]++;
                    }
                }


                    // If the line was found outside L1, evict something from L1 and move the line into L1
                if (hitLevel != __L1_CACHE_LEVEL){

                        // find out which line to replace
                    leastRecent = (cache[__L1_CACHE_LEVEL]->MOSTRECENT(setIdx[__L1_CACHE_LEVEL]) + 1) % assocCount[__L1_CACHE_LEVEL];

                        // get the victim
                    victim = content[__L1_CACHE_LEVEL][leastRecent];

                        // move the line into L1
                    if (assocCount[__L1_CACHE_LEVEL] >= __MAX_LINEAR_SEARCH_ASSOC){
                        insertInHash(clineIdx[__L1_CACHE_LEVEL], setCount[__L1_CACHE_LEVEL], assocCount[__L1_CACHE_LEVEL], 
                                     cache[__L1_CACHE_LEVEL]->highAssocHash, setIdx[__L1_CACHE_LEVEL], leastRecent);
                    }
                    content[__L1_CACHE_LEVEL][leastRecent] = clineIdx[__L1_CACHE_LEVEL];

                        // update most recent entries to the line we just updated
                    cache[__L1_CACHE_LEVEL]->MOSTRECENT(setIdx[__L1_CACHE_LEVEL]) = leastRecent;

                        // victimize all the way up the cache hierarchy
                    for (level = __L1_CACHE_LEVEL + 1; level < __VICTIM_CACHE_LEVELS; level++){

                        Address_t VCclineIdx = victim;
                        uint32_t VCsetIdx = (VCclineIdx % setCount[level]);
                        Address_t* VCcontent = cache[level]->content + (VCsetIdx * assocCount[level]);

                            // find where the victim goes
                        leastRecent = (cache[level]->MOSTRECENT(VCsetIdx) + 1) % assocCount[level];

                            // swap the entry being evicted and the victim from the lower level
                        victim = content[level][leastRecent];

                        if(assocCount[level] >= __MAX_LINEAR_SEARCH_ASSOC){
                            deleteFromHash(content[level][leastRecent],setCount[level],assocCount[level],cache[level]->highAssocHash,setIdx[level]);
                            insertInHash(victim,setCount[level],assocCount[level],cache[level]->highAssocHash,setIdx[level],leastRecent);
                        }
                        content[level][leastRecent] = VCclineIdx;

                        cache[level]->MOSTRECENT(VCsetIdx) = leastRecent;
                    }
                }
            } else {
#endif // VICTIM_CACHE

            for(level=0;level<levelCount;level++){
                register Cache* cache = &(memoryHierarchy->levels[level]);
                register Attribute_t sizeInBits = cache->attributes[line_size_in_bits];
                register Attribute_t setCount  = cache->attributes[number_of_sets];
                register Attribute_t assocCount = cache->attributes[set_associativity];

                register Address_t clineIdx = currentAddress >> sizeInBits;
                register uint32_t setIdx = (clineIdx % setCount);

                register Address_t* content = cache->content + (setIdx * assocCount);

                register uint16_t mostRecent = 0;
                register uint32_t i = 0;
                register AccessStatus status = cache_miss;

                if(assocCount >= __MAX_LINEAR_SEARCH_ASSOC){
                    i = findInHash(clineIdx,setCount,assocCount,cache->highAssocHash,setIdx);
                    if(i < assocCount){
                        status = cache_hit;
                        mostRecent = i;
                    }
                } else {
                    for(i=0;i<assocCount;i++){
                        if(content[i] == clineIdx){
                            status = cache_hit;
                            mostRecent = i;
                            break;
                        }
                    }
                }

                if(status == cache_hit){
                    cache->hitMissCounters[cache_hit]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,level,cache_hit)]++;
                } else {
                    cache->hitMissCounters[cache_miss]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,level,cache_miss)]++;

                    register Attribute_t replPolicy = cache->attributes[replacement_policy];
                    if(replPolicy == repl_ran){
                        mostRecent = getRandomNumber();
                    } else if(replPolicy == repl_lru){
                        mostRecent = (cache->MOSTRECENT(setIdx) + 1) % assocCount;
                    } else if(replPolicy == repl_dir){
                        mostRecent = 0;
                    }

                    if(assocCount >= __MAX_LINEAR_SEARCH_ASSOC){
                        deleteFromHash(content[mostRecent],setCount,assocCount,cache->highAssocHash,setIdx);
                        insertInHash(clineIdx,setCount,assocCount,cache->highAssocHash,setIdx,mostRecent);
                    }
                    content[mostRecent] = clineIdx;
                }
                cache->MOSTRECENT(setIdx) = mostRecent;

#ifdef STRIDE_VS_RANDOM
                if(!level && isStrideCheckSystem){
                    register uint32_t newStrideCheckLine = (setIdx * assocCount) + mostRecent;
                    register uint32_t oldStrideCheckLine = currentBlock->lastCacheLines[currentMemOp];
                    if(newStrideCheckLine != oldStrideCheckLine){
                        currentBlock->strideCounters[random_access]++;
                        currentBlock->lastCacheLines[currentMemOp] = newStrideCheckLine;
                    } else {
                        currentBlock->strideCounters[strided_access]++;
                    }
                }
#endif

                if(status == cache_hit){
                    break;
                }
            }
#ifdef VICTIM_CACHE
            }
#endif // VICTIM_CACHE

        }
    }
}

void MetaSim_simulFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){
#ifdef IGNORE_ACTUAL_SIMULATION
    register BufferEntry* entries = (BufferEntry*)base;
    entries->lastFreeIdx = 1;
    return;
#else
    register uint32_t      i;

    register BufferEntry* entries = (BufferEntry*)base;
    register Attribute_t startIndex = 1;
    register Attribute_t lastIndex = entries->lastFreeIdx;
    lastIndex--;

    totalNumberOfAccesses += lastIndex;

    if(blockCount && (saturatedBlockCount == blockCount)){
        entries->lastFreeIdx = 1;
        return;
    }

    lastSaturationAccess = totalNumberOfAccesses;

#ifndef NO_SAMPLING_MODE
    if(currentSamplingStatus == ignoring_accesses){
        alreadyIgnored += lastIndex;
        if(alreadyIgnored > __WHICH_IGNORING_VALUE){
            currentSamplingStatus = sampling_accesses;
            alreadySampled = alreadyIgnored - __WHICH_IGNORING_VALUE;
#ifdef EXTENDED_SAMPLING
            if(alreadySampled > lastIndex){
                startIndex = 1;
            } else {
#endif
            startIndex = (lastIndex - alreadySampled + 1);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = rand_value;
            rand_value = (Attribute_t)(random() % (SEGMENT_COUNT+1));
#ifdef DEBUG_RUN_1
            fprintf(stdout,"* %12lld IGNORED > %8lld sampled %8lld ignored %8lld [%6d,%6d] rand (%2d,%2d)\n",
                    totalNumberOfAccesses,__WHICH_IGNORING_VALUE,alreadySampled,alreadyIgnored,
                    startIndex,lastIndex,tmp_rand,rand_value);
#endif
            __WHICH_IGNORING_VALUE = __IGNORING_INTERVAL_MAX - (tmp_rand*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
            __WHICH_IGNORING_VALUE += (Counter_t)(rand_value*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
#endif
            alreadyIgnored = 0;
        } else {
            entries->lastFreeIdx = 1;
            return;
        }
    } else {
        alreadySampled += lastIndex;
        if(alreadySampled > __WHICH_SAMPLING_VALUE){
            currentSamplingStatus = ignoring_accesses;
            alreadyIgnored = alreadySampled - __WHICH_SAMPLING_VALUE;
#ifdef EXTENDED_SAMPLING
            if(alreadyIgnored > lastIndex){
                lastIndex = 0;
            } else {
#endif
            lastIndex = (lastIndex - alreadyIgnored);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = (random() % (SEGMENT_COUNT+1)) + (SEGMENT_COUNT/2);
#ifdef DEBUG_RUN_1
            fprintf(stdout,"- %12lld SAMPLED > %8lld ignored %8lld sampled %8lld [%6d,%6d] rand (%2d)\n",
                    totalNumberOfAccesses,__WHICH_SAMPLING_VALUE,alreadyIgnored,alreadySampled,
                    startIndex,lastIndex,tmp_rand);
#endif
            __WHICH_SAMPLING_VALUE = (Counter_t)((1.0*tmp_rand/SEGMENT_COUNT)*__SAMPLING_INTERVAL_MAX);
#endif
            alreadySampled = 0;
        }
    }
#endif /* NO_SAMPLING_MODE */

#ifdef EXTENDED_SAMPLING
    if(lastIndex < startIndex){
        entries->lastFreeIdx = 1;
#ifdef DEBUG_RUN_1
        fprintf(stdout,"DONE lastIndex < startIndex\n");
#endif
        return;
    }
#endif

    if(!blocks){
        char      appName[__MAX_STRING_SIZE];
        uint32_t  phaseId = 0;
        char      extension[__MAX_STRING_SIZE];

        sscanf(comment,"%s %u %s %u",appName,&phaseId,extension,&blockCount);
        blocks = (BasicBlockInfo*)malloc(sizeof(BasicBlockInfo) * blockCount);
        bzero(blocks,sizeof(BasicBlockInfo)*blockCount);
        initCaches();
#ifdef STRIDE_VS_RANDOM
        uint16_t* memopInfo = (uint16_t*)(entries + entryCount);
        initLastAccesses(memopInfo);
#endif
#ifdef COUNT_BB_EXECCOUNT
        uint64_t* counters = (uint64_t*)(entries + entryCount);
        for(i=0;i<blockCount;i++){
            blocks[i].counter = (counters + i);
        }
#endif
    }

    totalNumberOfSamples += (lastIndex-startIndex+1);

    register int32_t lastInvalidEntry = startIndex-1;

    register BasicBlockInfo* previousBlock = NULL;
    for(i=startIndex;i<=lastIndex;i++){
        register BufferEntry* currentEntry = (entries + i);
        register BasicBlockInfo* currentBlock = (blocks + currentEntry->blockId);
        register Attribute_t currentMemOp = currentEntry->memOpId;

        if(!currentMemOp || (currentBlock != previousBlock)){
            currentBlock->visitCount++;
            previousBlock = currentBlock;
        }
        if(currentBlock->visitCount > __MAXIMUM_BLOCK_VISIT){
            currentEntry->address = 0x0;
            if(lastInvalidEntry < 0){
                lastInvalidEntry = i;
            }
        } else {
            currentBlock->sampleCount++;
            if(lastInvalidEntry > -1){
                entries[lastInvalidEntry].memOpId = i;
                lastInvalidEntry = -1;
            }
            if(currentBlock->visitCount == __MAXIMUM_BLOCK_VISIT){
                currentBlock->saturationPoint = totalNumberOfAccesses;
                saturatedBlockCount++;
            }
        }
    }
    if(lastInvalidEntry > -1){
        entries[lastInvalidEntry].memOpId = i;
    }

    processSamples(entries,startIndex,lastIndex);

    entries->lastFreeIdx = 1;
#endif
}

float getPercentage(Counter_t val1,Counter_t val2){
    if(!val2){
        return 0.0;
    }
    return ((100.0*val1)/val2);
}

float getHitPercentage(BasicBlockInfo* block,uint32_t sys,uint32_t lvl){
    Counter_t total = block->hitMissCounters[STATUS_IDX(sys,lvl,cache_hit)] + 
                      block->hitMissCounters[STATUS_IDX(sys,lvl,cache_miss)];
    return getPercentage(block->hitMissCounters[STATUS_IDX(sys,lvl,cache_hit)],total);
}

void MetaSim_endFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){
#ifdef IGNORE_ACTUAL_SIMULATION
    BufferEntry* entries = (BufferEntry*)base;
    return;
#else
    uint32_t      i;
    uint32_t      j;
    uint32_t      k;

    BufferEntry* entries = (BufferEntry*)base;
    uint32_t lastIndex = entries->lastFreeIdx;
    lastIndex--;

    fprintf(stdout,"MetaSim_endFuncCall(0x%p,%d,%s,%d)\n",base,entryCount,comment,entries->lastFreeIdx);
#ifdef DEBUG_RUN_2
    fprintf(stdout,"HashSearches(%lld) Traversals(%lld)\n",howManyHashSearches,howManyHashTraverses);
#endif

    MetaSim_simulFuncCall_Simu(base,entryCount,comment);

    Counter_t processedSampleCount = 0;
    FILE* fp = openOutputFile(comment);
    if(fp){
        BasicBlockInfo* currentBlock = NULL;
        for(i=0;i<blockCount;i++){
            currentBlock = (blocks + i);
            if(blocks && currentBlock->sampleCount){
                processedSampleCount += currentBlock->sampleCount;
            }
        }
        fprintf(fp,"# buffer    = %d\n",entryCount);
        fprintf(fp,"# total     = %lld\n",totalNumberOfAccesses);
        fprintf(fp,"# sampled   = %lld\n",totalNumberOfSamples); 
        fprintf(fp,"# processed = %lld\n",processedSampleCount); 
        fprintf(fp,"# saturate  = %lld\n",lastSaturationAccess); 

#ifdef NO_SAMPLING_MODE
        fprintf(fp,"# sampling  = no\n");
#else
        fprintf(fp,"# sampling  = yes\n");
#ifdef EXTENDED_SAMPLING
        fprintf(fp,"# extended-sampling  = yes (intervals are averages)\n");
#endif
        fprintf(fp,"# ignoreint = %lld\n",(uint64_t)__IGNORING_INTERVAL_MAX);
        fprintf(fp,"# sampleint = %lld\n",(uint64_t)__SAMPLING_INTERVAL_MAX);
#endif

        fprintf(fp,"# maxvisits = %lld\n",(uint64_t)__MAXIMUM_BLOCK_VISIT);

#ifdef PER_SET_RECENT
        fprintf(fp,"# recentptr = per set\n");
#else
        fprintf(fp,"# recentptr = per cache\n");
#endif
        fprintf(fp,"#\n");
        for(j=0;j<systemCount;j++){
            MemoryHierarchy* memoryHierarchy = (systems + j);
            fprintf(fp,"# sysid%d =",memoryHierarchy->index);
            for(k=0;k<memoryHierarchy->levelCount;k++){
                Cache* cache = &(memoryHierarchy->levels[k]);
                Counter_t total = cache->hitMissCounters[cache_hit]+cache->hitMissCounters[cache_miss];
                fprintf(fp," l%1d[%lld,%lld(%5.2f)]", k,
                        total,
                        cache->hitMissCounters[cache_hit],
                        getPercentage(cache->hitMissCounters[cache_hit],total));
            }
            fprintf(fp,"\n");
        }

        fprintf(fp,"#\n");
#ifdef COUNT_BB_EXECCOUNT
        fprintf(fp,"#block <seqid> <visitcount> <samplingvisits> <samplecount>\n");
#else
        fprintf(fp,"#block <seqid> <visitcount> <samplecount>\n");
#endif
        fprintf(fp,"#\tsys <sysid> lvl <cachelvl> <hitcount> <miscount> <hitpercent>\n");
        fprintf(fp,"#\tstride <stridecount> <randcount> <randpercent>\n");
        fprintf(fp,"#\tsaturation <satpercent>\n");

        for(i=0;i<blockCount;i++){
            currentBlock = (blocks + i);
            if(blocks && currentBlock->sampleCount){

#ifdef COUNT_BB_EXECCOUNT
                fprintf(fp,"block\t%d\t%lld\t%lld\t%lld\n",i,*(currentBlock->counter),
                        currentBlock->visitCount,currentBlock->sampleCount);
#else
                fprintf(fp,"block\t%d\t%lld\t%lld\n",i,currentBlock->visitCount,currentBlock->sampleCount);
#endif
                for(j=0;j<systemCount;j++){
                    MemoryHierarchy* memoryHierarchy = (systems + j);
                    for(k=0;k<memoryHierarchy->levelCount;k++){
                        Cache* cache = &(memoryHierarchy->levels[k]);
                        Counter_t total = currentBlock->hitMissCounters[STATUS_IDX(j,k,cache_hit)] + 
                                          currentBlock->hitMissCounters[STATUS_IDX(j,k,cache_miss)];
                        fprintf(fp,"\tsys\t%3d\tlvl\t%1d\t%lld\t%lld\t%10.6f\n",
                            memoryHierarchy->index,k+1,
                            currentBlock->hitMissCounters[STATUS_IDX(j,k,cache_hit)],
                            currentBlock->hitMissCounters[STATUS_IDX(j,k,cache_miss)],
                            getHitPercentage(currentBlock,j,k));
                    }
                }

#ifdef STRIDE_VS_RANDOM
                Counter_t strideCount = currentBlock->strideCounters[strided_access];
                Counter_t randomCount = currentBlock->strideCounters[random_access];
                fprintf(fp,"\tstride\t%lld\t%lld\t%10.6f\n",strideCount,randomCount,
                                             getPercentage(randomCount,strideCount+randomCount)/100.0);
#endif

                if(!currentBlock->saturationPoint){
                    fprintf(fp,"\tsaturation\t%5.2f\n",100.0);
                } else {
                    fprintf(fp,"\tsaturation\t%5.2f\n",getPercentage(currentBlock->saturationPoint,
                                                                   totalNumberOfAccesses));
                }
#ifdef COUNT_BB_EXECCOUNT
            } else if(blocks) {
                fprintf(fp,"## block\t%d\t%lld\t%lld\t%lld\n",i,*(currentBlock->counter),
                        currentBlock->visitCount,currentBlock->sampleCount);
#endif
            }
        }

        fclose(fp);
    }
#endif
}
