#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <NewCacheSimulator.h>

/*** This method is called by the instrumentation when the address buffer
     is full. The address buffer is a list of records of type AddressEntry.
     Each AddressEntry contains address, basic block sequence id, memory instruction
     index within the block of each memory access. The first entry in the address
     stream buffer entry does not contain a valid address. Instead, the blockId
     (also defined as lastFreeIdx) contains valid data and it points to the first index
     in the buffer which is empty. Thus, at the end of simulation this field has
     to be set to 1 (addressStream->lastFreeIdx = 1) to indicate buffer is processed.
     Moreover, the valid data is in the indecies from 1 to (addressStream->lastFreeIdx-1).

     Note that for 32-bit and 64-bit executables, the size of virtiual address changes
     thus the address type is set to be 32-bit or 64-bit unsigned integer depending on
     the executable. The macro METASIM_32_BIT_LIB can be used as in NewCacheSimulator.h
     for correct representation of 32bit and 64bit libraries.

     arguments
     @ char* base : pointer to the beginning of address stream buffer. each entry is
                    a record of AddressEntry. 
     @ uint32_t entryCount : size of the address stream buffer. Note that it is not
                    the number of valid entries. It is the size of buffer in elements.
     @ const char* comment : a string contains additional information passed from the
                    executable to this function (name of the exec, suffix, and so on).
***/

void MetaSim_simulFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){

    AddressEntry* addressStream = (AddressEntry*)base;
    Attribute_t   startIndex = 1;
    Attribute_t   lastIndex = addressStream->lastFreeIdx;
    lastIndex--;

    printf("Address stream will be processed here. Args are [0x%x] [%d] [%s] -- last entry at (%d)\n",
            base,entryCount,comment,lastIndex); 

    processSamples(addressStream,startIndex,lastIndex);

    /*** End of simulation. Set the lastFreeIdx to 1 to 
         set address stream buffer to be empty. ***/
    addressStream->lastFreeIdx = 1;
}

/*** This method is called at program termination. So the results of
     simulations is printed in this method. Arguments are the same as in 
     MetaSim_simulFuncCall_Simu function above.
     The users are expected to include their code in this module
     to save/print/post-process simulation results, as shown below.
***/

void MetaSim_endFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){

    AddressEntry* addressStream = (AddressEntry*)base;
    uint32_t      lastIndex = addressStream->lastFreeIdx;
    lastIndex--;

    /*** Simulate the remaining address stream
         (since last call to MetaSim_simulFuncCall_Simu) ***/
    MetaSim_simulFuncCall_Simu(base,entryCount,comment);

    /*** Process/print your cache simulation results (termination)
              |  |   |   | 
              V  V   V   V  
    ****/

    printf("Application termination. Print your simulation results\n");
}

/*** This method needs to be implemented by users according to their simulation needs.
     Currently the method goes over each address in the stream and stores block and memory
     operation ids.
***/

void processSamples(AddressEntry* addressStream,Attribute_t startIndex,Attribute_t lastIndex)
{
    /*** Your cache simulation methodology (For the given address sequence)
                   |  |   |   |
                   V  V   V   V
    ***/

    /*** Iterate over the address entries in the address stream ***/
    uint32_t i = 0;
    for(i=startIndex;i<=lastIndex;i++){
        /*** Each entry contains virtual address (32 bit or 64 bit depending
             on the exceutable's being 32bit or 64bit), sequence id of the block
             which requested the access and the memory operation index within the block
             that requested the address ***/

        AddressEntry* currentEntry = (addressStream + i);
        Attribute_t   blockSeqId = currentEntry->blockId;
        Attribute_t   memOpIdx = currentEntry->memOpId;

        /*** Now we have an address and its requestor basic block and memory op
             id. Simulate using your own modules.***/

    }
}
