#ifndef _AddressSpaceCoverage_h_
#define _AddressSpaceCoverage_h_

typedef uint32_t     Attribute_t;
typedef uint64_t     Counter_t;    

#ifdef METASIM_32_BIT_LIB
    typedef uint32_t     Address_t;
#else
    typedef uint64_t     Address_t;
#endif

typedef struct {
    Attribute_t blockId;
    Attribute_t memOpId;
#ifdef METASIM_32_BIT_LIB
    Attribute_t unused;
#endif
    Address_t   address;
} AddressEntry;

#define lastFreeIdx blockId

void processSamples(AddressEntry*,Attribute_t,Attribute_t);

/*** Define data structures needed for cache 
     simulation functions, such as to keep track of
     hit/miss counts, memory hierarchies and etc. 
     |            |           |             |
     V            V           V             V
***/

#define __MAX_STRING_SIZE             1024

  #ifdef FINE_GRAIN_SAMPLING
    #define __SAMPLING_INTERVAL_MAX     100000
    #define __IGNORING_INTERVAL_MAX    1000000
  #else
    #define __SAMPLING_INTERVAL_MAX    1000000
    #define __IGNORING_INTERVAL_MAX   10000000
  #endif

  #ifdef EXTENDED_SAMPLING
    Counter_t sampling_interval_max = __SAMPLING_INTERVAL_MAX;
    Counter_t ignoring_interval_max = __IGNORING_INTERVAL_MAX;
    #define __WHICH_SAMPLING_VALUE  sampling_interval_max
    #define __WHICH_IGNORING_VALUE  ignoring_interval_max
    #define SEGMENT_COUNT 10  /* needs to divide both intervals */
    Attribute_t rand_value = SEGMENT_COUNT;
  #else
    #define __WHICH_SAMPLING_VALUE  __SAMPLING_INTERVAL_MAX
    #define __WHICH_IGNORING_VALUE  __IGNORING_INTERVAL_MAX
  #endif

typedef enum {
    sampling_accesses = 0,
    ignoring_accesses
} SamplingStatus;

SamplingStatus currentSamplingStatus = ignoring_accesses;
Counter_t alreadyIgnored = 0;
Counter_t alreadySampled = 0;

typedef enum {
   no_sampling = 0,
   event_sampling,
   address_sampling,
   Total_SamplingMethods
} AddressSpaceSampling;


#endif
