#include <SectHeader.h>
#include <Iterator.h>
#include <Function.h>
#include <LineInfoFinder.h>
#include <XCoffFile.h>
#include <CommonMethods.h>
#include <Loop.h>
#include <fstream>
using namespace std;


char*  CommonMethods::staticPrintExtension = "static";

CommonMethods::CommonMethods(XCoffFile* xcoff,char* extension,uint32_t phaseNo,uint32_t callCount,char* inpFile)
    : XCoffFileGen(xcoff,extension,phaseNo,inpFile),libraryFunctionCount(callCount),
      blocksToInstrument(NULL),terminationBlock(NULL),numOfBlocksFiltered(0),
      extendedBeginAddr(0),
      execTOCAddress(0),tocXorUpdateVal(0),tocDistUpdateVal(0),
      sharedLibraryInfoStr(NULL),enableExtendedPrint(false)
{
}

BasicBlock* CommonMethods::findTerminationBlock(){

    uint32_t allBlocksCount = xcoffFile->getNumberOfBlocks();
    if(!allBlocksCount)
        return NULL;

    BasicBlock** allBlocks = new BasicBlock*[allBlocksCount];
    xcoffFile->getAllBlocks(allBlocks);

    BasicBlock* mpiexit = NULL;
    BasicBlock* regexit = NULL;
    BasicBlock* pmacexit = NULL;

    for (uint32_t i = 0; i < allBlocksCount; i++){
        BasicBlock* bb = allBlocks[i];
        if (bb->isTrace() || bb->isJumpTable() || bb->getIndex())
            continue;
        Function* owner = bb->getFlowGraph()->getFunction();
        if (!pmacexit && owner->isAnySymbolA(".end_pmac_fromc")){
            pmacexit = bb;
        } 
        if (!mpiexit && 
            (owner->isAnySymbolA(".mpi_finalize") ||
             owner->isAnySymbolA(".MPI_Finalize"))){
            mpiexit = bb;
        } 
        if (!regexit && owner->isAnySymbolA(".exit")){
            regexit = bb;
        }
    }
    delete[] allBlocks;

    if (pmacexit) {
        PRINT_INFOR("The exit function .end_pmac_fromc is taken");
        return pmacexit;
    }
    if (mpiexit){
        PRINT_INFOR("The exit function .mpi_finalize is taken");
        return mpiexit;
    }
    if (regexit){
        PRINT_INFOR("The exit function .exit is taken");
        return regexit;
    }
    return NULL;
}

void CommonMethods::printInstrumentationPoints(){

    char instBlockFileName[__MAX_STRING_SIZE];
    if (!getPhaseIndex()){
        sprintf(instBlockFileName,"%s.%s.%s",xcoffFile->getXCoffFileName(),getFileExtension(),staticPrintExtension);
    } else {
        sprintf(instBlockFileName,"%s.phase.%d.%s.%s",xcoffFile->getXCoffFileName(),getPhaseIndex(),
                getFileExtension(),staticPrintExtension);
    }

    ofstream instBlockFile;
    instBlockFile.open(instBlockFileName);
    if (!instBlockFile){
         PRINT_ERROR("Instrumented block infor can not be opened [%s]",instBlockFileName);
    }

    uint32_t memOpCount = 0;
    uint32_t fpOpCount = 0;
    uint32_t insnCount = 0;

    for (uint32_t i=0;i<numberOfInstPoints;i++){
        memOpCount += blocksToInstrument[i]->getNumOfMemoryOps();
        fpOpCount += blocksToInstrument[i]->getNumOfFloatPOps();
        insnCount += blocksToInstrument[i]->getInstructionCount();
    }

    TextSection* lineInfoSect = (TextSection*)xcoffFile->getRawSection(xcoffFile->getTextSectionIndex());
    LineInfoFinder* lineInfoFinder = lineInfoSect->getLineInfoFinder();

    instBlockFile << "# appname   = " << xcoffFile->getXCoffFileName() << endl;
    instBlockFile << "# appsize   = " << xcoffFile->getFileSize() << endl;
    instBlockFile << "# extension = " << getFileExtension() << endl;
    instBlockFile << "# phase     = " << getPhaseIndex() << endl;
    instBlockFile << "# type      = " << getGenClassName() << endl;
    instBlockFile << "# candidate = " << xcoffFile->getNumberOfBlocks() << endl;
    instBlockFile << "# blocks    = " << numberOfInstPoints << endl;
    instBlockFile << "# memops    = " << memOpCount << endl;
    instBlockFile << "# fpops     = " << fpOpCount << endl;
    instBlockFile << "# insns     = " << insnCount << endl;
    instBlockFile << "# buffer    = " << howManyBufferEntry() << endl;
    instBlockFile << "# library   = " << getPathToInstLib() << endl;
    instBlockFile << "# libTag    = " << VersionTag::getLibraryVersionTag() << endl;
    instBlockFile << "# " << additionalInfoToPrint() << endl;
    if(isExtendedPrint()){
        instBlockFile << "# <sequence> <block_uid> <memop> <fpop> <insn> <line> <fname>";
        instBlockFile << " <loopcnt> <loopid> <ldepth> <hex_uid> <vaddr>" << endl;
    } else {
        instBlockFile << "# <sequence> <block_uid> <memop> <fpop> <insn> <line> <fname>";
        instBlockFile << " <hex_uid> <vaddr>" << endl;
    }

    for (uint32_t i = 0; i < numberOfInstPoints; i++){
        BasicBlock* bb = blocksToInstrument[i];
        uint64_t unq = bb->getHashCode().getValue();
        Function* owner = bb->getFlowGraph()->getFunction();
        char names[__MAX_STRING_SIZE];
        names[0] = '\0';
        owner->getAllSymbolNames(names,__MAX_STRING_SIZE,true);
        if(!strlen(names)){
            strcpy(names,"__name__unknown__");
        }
        char* absFile = NULL;
        uint32_t absLine  =0;
        if(lineInfoFinder){
            uint64_t addr = bb->getBaseAddress();
            absFile = lineInfoFinder->getFileName(addr);
            absLine = lineInfoFinder->getLineNumberInFile(addr);
        }

        uint32_t loopId = Invalid_UInteger_ID;
        uint32_t loopDepth = 0;
        uint32_t loopCount = bb->getFlowGraph()->getNumberOfLoops();

        if(isExtendedPrint()){
            for(uint32_t j=0;j<loopCount;j++){
                Loop* currLoop = bb->getFlowGraph()->getLoop(j);
                if(currLoop->isBlockIn(bb->getIndex())){
                    loopDepth++;
                    loopId = currLoop->getIndex();
                }
            }
        }

        instBlockFile << i << "\t" << unq 
                      << "\t" << bb->getNumOfMemoryOps()
                      << "\t" << bb->getNumOfFloatPOps()
                      << "\t" << bb->getInstructionCount()
                      << "\t" << (absFile ? absFile : "__FILE_UNK__") << ":" << absLine 
                      << "\t" << names << "\t#";
        if(isExtendedPrint()){
            instBlockFile << "\t" << loopCount << "\t" << (int32_t)loopId << "\t" << loopDepth;
        }
        instBlockFile << "\t 0x" << hex << unq << "\t0x" << bb->getBaseAddress() << dec 
                      << endl;    
    }
    instBlockFile.close();

    char sysCommand[__MAX_STRING_SIZE];
    sprintf(sysCommand,"chmod a-w %s",instBlockFileName);
//  system(sysCommand);
}

void CommonMethods::reserveDataForInstrumentation(){

    extendedDataSize = 0;
    extendedBeginAddr = 0;
    dataBufferEntryCount = 0;

    uint64_t dataSectionBeginAddr = xcoffFile->getDataSectionVAddr();
    uint32_t dataSectionSize = xcoffFile->getDataSectionSize();

    execTOCAddress = xcoffFile->getTOCAddress();

    uint64_t currentFreeAddress = dataSectionBeginAddr + dataSectionSize;

    uint32_t incrementSize = 0;
    if (xcoffFile->is64Bit()){
        currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
        incrementSize = sizeof(uint64_t);
    } else {
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
        incrementSize = sizeof(uint32_t);
    }

    extendedBeginAddr = currentFreeAddress;

    for (uint32_t i = 0; i < getNumOfSharedLibFuncs(); i++){
        libraryFunctionStubAddrLocs[i] = currentFreeAddress;
        currentFreeAddress += incrementSize;
    }

    currentFreeAddress = nextAlignAddressWord(currentFreeAddress);

    saveAddressCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);
    saveAddressFPCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);

    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);

    saveAddressLinkReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressCtrReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressExpReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);

    saveAddressGprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_GPR_REGS);
    saveAddressFprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_FPR_REGS);
        
    char* ptr = xcoffFile->getXCoffFileName();
    char* sepPtr = strrchr(ptr,'/');
    if(sepPtr){
        ptr = (sepPtr + 1);
    }
    char tmpStr[__MAX_STRING_SIZE];
    if(!phaseIndex){
        sprintf(tmpStr,"%s 0 %s %d",ptr,getFileExtension(),
                getNumberOfInstPoints(NULL));
    } else {
        sprintf(tmpStr,"%s %d %s %d",ptr,getPhaseIndex(),getFileExtension(),
                getNumberOfInstPoints(NULL));
    }

    commentAddress = currentFreeAddress;
    sharedLibraryInfoStr = strdup(tmpStr);
    currentFreeAddress += (strlen(sharedLibraryInfoStr) + 1);

    dataBufferEntryCount = howManyBufferEntry();
    incrementSize = bytesPerBufferEntry();

    if(incrementSize == sizeof(uint32_t)){
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
    } else {
        currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
    }

    dataBufferAddress = currentFreeAddress;
    currentFreeAddress += (dataBufferEntryCount * incrementSize);

    extraBlockInfoAddress = currentFreeAddress;
    currentFreeAddress += spaceForBlockInfo();

    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);

    extendedDataSize = currentFreeAddress - (dataSectionBeginAddr + dataSectionSize);

    tocXorUpdateVal = 0;
    if(!Instruction::isInLoadOffsetInsnRange((int32_t)(dataBufferAddress-execTOCAddress))){
        tocXorUpdateVal = (uint32_t)(extendedBeginAddr ^ execTOCAddress);
        tocDistUpdateVal = (uint32_t)(extendedBeginAddr - execTOCAddress);
    }

    PRINT_INFOR("**** Additional data section inc  is %d ***",extendedDataSize);
    PRINT_INFOR("**** Number of entries to process is %18d ***",dataBufferEntryCount);
    PRINT_INFOR("**** Start address for toc        is %#18llx (%lld) ***",
        execTOCAddress,execTOCAddress-dataSectionBeginAddr);
    PRINT_INFOR("**** Start address for additional is %#18llx (%lld) ***",
        extendedBeginAddr,extendedBeginAddr-execTOCAddress);
    PRINT_INFOR("**** The tocXorUpdateVal for save is %x %d ***",
        tocXorUpdateVal,tocDistUpdateVal);
    PRINT_INFOR("**** Start address for the gprs   is %#18llx (%lld) ***",
        saveAddressGprRegs,saveAddressGprRegs-execTOCAddress);
    PRINT_INFOR("**** Start address for the comm   is %#18llx (%lld) ***",
        commentAddress,commentAddress-execTOCAddress);
    PRINT_INFOR("**** Start address for the buffer is %#18llx (%lld) ***",
        dataBufferAddress,dataBufferAddress-execTOCAddress);
    PRINT_INFOR("**** End address for the buffer   is %#18llx (%lld) ***",
        currentFreeAddress,currentFreeAddress-execTOCAddress);
}

void CommonMethods::updateTOCRegister(Instruction* insnBuffer,uint32_t* insnIndex){
    uint16_t tocUpdateValLow = FIRST_HALFWORD(tocXorUpdateVal);
    uint16_t tocUpdateValHigh = SECOND_HALFWORD(tocXorUpdateVal);
    if(tocXorUpdateVal){
        if(tocUpdateValHigh){
            insnBuffer[(*insnIndex)++] = Instruction::generateXorImmShifted(REG_TOC,REG_TOC,tocUpdateValHigh);
        }
        insnBuffer[(*insnIndex)++] = Instruction::generateXorImm(REG_TOC,REG_TOC,tocUpdateValLow);
    }
}
void CommonMethods::saveCTRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getCTRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromSPR(reg,REG_CTR);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreDouble(reg,REG_TOC,saveOffset);
}
void CommonMethods::restCTRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getCTRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadDouble(reg,REG_TOC,saveOffset);
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToSPR(reg,REG_CTR);
}
void CommonMethods::saveEXRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getEXRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromSPR(reg,REG_XER);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreDouble(reg,REG_TOC,saveOffset);
}
void CommonMethods::restEXRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getEXRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadDouble(reg,REG_TOC,saveOffset);
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToSPR(reg,REG_XER);
}
void CommonMethods::saveFPSCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getFPSCRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromFPSCR(reg);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreWordFloat(reg,REG_TOC,saveOffset);
}
void CommonMethods::restFPSCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset = getFPSCRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadWordFloat(reg,REG_TOC,saveOffset);
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToFPSCR(reg);
}
void CommonMethods::saveCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset   = getCRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromCR(reg);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreWord(reg,REG_TOC,saveOffset);
}
void CommonMethods::restCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset   = getCRSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadWord(reg,REG_TOC,saveOffset);
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToCR(reg);
}
void CommonMethods::saveLNKRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset   = getLNKSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromSPR(reg,REG_LNK);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreDouble(reg,REG_TOC,saveOffset);
}
void CommonMethods::restLNKRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg){
    int32_t saveOffset   = getLNKSaveOffset();
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadDouble(reg,REG_TOC,saveOffset);
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToSPR(reg,REG_LNK);
}


void CommonMethods::generateFuncCallFull(Instruction* insnBuffer,uint32_t* insnIndex,
                                                uint32_t reg,uint64_t stubAddr,uint64_t funcAddr)
{
//    PRINT_INFOR("Generating a function call with args %d %d %#llx %#llx", *insnIndex, reg, stubAddr, funcAddr);

    saveCRRegister(insnBuffer,insnIndex,reg);
    saveLNKRegister(insnBuffer,insnIndex,reg);

    updateTOCRegister(insnBuffer,insnIndex);

    uint64_t fromAddr = stubAddr + ((*insnIndex) * sizeof(uint32_t));
    uint64_t toAddr   = funcAddr;
    insnBuffer[(*insnIndex)++] = Instruction::generateCallToImmediate(fromAddr,toAddr);

    updateTOCRegister(insnBuffer,insnIndex);

    restLNKRegister(insnBuffer,insnIndex,reg);
    restCRRegister(insnBuffer,insnIndex,reg);
}

uint32_t CommonMethods::generateSharedLibFuncWrapper(
    uint32_t libFuncIdx,uint64_t funcCallAddr,uint32_t genBufferOffset,BaseGen* gen)
{


    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[byteCountForSharedLibFuncWrappers() / sizeof(uint32_t)];

    uint32_t scratchReg = 6;
    uint32_t fpScratchReg = 3;
    uint32_t savedReg = 31;
    uint32_t tocHoldReg = 30;
    uint32_t firstArgReg = 3;
    uint32_t secondArgReg = 4;
    uint32_t thirdArgReg = 5;

    ASSERT(scratchReg == 6);
    ASSERT(fpScratchReg == 3);
    ASSERT(savedReg == 31);
    ASSERT(tocHoldReg == 30);
    ASSERT(firstArgReg == 3);
    ASSERT(secondArgReg == 4);
    ASSERT(thirdArgReg == 5);

    int32_t saveOffset = 0x0;

    updateTOCRegister(insnBuffer,&insnCount);

    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(i,REG_TOC,saveOffset);
    }
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDoubleFloat(i,REG_TOC,saveOffset);
    }

    saveCTRRegister(insnBuffer,&insnCount,scratchReg);
    saveEXRRegister(insnBuffer,&insnCount,scratchReg);
    saveFPSCRRegister(insnBuffer,&insnCount,fpScratchReg);

    uint32_t bufferOffset = getDataOffsetForInst(0);
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,firstArgReg);
    insnBuffer[insnCount++] = Instruction::generateIncrement(firstArgReg,bufferOffset);

    uint32_t bufferCount = getDataBufferEntryCount();
    insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(secondArgReg,bufferCount);
    insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(secondArgReg,bufferCount);

    uint32_t commentOffset = getCommentAddressOffset();
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,thirdArgReg);
    insnBuffer[insnCount++] = Instruction::generateIncrement(thirdArgReg,commentOffset);

    insnBuffer[insnCount++] = Instruction::generateMoveFromSPR(savedReg,REG_LNK);
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,tocHoldReg);


    int32_t callOffset = funcCallAddr - execTOCAddress - tocDistUpdateVal;
    if(xcoffFile->is64Bit()){
        insnBuffer[insnCount++] = Instruction::generateSPIncrementDouble(-SAFE_FRAME_LOC*sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,callOffset);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(REG_TOC,scratchReg,sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,scratchReg,0);
    } else {
        insnBuffer[insnCount++] = Instruction::generateSPIncrementWord(-SAFE_FRAME_LOC*sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,REG_TOC,callOffset);
        insnBuffer[insnCount++] = Instruction::generateLoadWord(REG_TOC,scratchReg,sizeof(uint32_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,scratchReg,0);
    }
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(scratchReg,REG_CTR);
    insnBuffer[insnCount++] = Instruction::generateCallToCTR();

    insnBuffer[insnCount++] = Instruction::generateMoveReg(tocHoldReg,REG_TOC);
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(savedReg,REG_LNK);

    restFPSCRRegister(insnBuffer,&insnCount,fpScratchReg);
    restEXRRegister(insnBuffer,&insnCount,scratchReg);
    restCTRRegister(insnBuffer,&insnCount,scratchReg);

    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDoubleFloat(i,REG_TOC,saveOffset);
    }
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(i,REG_TOC,saveOffset);
    }
    updateTOCRegister(insnBuffer,&insnCount);

    insnBuffer[insnCount++] = Instruction::generateReturnToLnk();

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    return (insnCount*sizeof(uint32_t));
}
