#ifndef _CompareCacheSimulations_h_
#define _CompareCacheSimulations_h_

typedef uint32_t     Attribute_t;  
typedef uint64_t     Counter_t;    
#ifdef METASIM_32_BIT_LIB
  typedef uint32_t     Address_t;
#else
  typedef uint64_t     Address_t;
#endif

#ifdef NO_SAMPLING_MODE
  #define __MAXIMUM_BLOCK_VISIT       500000
#else
  #define __MAXIMUM_BLOCK_VISIT        50000
  #ifdef FINE_GRAIN_SAMPLING
    #define __SAMPLING_INTERVAL_MAX     100000
    #define __IGNORING_INTERVAL_MAX    1000000
  #else
    #define __SAMPLING_INTERVAL_MAX    1000000
    #define __IGNORING_INTERVAL_MAX   10000000
  #endif

  #ifdef EXTENDED_SAMPLING
    Counter_t sampling_interval_max = __SAMPLING_INTERVAL_MAX;
    Counter_t ignoring_interval_max = __IGNORING_INTERVAL_MAX;
    #define __WHICH_SAMPLING_VALUE  sampling_interval_max
    #define __WHICH_IGNORING_VALUE  ignoring_interval_max
    #define SEGMENT_COUNT 10  /* needs to divide both intervals */
    Attribute_t rand_value = SEGMENT_COUNT;
  #else
    #define __WHICH_SAMPLING_VALUE  __SAMPLING_INTERVAL_MAX
    #define __WHICH_IGNORING_VALUE  __IGNORING_INTERVAL_MAX
  #endif
#endif

#define __MAX_LEVEL                      3
#define __MAX_STRING_SIZE             1024

typedef enum {
    repl_lru = 0,
    repl_ran,
    repl_dir,
    Total_ReplacementPolicy
} ReplacementPolicy;

typedef enum {
    number_of_sets = 0,
    set_associativity,
    line_size_in_bits,
    replacement_policy,
    Total_CacheAttributes
} CacheAttributes;

typedef enum {
    cache_hit = 0,
    cache_miss,
    Total_AccessStatus
} AccessStatus;

typedef enum {
   no_sampling,
   event_sampling,
   address_sampling,
   Total_SamplingType
} SamplingType;
    
typedef struct {
    Attribute_t attributes[Total_CacheAttributes];
    Counter_t   hitMissCounters[Total_AccessStatus*Total_SamplingType];
#ifdef PER_SET_RECENT
    uint16_t*   mostRecent;
#else
    uint16_t    mostRecent[Total_SamplingType];
#endif
    Address_t*  content;
} Cache;

typedef struct {
    uint8_t   index;
    uint8_t   levelCount;
    Cache     levels[__MAX_LEVEL]; 
} MemoryHierarchy;

#ifndef NO_SAMPLING_MODE
typedef enum {
    sampling_accesses = 0,
    ignoring_accesses
} SamplingStatus;

SamplingStatus currentSamplingStatus = ignoring_accesses;
Counter_t alreadyIgnored = 0;
Counter_t alreadySampled = 0;
#endif

typedef struct {
    Attribute_t blockId;
    Attribute_t memOpId;
#ifdef METASIM_32_BIT_LIB
    Attribute_t unused;
#endif
    Address_t   address;
} BufferEntry;

#define lastFreeIdx blockId

#include <CacheStructures.h>


typedef struct {
#ifdef USE_SATURATION
    Counter_t saturationPoint;
#endif
    Counter_t visitCount[Total_SamplingType];
/*    Counter_t addressCount[__SYSTEM_COUNT*__MAX_LEVEL]; */
    Counter_t hitMissCounters[__SYSTEM_COUNT*__MAX_LEVEL*Total_AccessStatus*Total_SamplingType];
} BasicBlockInfo;

#endif /* _CompareCacheSimulations_h_ */
