#include <SectHeader.h>
#include <Iterator.h>
#include <Function.h>
#include <LineInfoFinder.h>
#include <XCoffFile.h>
#include <CustomDataExtender.h>

CustomDataExtender::CustomDataExtender (XCoffFile* xcoff, char* extension, char* inptName)
      : CommonMethods (xcoff, extension, 0, 1, NULL), 
        inputFile (inptName)
{
    extraDataSize = 0;
    binaryFile = 0;
}

void 
CustomDataExtender::verify (uint32_t written) {
    PRINT_INFOR ("Verifying the two files are identical");
}

uint32_t 
CustomDataExtender::bytesPerBufferEntry () {
    // parse the input file and determine how much text to include
    char buffer[__MAX_STRING_SIZE];
    if (extraDataSize)
       return (extraDataSize);

    extraDataSize += sizeof (uint32_t);
    FILE *fd = fopen (inputFile, "rt");
    if (fd == NULL) 
    {
       sprintf (buffer, "CustomDataExtender: cannot open for read input file %s\n", inputFile);
       extraDataSize += sizeof (uint32_t) + strlen (buffer);
       extraDataSize = nextAlignAddressWord (extraDataSize);
    } else
    {
       uint32_t l;
       while ((l=fread (buffer, sizeof(char), __MAX_STRING_SIZE, fd))>0)
       {
          // I detect if the file contains any non printable characters (if it
          // is considered a binary or text file)
          if (!binaryFile)
             for (uint32_t j=0 ; j<l ; ++j)
                if (!buffer[j] || buffer[j]>127)
                {
                   binaryFile = 1;
                   break;
                }
          extraDataSize += sizeof (uint32_t) + l;
          extraDataSize = nextAlignAddressWord (extraDataSize);
       }
       fclose (fd);
    }
    extraDataSize += sizeof (uint32_t);
    return (extraDataSize);
}

void 
CustomDataExtender::initializeReservedData (DataSection* dataSect, BaseGen* gen) 
{
    char buffer[__MAX_STRING_SIZE];

    gen->writeString (commentAddress-dataSect->getSectHeader()->GET(s_vaddr),
                     sharedLibraryInfoStr, strlen(sharedLibraryInfoStr) + 1);
    
    FILE *fd = fopen (inputFile, "rt");
    uint64_t currentFreeAddress = dataBufferAddress;
    gen->writeWord (currentFreeAddress-dataSect->getSectHeader()->GET(s_vaddr), binaryFile);
    currentFreeAddress += sizeof (uint32_t);
    if (fd == NULL) 
    {
       sprintf (buffer, "CustomDataExtender: cannot open for read input file %s\n", inputFile);
       gen->writeWord (currentFreeAddress- dataSect->getSectHeader()->GET(s_vaddr), strlen (buffer));
       currentFreeAddress += sizeof (uint32_t);
       gen->writeString (currentFreeAddress- dataSect->getSectHeader()->GET(s_vaddr), buffer, strlen(buffer));
       currentFreeAddress += strlen (buffer);
       currentFreeAddress = nextAlignAddressWord (currentFreeAddress);
    } else
    {
       uint32_t nreads = 0, l;
       while ((l=fread (buffer, sizeof(char), __MAX_STRING_SIZE, fd))>0)
       {
          gen->writeWord (currentFreeAddress- dataSect->getSectHeader()->GET(s_vaddr), l);
          currentFreeAddress += sizeof (uint32_t);
          gen->writeString (currentFreeAddress- dataSect->getSectHeader()->GET(s_vaddr), buffer, l);
          currentFreeAddress += l;
          currentFreeAddress = nextAlignAddressWord (currentFreeAddress);
          ++ nreads;
       }
       fclose (fd);
       fprintf (stdout, "Num reads from the input file %d\n", nreads);
    }
    gen->writeWord (currentFreeAddress, 0);
}

void 
CustomDataExtender::getSharedLibraryPathAndObj (char** path,char** name,char** obj)
{
    *path = getPathToInstLib();
    *obj  = "";
    if(xcoffFile->is64Bit())
        *name = "libprintExtD_64.a";
    else
        *name = "libprintExtD.a";
    ASSERT(path && name && obj);
}

void 
CustomDataExtender::selectInstrumentationPoints (char* blockFile)
{
    terminationBlock = findTerminationBlock();
    ASSERT(terminationBlock && "FATAL : There is no exit block");
    fprintf (stdout, "termination block at %llx, size %d bytes\n",
        terminationBlock->getBaseAddress(), terminationBlock->getSizeInBytes());

    char buffer[__MAX_STRING_SIZE];
    uint32_t howMany = 0;

    blocksToInstrument = new BasicBlock*[howMany+1];

    uint32_t lineNo = 0;

    bool insertAppExitBlock = true;

    if (insertAppExitBlock) {
        blocksToInstrument[numberOfInstPoints++] = terminationBlock;
    }

    if (!Instruction::isInLoadOffsetInsnRange (numberOfInstPoints-1)) {
        PRINT_ERROR("At cache simulation there can only be at most %d blocks instrumented",0x7fff);
    }

    PRINT_INFOR("**** Number of basic blocks considered in instrumentation %d ******",numberOfInstPoints);
}

uint32_t 
CustomDataExtender::byteCountForInst (uint32_t instPointIdx,
                 uint64_t instStubAddress,TextSection* textSect)
{
    ASSERT(extendedDataSize && dataBufferEntryCount);

    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);

    bool isApplicationExit = (block == terminationBlock);
    assert (isApplicationExit);

    uint32_t insnCount = 0;
    uint32_t minInsnCount = 15;
    uint32_t indexAccessInsnCount = 0;
    uint32_t tocUpdateInsnCount = 0;
    uint32_t exitInsnCount = 2;

    if (tocXorUpdateVal) {
        tocUpdateInsnCount = 2;
        if (SECOND_HALFWORD(tocXorUpdateVal)) {
            tocUpdateInsnCount = 2;
        }
    }
    int32_t counterOffset = getDataOffsetForInst(0);
    if (!Instruction::isInLoadOffsetInsnRange (counterOffset)) {
        indexAccessInsnCount = 5;
    } else {
        indexAccessInsnCount = 3;
    }

    int32_t tocLoadBackOffset = 20 * (xcoffFile->is64Bit() ? 2 : 1);

    uint32_t savedReg = 31;
    uint32_t scratchReg = 30;

    insnCount += minInsnCount;
    insnCount += tocUpdateInsnCount;
//    insnCount += indexAccessInsnCount;
    insnCount += tocUpdateInsnCount;  // for exit block
    insnCount += exitInsnCount;       // for exit block
  
    return (insnCount*sizeof(uint32_t));
}

uint32_t 
CustomDataExtender::generateCodeForInst (uint32_t instPointIdx,
              uint64_t instStubAddress,TextSection* textSect,
              BaseGen* gen,uint32_t genBufferOffset)
{
    ASSERT(extendedDataSize && dataBufferEntryCount);
    ASSERT(gen->getParsedBase() == (Base*)textSect);

    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);

    bool isApplicationExit = (block == terminationBlock);
    assert (isApplicationExit);

    uint32_t maxInsnCount = maxByteCountPerInst() / sizeof(uint32_t);

    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[maxInsnCount];

    uint32_t savedReg = 31;
    uint32_t scratchReg = 30;
    uint32_t extraReg = 0;
    ASSERT (savedReg > LAST_VOLATILE_REG);
    ASSERT (scratchReg > LAST_VOLATILE_REG);
    ASSERT (!extraReg);
    int32_t savedRegSaveOffset   = getGPRSaveOffset (savedReg);
    int32_t scratchRegSaveOffset = getGPRSaveOffset (scratchReg);
    int32_t counterOffset = getDataOffsetForInst (0);

    int32_t tocLoadBackOffset = 20 * (xcoffFile->is64Bit() ? 2 : 1);

    MemoryOperation* memop = block->getMemoryOp (0);
    ASSERT(memop);

    uint64_t instPointAddress = memop->getInsnAddr();

    ASSERT(textSect->inRange(instPointAddress));

    if (!Instruction::isInJumpInsnRange (instPointAddress,instStubAddress)) {
       PRINT_INFOR("mem op %#18llx,%d can not be instrumented due to branch offset limit",
                    block->getHashCode().getValue(),0);
       delete[] insnBuffer;
       return (0);
    }

    DEBUG(memop->print());
    memop->print();

    uint64_t instStubAddressForMemop = instStubAddress + (insnCount * sizeof(uint32_t));
    fprintf (stdout, "Stub starts at address %llx\n", instStubAddressForMemop);
    
    bool originalFirst = false;
    Instruction origInsn = memop->getInstruction();

    updateTOCRegister(insnBuffer,&insnCount);

    insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,savedRegSaveOffset);
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);

/*
    if (!Instruction::isInLoadOffsetInsnRange (counterOffset)) {
       insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(scratchReg,counterOffset);
       insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(scratchReg,counterOffset);
       insnBuffer[insnCount++] = Instruction::generateLoadWordIndx(savedReg,REG_TOC,scratchReg);

       insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);
       insnBuffer[insnCount++] = Instruction::generateStoreWordIndx(savedReg,REG_TOC,scratchReg);
    } else {
       insnBuffer[insnCount++] = Instruction::generateLoadWord(savedReg,REG_TOC,counterOffset);
       insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);
       insnBuffer[insnCount++] = Instruction::generateStoreWord(savedReg,REG_TOC,counterOffset);
    }
 */
    saveCRRegister(insnBuffer,&insnCount,scratchReg);

    saveLNKRegister(insnBuffer,&insnCount,scratchReg);


    updateTOCRegister(insnBuffer,&insnCount);

    uint64_t fromAddr = instStubAddress + (insnCount * sizeof(uint32_t));
    uint64_t toAddr   = libraryFuncStubAddrs[0];
    insnBuffer[insnCount++] = Instruction::generateCallToImmediate(fromAddr,toAddr);
    insnBuffer[insnCount++] = Instruction::generateLoadImmediate(savedReg,1);

    updateTOCRegister(insnBuffer,&insnCount);

    restLNKRegister(insnBuffer,&insnCount,scratchReg);
    restCRRegister (insnBuffer,&insnCount,scratchReg);

    insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,savedRegSaveOffset);

    updateTOCRegister(insnBuffer,&insnCount);

    if (!originalFirst) {
       insnBuffer[insnCount++] = origInsn;
    }

    fromAddr = instStubAddress + (insnCount * sizeof(uint32_t));
    toAddr   = instPointAddress + sizeof(uint32_t);
    insnBuffer[insnCount++] = Instruction::generateJumpInsn(fromAddr,toAddr);

    insnBuffer[insnCount++] = Instruction((instPointIdx << 16) | 0);

    fromAddr = instPointAddress;
    toAddr   = instStubAddressForMemop;
    Instruction origJumpToStub = Instruction::generateJumpInsn(fromAddr,toAddr);
    gen->writeWord(instPointAddress-textSect->getSectHeader()->GET(s_vaddr),origJumpToStub.bits());

    ASSERT (insnCount <= maxInsnCount);
    gen->writeInstructions (genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    return (insnCount*sizeof(uint32_t));
}
