#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include "PrintExtendedData.h"

FILE* openOutputFile(const char* comment, uint32_t binaryFile)
{
    uint32_t rankId = 0;

    char strBuffer[__MAX_STRING_SIZE];

    uint32_t processId = getpid();
    uint32_t hostId = gethostid();

    FILE* fp = fopen("RankPid","r");
    uint32_t entryFound = 0;
    if (fp) {
        while (fgets(strBuffer,1024,fp)) {
            strBuffer[1023] = '\0';
            uint32_t corresPid = 0,corresHid = 0;
            sscanf(strBuffer,"%u %u %u",&rankId,&corresPid,&corresHid);
            if ((corresPid == processId) && (corresHid == hostId)) {
                entryFound = 1;
                break;
            }
        }
        fclose(fp);
    }

    char appName[__MAX_STRING_SIZE];
    uint32_t phaseId = 0;
    char extension[__MAX_STRING_SIZE];

    sscanf(comment,"%s %u %s",appName,&phaseId,extension);

    if (phaseId > 0) 
       if (entryFound)
          sprintf (strBuffer,"%s.phase.%d.meta_%04d.%s",appName,phaseId,rankId,extension);
       else
          sprintf (strBuffer,"%s.phase.%d.meta_%04d.%04d.%s",appName,phaseId,processId,hostId,extension);
    else
       if (entryFound)
          sprintf (strBuffer,"%s.meta_%04d.%s",appName,rankId,extension);
       else
          sprintf (strBuffer,"%s.meta_%04d.%04d.%s",appName,processId,hostId,extension);
    
    fp = fopen(strBuffer,"w");
    if(!fp){
        fprintf(stderr,"Error : Can not open %s to write results\n",strBuffer);
        return NULL;
    }

    if (!binaryFile)
    {
       fprintf(fp,"# appname   = %s\n",appName);
       fprintf(fp,"# extension = %s\n",extension);
       fprintf(fp,"# phase     = %d\n",phaseId);
       fprintf(fp,"# rank      = %d\n",rankId);
    }

    return fp;
}

/*** This method is called at program termination. 
***/

void Ohman_endFuncCall (char* base, uint32_t entryCount, const char* comment)
{
    uint32_t res;
    uint32_t len;
    Address_t mask;
    uint32_t numWrites = 0;
    uint32_t binaryFile = 0;
    char buffer[__MAX_STRING_SIZE];
    
    mask = ~((Address_t)(sizeof(uint32_t)-1));
    if (base)
    {
       binaryFile = *((uint32_t*)base);
       base += sizeof (uint32_t);
    }
       
    FILE* fd = openOutputFile (comment, binaryFile);
    if (fd && base)
    {
       len = *((uint32_t*)base);
       while (len)
       {
          ++ numWrites;
          base += sizeof (uint32_t);
          res = fwrite (base, sizeof(char), len, fd);
          if (res != len)
             fprintf (stderr, "Error when dumping extended data. Wrote %d bytes instead of the requested %d.\n",
                     res, len);
          base += len;
          /* now align base to next word */
          base = (char*)(((Address_t)(base+sizeof(uint32_t)-1)) & mask);
          len = *((uint32_t*)base);
       }
    }
    if (fd && !binaryFile)
       fprintf (fd, "\n# number of writes = %d\n", numWrites);
    fclose (fd);
/*    printf("Application termination. Print your simulation results\n");
 */
}

