#include <BinaryFile.h>
#include <FileHeader.h>

bool FileHeader::verify(uint16_t targetSize){
    if(GET(f_opthdr) != targetSize){
        PRINT_ERROR("This is not executable, aux-header size does not match %d %d",GET(f_opthdr),targetSize);
        return false;
    }
    if(!(GET(f_flags) & F_EXEC)){
        PRINT_ERROR("Currently only the executable files are instrumented");
        return false;
    }
    if(GET(f_flags) & F_LNNO){
        PRINT_ERROR("Failing since the executable seems to be stripped");
        return false;
    }
    return true;
}

void FileHeader::print() { 
    PRINT_INFOR("HEADER");
    PRINT_INFOR("\tMagic : %#x",GET(f_magic));
    PRINT_INFOR("\t#Sect : %d",GET(f_nscns));
    if(GET(f_nsyms)){
        PRINT_INFOR("\tSymAt : %#llx",GET(f_symptr));
        PRINT_INFOR("\t#Syms : %d (%d bytes)",GET(f_nsyms),GET(f_nsyms) * Size__NN_bit_SymbolTable_Entry);
    }
    PRINT_INFOR("\t[%#llx --- %#llx]",GET(f_symptr),GET(f_symptr)+GET(f_nsyms) * Size__NN_bit_SymbolTable_Entry);
    PRINT_INFOR("\tFlags : %#x LOADONLY SHROBJ DYNLOAD][R PATCH AR32W R] [R R R R] [R LNNO EXEC RELFLG]",GET(f_flags));
}

void FileHeader::initFilePointers(BinaryInputFile* binaryInputFile){
    if(GET(f_nsyms)){
        symbolTablePtr = binaryInputFile->fileOffsetToPointer(GET(f_symptr));
    }
}

uint32_t FileHeader32::read(BinaryInputFile* binaryInputFile){
    setFileOffset(binaryInputFile->currentOffset());

    if(!binaryInputFile->copyBytesIterate(&entry,Size__32_bit_File_Header)){
        PRINT_ERROR("File header (32) can not be read");
    }

    verify(Size__32_bit_Auxilary_Header);
    initFilePointers(binaryInputFile);

    return Size__32_bit_File_Header;
}

uint32_t FileHeader64::read(BinaryInputFile* binaryInputFile){
    setFileOffset(binaryInputFile->currentOffset());

    if(!binaryInputFile->copyBytesIterate(&entry,Size__64_bit_File_Header)){
        PRINT_ERROR("File header (64) can not be read");
    }

    verify(Size__64_bit_Auxilary_Header);
    initFilePointers(binaryInputFile);

    return Size__64_bit_File_Header;
}
