#include <SymbolTable.h>
#include <RelocationTable.h>
#include <XCoffFile.h>
#include <BinaryFile.h>

RelocationTable::RelocationTable(char* ptr,uint32_t s,XCoffFile* xcoff)
        : Base(XCoffClassTypes_relocation),
          relocationPtr(ptr),numOfRelocations(s),
          symbolTable(NULL),xCoffFile(xcoff)
{
    if(getXCoffFile()->is64Bit()){
        sizeInBytes = numOfRelocations * Size__64_bit_RelocationTable_Entry;
    } else {
        sizeInBytes = numOfRelocations * Size__32_bit_RelocationTable_Entry;
    }
    relocations = new Relocation*[numOfRelocations];
}

void Relocation::print(SymbolTable* symbolTable,uint32_t index){
    PRINT_INFOR("\tRLC [%3d](adr %#llx)(sym %9d)(sze %3d)(typ %3d)",
                index,GET(r_vaddr),GET(r_symndx),GET(r_rsize),GET(r_rtype));
    if(symbolTable){
        symbolTable->printSymbol(GET(r_symndx));
    }
}

void RelocationTable::print(){
    PRINT_INFOR("RELOCATIONTABLE");
    PRINT_INFOR("\tCount : %d",numOfRelocations);

    PRINT_INFOR("\tReloc :");
    for(uint32_t i = 0;i<numOfRelocations;i++){
        relocations[i]->print(symbolTable,i);
    }
}

uint32_t RelocationTable::read(BinaryInputFile* binaryInputFile){

    PRINT_DEBUG("Reading the Relocation table");

    binaryInputFile->setInPointer(relocationPtr);
    setFileOffset(binaryInputFile->currentOffset());

    uint32_t currSize = Size__32_bit_RelocationTable_Entry;
    if(getXCoffFile()->is64Bit()){
        currSize = Size__64_bit_RelocationTable_Entry;
    }

    uint32_t ret = 0;
    for(uint32_t i = 0;i<numOfRelocations;i++){
        if(getXCoffFile()->is64Bit()){
            relocations[i] = new Relocation64();
        } else {
            relocations[i] = new Relocation32();
        }
        binaryInputFile->copyBytesIterate(relocations[i]->charStream(),currSize);
        ret += currSize;
    }

    ASSERT((sizeInBytes == ret) && "FATAL : Somehow the number of read does not match");

    return sizeInBytes;
}
