#include <DemangleWrapper.h>
#include <StringTable.h>
#include <BinaryFile.h>

char* StringTable::getString(uint32_t offset){
    PRINT_DEBUG("Offset is %d in %d",offset,sizeInBytes);
    ASSERT(offset < sizeInBytes); 
    if(!offset)
        return "";
    return stringTablePtr+offset; 
}

void StringTable::print() { 
    PRINT_INFOR("STRINGTABLE");
    if(stringTablePtr && sizeInBytes){
        PRINT_INFOR("\tSize  : %d",sizeInBytes);
    } else {
        PRINT_INFOR("\tSize  : 0 EMPTY");
    }

    PRINT_INFOR("\tStrs  :");
    for(uint32_t currByte = sizeof(uint32_t);currByte<sizeInBytes;currByte++){
        char* ptr = (char*)(stringTablePtr+currByte);
        DemangleWrapper wrapper;
        char* demangled = wrapper.demangle_combined(ptr);
        ASSERT(demangled && "FATAL : demangling should always return non-null pointer");

        PRINT_INFOR("%9d %s --- %s",currByte,ptr,demangled);

        currByte += strlen(ptr);
    }
}

uint32_t StringTable::read(BinaryInputFile* binaryInputFile){
    
    binaryInputFile->setInPointer(stringTablePtr);
    setFileOffset(binaryInputFile->currentOffset());

    binaryInputFile->copyBytes(&sizeInBytes,sizeof(uint32_t));
    DEBUG(
    uint32_t left = binaryInputFile->bytesLeftInBuffer();
    PRINT_DEBUG("%d bytesleft to the end and total string table size is %d",left,sizeInBytes);
    ASSERT((left == sizeInBytes) && "FATAL : Number of bytes in string table does not match");
    );
    return sizeInBytes;
}
