#include <Instruction.h>
#include <FileHeader.h>
#include <AOutHeader.h>
#include <SectHeader.h>
#include <RawSection.h>
#include <SymbolTable.h>
#include <StringTable.h>
#include <RelocationTable.h>
#include <LineInfoTable.h>
#include <XCoffFile.h>
#include <Iterator.h>
#include <BinaryFile.h>
#include <Function.h>

TIMER(
	extern double cfg_s1;
	extern double cfg_s2;
	extern double cfg_s3;
	extern double cfg_s4;
	extern double cfg_s5;
);

#define ISXCOFF64(__a) ((0x01ef == (__a)) || (0x01f7 == (__a)))
#define ISXCOFF32(__a) (0x01df == (__a))

DEBUG(
uint32_t readBytes = 0;
);

#include <BitSet.h>
void XCoffFile::testBitSet(){
    PRINT_INFOR("Testing BitSet functionality");
    BitSet<BasicBlock*>** foo = new BitSet<BasicBlock*>*[3];

    foo[0] = new BitSet<BasicBlock*>(23);
    foo[1] = new BitSet<BasicBlock*>(23);
    foo[0]->print();
    foo[0]->setall();
    foo[0]->print();

    foo[0]->remove(0);
    foo[0]->remove(1);
    foo[1]->insert(1);
    foo[1]->insert(3);

    foo[0]->print();
    foo[1]->print();

    *foo[0] |= *foo[1];
    
    foo[0]->print();
    foo[1]->print();
}


void XCoffFile::parse(){

	TIMER(double t1 = timer());	

    binaryInputFile.readFileInMemory(xcoffFileName); 

    uint16_t magic = 0x0;
    if(!binaryInputFile.copyBytes(&magic,sizeof(uint16_t))){
        PRINT_ERROR("The magic number can not be read\n");
    }

    if(ISXCOFF64(magic)){
        PRINT_INFOR("The executable is 64-bit");
        is64BitFlag = true;
    } else if(ISXCOFF32(magic)){
        PRINT_INFOR("The executable is 32-bit");
    } else {
        PRINT_ERROR("The magic number is not a valid one [%#x]",magic);
    }

    PRINT_INFOR("Parsing the executable");
    readFileHeader();
    readAuxilaryHeader();
    readSectionHeaders();
    readRawSectionData();

    DebugSection* debugRawSect = NULL;
    for(uint32_t i=1;i<=numberOfSections;i++){
        if(rawSections[i] && rawSections[i]->IS_SECT_TYPE(DEBUG)){
            ASSERT(!debugRawSect && "FATAL: Why are there more than one debug raw data");
            debugRawSect = (DebugSection*)(rawSections[i]);
        }
    }
    readSymbolStringTable(debugRawSect);

    for(uint32_t i=1;i<=numberOfSections;i++){
        rawSections[i]->setSymbolTable(symbolTable);
    }

    readRelocLineInfoTable();

	TIMER(double t2 = timer();PRINT_INFOR("___timer: Parsing Step I parse XCoff    : %.2f",t2-t1);t1=t2);

    findFunctions();

	TIMER(t2 = timer();PRINT_INFOR("___timer: Parsing Step II findFunctions : %.2f",t2-t1);t1=t2);

    generateCFGs();

	TIMER(t2 = timer();PRINT_INFOR("___timer: Parsing Step III generateCFGs : %.2f",t2-t1);t1=t2);
	TIMER(PRINT_INFOR("    ___timer: CFG Step I basic blocks   : %.6f",cfg_s1));
    TIMER(PRINT_INFOR("    ___timer: CFG Step I Edges          : %.6f",cfg_s2));
    TIMER(PRINT_INFOR("    ___timer: CFG Step I Graph          : %.6f",cfg_s3));
    TIMER(PRINT_INFOR("    ___timer: CFG Step I Unreachable    : %.6f",cfg_s4));
    TIMER(PRINT_INFOR("    ___timer: CFG Step I Dominator      : %.6f",cfg_s5));

    findMemoryFloatOps();

	TIMER(t2 = timer();PRINT_INFOR("___timer: Parsing Step IV memoryFloatOp : %.2f",t2-t1);t1=t2);

    for(uint32_t i=1;i<=numberOfSections;i++){
        numberOfFunctions += rawSections[i]->getNumberOfFunctions();
        numberOfBlocks    += rawSections[i]->getNumberOfBlocks();
        numberOfMemoryOps += rawSections[i]->getNumberOfMemoryOps();
        numberOfFloatPOps += rawSections[i]->getNumberOfFloatPOps();
    }
}

void XCoffFile::readFileHeader() {
    PRINT_INFOR("Parsing the header");
    if(is64Bit()){
        fileHeader = new FileHeader64();
    } else {
        fileHeader = new FileHeader32();
    }
    ASSERT(fileHeader);
    fileHeader->read(&binaryInputFile);
    DEBUG(
        readBytes += fileHeader->getSizeInBytes();
        ASSERT(binaryInputFile.alreadyRead() == readBytes);
    );
}

void XCoffFile::readAuxilaryHeader(){
    PRINT_INFOR("Parsing the auxilary header");
    if(is64Bit()){
        aOutHeader = new AOutHeader64();
    } else {
        aOutHeader = new AOutHeader32();
    }
    ASSERT(aOutHeader);
    aOutHeader->read(&binaryInputFile);
    DEBUG(
        readBytes += aOutHeader->getSizeInBytes();
        ASSERT(binaryInputFile.alreadyRead() == readBytes);
    );
    textSectionIndex = aOutHeader->GET(o_sntext);
    dataSectionIndex = aOutHeader->GET(o_sndata);
    bssSectionIndex = aOutHeader->GET(o_snbss);
    loaderSectionIndex = aOutHeader->GET(o_snloader);
    tocSectionIndex = aOutHeader->GET(o_sntoc);
    entrySectionIndex = aOutHeader->GET(o_snentry);
}

void XCoffFile::readSectionHeaders(){
    PRINT_INFOR("Parsing the section headers");
    numberOfSections = fileHeader->GET(f_nscns);
    ASSERT(numberOfSections && "FATAL : This file has no sections!!!!!");
    sectHeaders = new SectHeader*[numberOfSections + 1];
    ASSERT(sectHeaders);

    sectHeaders[0] = NULL;
    for(uint32_t i=1;i<=numberOfSections;i++){
        if(is64Bit()){
            sectHeaders[i] = new SectHeader64(i);
        } else {
            sectHeaders[i] = new SectHeader32(i);
        }
        ASSERT(sectHeaders[i]);
        sectHeaders[i]->read(&binaryInputFile);
        DEBUG(
            readBytes += sectHeaders[i]->getSizeInBytes();
            ASSERT(binaryInputFile.alreadyRead() == readBytes);
        );
    }
}

void XCoffFile::processOverflowSections(){
    PRINT_INFOR("Processing the overflow sections");
    for(uint32_t i=1;i<=numberOfSections;i++){
        if(sectHeaders[i]->IS_SECT_TYPE(OVRFLO)){
            uint32_t whichSection = sectHeaders[i]->GET(s_nreloc);
            ASSERT((whichSection <= numberOfSections) && "FATAL : Which section is this??\n");
            ASSERT(!sectHeaders[whichSection]->getOverFlowSection() && "FATAL : More than 1 overflow??\n");
            sectHeaders[whichSection]->setOverFlowSection(sectHeaders[i]);
        }
    }
}

void XCoffFile::readRawSectionData(){

    PRINT_INFOR("Parsing the raw section data");

    rawSections = new RawSection*[numberOfSections + 1];
    ASSERT(rawSections);

    rawSections[0] = NULL;
    for(uint32_t i=1;i<=numberOfSections;i++){
        rawSections[i] = RawSection::newRawSection(sectHeaders[i],this);
        if(rawSections[i]){
            rawSections[i]->read(&binaryInputFile);
        }
    }
}

void XCoffFile::readSymbolStringTable(DebugSection* debugRawSect){

    PRINT_INFOR("Parsing the string and symbol tables");

    uint32_t numberOfSyms = fileHeader->GET(f_nsyms);
    if(numberOfSyms){
        char* startPtr = fileHeader->getSymbolTablePtr();
        ASSERT(startPtr && "FATAL : Somehow the symbol table exists but the data does not");

        char* stringTablePtr = startPtr + (numberOfSyms * Size__NN_bit_SymbolTable_Entry);
        if(binaryInputFile.isInBuffer(stringTablePtr)){
            stringTable = new StringTable(stringTablePtr);
            stringTable->read(&binaryInputFile);
        } else {
            ASSERT(NULL && "FATAL : Some how the string table does not exist");
        }

        symbolTable = new SymbolTable(startPtr,numberOfSyms,this);
        symbolTable->setStringTable(stringTable);
        symbolTable->setDebugSection(debugRawSect);
        symbolTable->read(&binaryInputFile);
    } else {
        ASSERT(numberOfSyms && "FATAL : At this moment we can not handle executables with no symbol table");
    }
}

void XCoffFile::readRelocLineInfoTable(){

    PRINT_INFOR("Parsing the LineInformation and Relocation tables");

    processOverflowSections();

    for(uint32_t i=1;i<=numberOfSections;i++){
        RelocationTable* relocationTable = sectHeaders[i]->readRelocTable(&binaryInputFile,this);
        if(relocationTable){
            relocationTable->setSymbolTable(symbolTable);
        }
        LineInfoTable* lineInfoTable = sectHeaders[i]->readLineInfoTable(&binaryInputFile,this);
        if(lineInfoTable){
            lineInfoTable->setSymbolTable(symbolTable);
        }
    }
}

void XCoffFile::briefPrint() 
{ 
    if(fileHeader) 
        fileHeader->print(); 
    if(aOutHeader) 
        aOutHeader->print(); 
    for(uint32_t i=1;i<=numberOfSections;i++){
        sectHeaders[i]->print();
    }
}

void XCoffFile::print(){

    if(fileHeader) fileHeader->print();
    if(aOutHeader) aOutHeader->print();
    for(uint32_t i=1;i<=numberOfSections;i++){
        sectHeaders[i]->print();
        rawSections[i]->print();
    }
    for(uint32_t i=1;i<=numberOfSections;i++){
        if(sectHeaders[i]->getRelocationTable())
            sectHeaders[i]->getRelocationTable()->print();
        if(sectHeaders[i]->getLineInfoTable())
            sectHeaders[i]->getLineInfoTable()->print();
    }
    if(symbolTable) symbolTable->print();
    if(stringTable) stringTable->print();
}

void XCoffFile::displaySymbols(){

    ASSERT(rawSections && "FATAL : Raw data is not read");
    ASSERT(symbolTable && "FATAL : Symbol table is missing");

    uint32_t numberOfSymbols = symbolTable->getNumberOfSymbols();
    Symbol** addressSymbols = new Symbol*[numberOfSymbols];
    numberOfSymbols = symbolTable->filterSortAddressSymbols(addressSymbols,numberOfSymbols);

    if(numberOfSymbols){
        for(uint32_t i=1;i<=numberOfSections;i++){
            if(!rawSections[i]->IS_SECT_TYPE(OVRFLO)){
                rawSections[i]->displaySymbols(addressSymbols,numberOfSymbols);
            }
        }
    }
    delete[] addressSymbols;
}

void XCoffFile::findFunctions(){
    ASSERT(symbolTable && "FATAL : Symbol table is missing");
    ASSERT(rawSections && "FATAL : Raw data is not read");
    for(uint32_t i=1;i<=numberOfSections;i++){
        rawSections[i]->findFunctions();
    }
}

void XCoffFile::generateCFGs(){
    ASSERT(rawSections && "FATAL : Raw data is not read");
    for(uint32_t i=1;i<=numberOfSections;i++){
        rawSections[i]->generateCFGs();
    }
}

void XCoffFile::findMemoryFloatOps(){
    ASSERT(rawSections && "FATAL : Raw data is not read");
    for(uint32_t i=1;i<=numberOfSections;i++){
        rawSections[i]->findMemoryFloatOps();
    }
}

RawSection* XCoffFile::findRawSection(uint64_t addr){
    ASSERT(rawSections && "FATAL : Raw data is not read");
    for(uint32_t i=1;i<=numberOfSections;i++){
        if(rawSections[i]->inRange(addr))
            return rawSections[i];
    }
    return NULL;
}

uint64_t XCoffFile::getTOCAddress(){
    return aOutHeader->GET(o_toc);
}

uint64_t XCoffFile::readTOC(int32_t offset){
    RawSection* tocSection = rawSections[tocSectionIndex];
    ASSERT(tocSection && "FATAL: TOC is missing");
    AddressIterator ait = tocSection->getAddressIterator();
    uint64_t tocAddr = aOutHeader->GET(o_toc);
    tocAddr += offset;
    ait.skipTo(tocAddr);
    return tocSection->readBytes(&ait);
}

RawSection* XCoffFile::getBSSSection(){
    if(bssSectionIndex)
        return rawSections[bssSectionIndex];
    return NULL;
}

RawSection* XCoffFile::getLoaderSection(){
    if(loaderSectionIndex)
        return rawSections[loaderSectionIndex];
    return NULL;
}

RawSection* XCoffFile::getTOCSection(){
    return rawSections[tocSectionIndex];
}

BasicBlock* XCoffFile::findBasicBlock(HashCode* hashCode){
    ASSERT(hashCode->isBlock() && "FATAL: The given hashcode for the block is incorrect");

    uint32_t sectionNo = hashCode->getSection();
    uint32_t functionNo = hashCode->getFunction(); 
    uint32_t blockNo = hashCode->getBlock();

    if(sectionNo >= numberOfSections)
        return NULL;

    Function* func = rawSections[sectionNo]->getFunction(functionNo);
    if(!func)
        return NULL;

    return func->getBlock(blockNo);
}

uint32_t XCoffFile::getAllBlocks(BasicBlock** arr){
    uint32_t ret = 0;
    for(uint32_t i=1;i<=numberOfSections;i++){
        uint32_t n = rawSections[i]->getAllBlocks(arr);
        arr += n;
        ret += n;
    }
    return ret;
}

RelocationTable* XCoffFile::getRelocationTable(uint32_t idx){ 
    return sectHeaders[idx]->getRelocationTable(); 
}
LineInfoTable* XCoffFile::getLineInfoTable(uint32_t idx){ 
    return sectHeaders[idx]->getLineInfoTable(); 
}
uint32_t XCoffFile::getFileSize() { 
    return binaryInputFile.getSize(); 
}
uint64_t XCoffFile::getDataSectionVAddr(){
    return sectHeaders[dataSectionIndex]->GET(s_vaddr);
}
uint32_t XCoffFile::getDataSectionSize(){
    return sectHeaders[dataSectionIndex]->GET(s_size);
}
uint64_t XCoffFile::getBSSSectionVAddr(){
    if(bssSectionIndex)
        return sectHeaders[bssSectionIndex]->GET(s_vaddr);
    return 0;
}
uint64_t XCoffFile::getTextSectionVAddr(){
    return sectHeaders[textSectionIndex]->GET(s_vaddr);
}

void XCoffFile::setLineInfoFinder(){
    PRINT_INFOR("Building LineInfoFinders");
    for (uint32_t i = 1; i <= numberOfSections; i++){
        rawSections[i]->buildLineInfoFinder();
    }
}

void XCoffFile::findLoops(){
    PRINT_INFOR("Finding Loops");
    for (uint32_t i = 1; i <= numberOfSections; i++){
        rawSections[i]->buildLoops();
    }
}

