#include <SectHeader.h>
#include <Iterator.h>
#include <Function.h>
#include <LineInfoFinder.h>
#include <XCoffFile.h>
#include <BasicBlockTracer.h>

bool BasicBlockTracer::filterBlockForInst(BasicBlock* bb){
    if (!bb)
        return true;
    if (bb->isTrace() || bb->isJumpTable())
        return true;
    return false;
}


/* initialize any extra data being put into the data section */
void BasicBlockTracer::initializeReservedData(DataSection* dataSect,BaseGen* gen){
    uint32_t lineno = 0;
    char* functionNameOfBlock = NULL;
    LineInfoFinder* LIF = ((TextSection*)xcoffFile->getRawSection(xcoffFile->getTextSectionIndex()))->getLineInfoFinder();
    uint64_t baseAddress;

    for (uint32_t i = 0; i < numberOfInstPoints; i++){
            /* get the address, line number and function name of this block */
        baseAddress = blocksToInstrument[i]->getBaseAddress();
        if (LIF){
            lineno = LIF->getLineNumberInFunction(baseAddress);
            functionNameOfBlock = LIF->getFunctionName(baseAddress);
        } 

            /* write the info collected for this block to the data section */
        gen->writeWord(allBlockInfo[i]-dataSect->getSectHeader()->GET(s_vaddr),i);
        gen->writeWord(allBlockInfo[i]-dataSect->getSectHeader()->GET(s_vaddr)+sizeof(uint32_t),lineno);
        gen->writeString(allBlockInfo[i]-dataSect->getSectHeader()->GET(s_vaddr)+(2*sizeof(uint32_t)),functionNameOfBlock,strlen(functionNameOfBlock));
        gen->writeString(allBlockInfo[i]-dataSect->getSectHeader()->GET(s_vaddr)+(2*sizeof(uint32_t))+strlen(functionNameOfBlock),"\0",sizeof(char));

        PRINT_DEBUG("Writing [%d,%d,%s] (length %d) to data section at location %#lx", i, lineno, functionNameOfBlock, 
            (2*sizeof(uint32_t))+strlen(functionNameOfBlock)+sizeof(char), allBlockInfo[i]);
    }
}

uint32_t BasicBlockTracer::generateSharedLibFuncWrapper(
        uint32_t libFuncIdx,uint64_t funcCallAddr,uint32_t genBufferOffset,BaseGen* gen)
{
    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[byteCountForSharedLibFuncWrappers() / sizeof(uint32_t)];

    PRINT_DEBUG("Generating library stub for library function at %#llx", funcCallAddr);

    uint32_t scratchReg = 6;
    uint32_t fpScratchReg = 3;
    uint32_t savedReg = 31;
    uint32_t tocHoldReg = 30;
    uint32_t firstArgReg = 3;

    ASSERT(scratchReg == 6);
    ASSERT(fpScratchReg == 3);
    ASSERT(savedReg == 31);
    ASSERT(tocHoldReg == 30);
    ASSERT(firstArgReg == 3);

    int32_t saveOffset = 0x0;

    updateTOCRegister(insnBuffer,&insnCount);

    for (uint32_t i = 0; i <= LAST_VOLATILE_REG; i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(i,REG_TOC,saveOffset);
    }
    for (uint32_t i = 0; i <= LAST_VOLATILE_REG; i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDoubleFloat(i,REG_TOC,saveOffset);
    }

    saveCTRRegister(insnBuffer,&insnCount,scratchReg);
    saveEXRRegister(insnBuffer,&insnCount,scratchReg);
    saveFPSCRRegister(insnBuffer,&insnCount,fpScratchReg);

        /* load the saved block info pointer offset (saved in function stub) into the scratch register */
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,getBlockPointerOffsetSaveOffset());
        /* copy the TOC location to a function argument register */
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,firstArgReg);
        /* add the block info pointer offset to the TOC location to get the location of the block info */
    insnBuffer[insnCount++] = Instruction::generateAdd(firstArgReg,firstArgReg,scratchReg);

    insnBuffer[insnCount++] = Instruction::generateMoveFromSPR(savedReg,REG_LNK);
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,tocHoldReg);


    int32_t callOffset = funcCallAddr - execTOCAddress - tocDistUpdateVal;
    if (xcoffFile->is64Bit()){
        insnBuffer[insnCount++] = Instruction::generateSPIncrementDouble(-SAFE_FRAME_LOC*sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,callOffset);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(REG_TOC,scratchReg,sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,scratchReg,0);
    } else {
        insnBuffer[insnCount++] = Instruction::generateSPIncrementWord(-SAFE_FRAME_LOC*sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,REG_TOC,callOffset);
        insnBuffer[insnCount++] = Instruction::generateLoadWord(REG_TOC,scratchReg,sizeof(uint32_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,scratchReg,0);
    }
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(scratchReg,REG_CTR);
    insnBuffer[insnCount++] = Instruction::generateCallToCTR();

    insnBuffer[insnCount++] = Instruction::generateMoveReg(tocHoldReg,REG_TOC);
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(savedReg,REG_LNK);

    restFPSCRRegister(insnBuffer,&insnCount,fpScratchReg);
    restEXRRegister(insnBuffer,&insnCount,scratchReg);
    restCTRRegister(insnBuffer,&insnCount,scratchReg);

    for (uint32_t i = 0; i <= LAST_VOLATILE_REG; i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDoubleFloat(i,REG_TOC,saveOffset);
    }
    for (uint32_t i = 0; i <= LAST_VOLATILE_REG; i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(i,REG_TOC,saveOffset);
    }
    updateTOCRegister(insnBuffer,&insnCount);

    insnBuffer[insnCount++] = Instruction::generateReturnToLnk();

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    return (insnCount*sizeof(uint32_t));
}

void BasicBlockTracer::reserveDataForInstrumentation(){

    extendedDataSize = 0;
    extendedBeginAddr = 0;
    dataBufferEntryCount = 0;

    uint64_t dataSectionBeginAddr = xcoffFile->getDataSectionVAddr();
    uint32_t dataSectionSize = xcoffFile->getDataSectionSize();

    execTOCAddress = xcoffFile->getTOCAddress();

    uint64_t currentFreeAddress = dataSectionBeginAddr + dataSectionSize;

    uint32_t incrementSize = 0;
    if (xcoffFile->is64Bit()){
        currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
        incrementSize = sizeof(uint64_t);
    } else {
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
        incrementSize = sizeof(uint32_t);
    }

    extendedBeginAddr = currentFreeAddress;

    for (uint32_t i = 0; i < getNumOfSharedLibFuncs(); i++){
        libraryFunctionStubAddrLocs[i] = currentFreeAddress;
        currentFreeAddress += incrementSize;
    }

    currentFreeAddress = nextAlignAddressWord(currentFreeAddress);

    saveAddressCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);
    saveAddressFPCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);

    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);

    saveAddressLinkReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressCtrReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressExpReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);

    saveAddressGprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_GPR_REGS);
    saveAddressFprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_FPR_REGS);
        
    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);


        /* the offset of the block info pointer that will be saved for each library call */
    blockPointerOffset = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);

        /* reserve the locations for the information for each block */
    allBlockInfo = new uint32_t[numberOfInstPoints];
    uint32_t totalNameSizes = 0;
    char* functionNameOfBlock = NULL;
    uint32_t lineno = 0;
    LineInfoFinder* LIF = ((TextSection*)xcoffFile->getRawSection(xcoffFile->getTextSectionIndex()))->getLineInfoFinder();

    for (uint32_t i = 0; i < numberOfInstPoints; i++){
        uint64_t baseAddress = blocksToInstrument[i]->getBaseAddress();
        if (LIF){
            lineno = LIF->getLineNumberInFunction(baseAddress);
            functionNameOfBlock = LIF->getFunctionName(baseAddress);
        }
        totalNameSizes += (strlen(functionNameOfBlock) + 1);
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
        allBlockInfo[i] = currentFreeAddress;
        currentFreeAddress += ((2 * sizeof(uint32_t)) + strlen(functionNameOfBlock) + 1);
    }

    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
    extendedDataSize = currentFreeAddress - (dataSectionBeginAddr + dataSectionSize);

    tocXorUpdateVal = 0;
    if (!Instruction::isInLoadOffsetInsnRange((int32_t)(dataBufferAddress-execTOCAddress))){
        tocXorUpdateVal = (uint32_t)(extendedBeginAddr ^ execTOCAddress);
        tocDistUpdateVal = (uint32_t)(extendedBeginAddr - execTOCAddress);
    }
}

BasicBlockTracer::BasicBlockTracer(XCoffFile* xcoff,char* extension,uint32_t phaseNo)
    : CommonMethods(xcoff,extension,phaseNo,1,NULL)
{
}

void BasicBlockTracer::selectInstrumentationPoints(char* inpFile){
    
    terminationBlock = findTerminationBlock();
    ASSERT(terminationBlock && "FATAL : There is no exit block");

    uint32_t allBlocksCount = xcoffFile->getNumberOfBlocks();

    bool insertAppExitBlock = true;

    if (allBlocksCount){
        blocksToInstrument = new BasicBlock*[allBlocksCount];
        xcoffFile->getAllBlocks(blocksToInstrument);

        for (uint32_t i = 0; i < allBlocksCount; i++){
            BasicBlock* bb = blocksToInstrument[i];
            if (filterBlockForInst(bb)){
                PRINT_DEBUG("Filtering basic block %#12llx",bb->getHashCode().getValue());
                numOfBlocksFiltered++;
                continue;
            }
            if (insertAppExitBlock && (bb == terminationBlock)){
                PRINT_INFOR("The exit block sequence id will be %#x",numberOfInstPoints);
                insertAppExitBlock = false;
            }
            blocksToInstrument[numberOfInstPoints++] = bb;
        }
    } else {
        PRINT_ERROR("There is no basic block identified in the executable yet, is executable parsed");
    }

    if (!numberOfInstPoints){
        PRINT_ERROR("There is no basic block passed the filter");
    }

    if (insertAppExitBlock){
        blocksToInstrument[numberOfInstPoints++] = terminationBlock;
    }
    PRINT_INFOR("**** Number of basic blocks considered in instrumentation %d ******",numberOfInstPoints);
}

uint32_t BasicBlockTracer::byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect){

    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);
    ASSERT(extendedDataSize);

    uint64_t instPointAddress = 0;

    if (!block->findFirstInstPoint(&instPointAddress)){
        PRINT_INFOR("Basic Block %#18llx does not have suitable location for inst",
                    block->getHashCode().getValue());
        return 0;
    }
    ASSERT(textSect->inRange(instPointAddress));

    if (!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress)){
        PRINT_INFOR("Basic Block %#18llx can not be instrumented due to branch offset limit",
                    block->getHashCode().getValue());
        return 0;
    }
    uint32_t numberOfInstructions = 0;

    uint32_t origInstructionSize = 0;
    uint32_t updateTOCSize = 0;
    uint32_t saveRestSize = 0;
    uint32_t blockPointSize = 0;
    uint32_t funcCallSize = 0;

    origInstructionSize++;

    uint16_t tocUpdateValLow = FIRST_HALFWORD(tocXorUpdateVal);
    uint16_t tocUpdateValHigh = SECOND_HALFWORD(tocXorUpdateVal);
    if(tocXorUpdateVal){
        if(tocUpdateValHigh){
            updateTOCSize += 4;
        }
        updateTOCSize += 4;
    }

    saveRestSize += 2;
    blockPointSize += 3;
    funcCallSize += 9;        
    saveRestSize += 2;
    funcCallSize += 2;

    numberOfInstructions = origInstructionSize + updateTOCSize + saveRestSize + blockPointSize + funcCallSize;    
    PRINT_DEBUG("Found %d instructions in the stub for inst point %d", numberOfInstructions, instPointIdx); 

    return numberOfInstructions * sizeof(uint32_t);
}

uint32_t BasicBlockTracer::generateCodeForInst(uint32_t instPointIdx,
                                                    uint64_t instStubAddress,TextSection* textSect,
                                                    BaseGen* gen,uint32_t genBufferOffset)
{
    ASSERT(gen->getParsedBase() == (Base*)textSect);
    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);
    ASSERT(extendedDataSize);

    uint64_t instPointAddress = 0;

    if (!block->findFirstInstPoint(&instPointAddress)){
        PRINT_INFOR("Basic Block %#18llx does not have suitable location for inst",
                    block->getHashCode().getValue());
        return 0;
    }
    ASSERT(textSect->inRange(instPointAddress));

    if (!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress)){
        PRINT_INFOR("Basic Block %#18llx can not be instrumented due to branch offset limit",
                    block->getHashCode().getValue());
        return 0;
    }

    AddressIterator ait = block->getInstructionIterator();
    ait.skipTo(instPointAddress);
    Instruction origInsn = textSect->readInstruction(&ait);

    bool isApplicationExit = (block == terminationBlock);

    uint32_t savedReg = 31;
    uint32_t scratchReg = 30;
    ASSERT(savedReg > LAST_VOLATILE_REG);
    ASSERT(scratchReg > LAST_VOLATILE_REG);

    int32_t savedRegSaveOffset = getGPRSaveOffset(savedReg);
    int32_t scratchRegSaveOffset = getGPRSaveOffset(scratchReg);
    int32_t blockPointerOffsetSaveOffset = getBlockPointerOffsetSaveOffset();
    int32_t blockPointerOffset = getBlockPointerOffset(instPointIdx);

    uint32_t maxInsnCount = (maxByteCountPerInst() / sizeof(uint32_t));

    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[maxInsnCount];

    insnBuffer[insnCount++] = origInsn;

    updateTOCRegister(insnBuffer,&insnCount);

    insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,savedRegSaveOffset);
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);

    insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(scratchReg,blockPointerOffset);
    insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(scratchReg,blockPointerOffset);
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,blockPointerOffsetSaveOffset);

    generateFuncCallFull(insnBuffer,&insnCount,savedReg,instStubAddress,libraryFuncStubAddrs[0]);

    insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,savedRegSaveOffset);

    updateTOCRegister(insnBuffer,&insnCount);

    uint64_t fromAddr = instStubAddress + (insnCount * sizeof(uint32_t));
    uint64_t toAddr   = instPointAddress + sizeof(uint32_t);
    insnBuffer[insnCount++] = Instruction::generateJumpInsn(fromAddr,toAddr);
    insnBuffer[insnCount++] = Instruction(instPointIdx);

    ASSERT(insnCount <= maxInsnCount);

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    fromAddr = instPointAddress;
    toAddr   = instStubAddress;
    Instruction origJumpToStub = Instruction::generateJumpInsn(fromAddr,toAddr);
    gen->writeWord(instPointAddress-textSect->getSectHeader()->GET(s_vaddr),origJumpToStub.bits());

    PRINT_DEBUG("Generating function stub with %d instructions", insnCount);
    return (insnCount*sizeof(uint32_t));
}

void BasicBlockTracer::getSharedLibraryPathAndObj(char** path,char** name,char** obj){
    *path = getPathToInstLib();
    *obj  = "";
    if(xcoffFile->is64Bit()){
        *name = "libtracFunc_64.a";
    } else {
        *name = "libtracFunc.a";
    }
    ASSERT(path && name && obj);
}


