#ifndef _CacheSimulator_h_
#define _CacheSimulator_h_

#include <CommonMethods.h>

class CacheSimulator : public CommonMethods {
protected:
    bool     enableBBCounting;
    bool     enableLoopInclusion;
    char*    additionalInfo;
public:
    CacheSimulator(XCoffFile* xcoff,char* extension,uint32_t phaseNo,char* blockFile,bool countBB=false,
                   bool loopIncl=false);
    virtual ~CacheSimulator() { if(additionalInfo) delete[] additionalInfo; }

    void selectInstrumentationPoints(char* blockFile);

    uint32_t maxByteCountPerInst()        { return 64 * sizeof(uint32_t); }
    uint32_t byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect);
    uint32_t generateCodeForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect,
                                 BaseGen* gen,uint32_t genBufferOffset);
    uint32_t bytesPerBufferEntry() { return (2*sizeof(uint64_t)); }
#define BUFFER_SIZE 0x10000
    uint32_t howManyBufferEntry()  { return BUFFER_SIZE; }
    bool     filterBlockForInst(BasicBlock* bb);
    char*    getSharedLibFuncName(uint32_t i){
        ASSERT(i < getNumOfSharedLibFuncs());
        if(i == 0){
            return "MetaSim_endFuncCall_Simu";
        } else {
            return "MetaSim_simulFuncCall_Simu";
        }
        return NULL;
    }
    char* getGenClassName() { return (enableBBCounting ? "SimulatorWithCounters" : "CacheSimulator"); }
    uint32_t spaceForBlockInfo()   { 
        uint32_t ret = 0;
        if(enableBBCounting){
            ret = (uint32_t)(sizeof(uint64_t)*numberOfInstPoints);
        } else {
            ret = (uint32_t)(sizeof(uint16_t)*numberOfInstPoints);
        }
        return ret;
    }
    void getSharedLibraryPathAndObj(char** path,char** name,char** obj);
    void initializeReservedData(DataSection* dataSect,BaseGen* gen);

    int32_t getExtraBlockInfoAddress(uint32_t idx){
        int32_t ret = 0;
        if(enableBBCounting){
            ret = (int32_t)(extraBlockInfoAddress + (idx*sizeof(uint64_t)) - execTOCAddress - tocDistUpdateVal);
        } else {
            ret = (int32_t)(extraBlockInfoAddress + (idx*sizeof(uint16_t)) - execTOCAddress - tocDistUpdateVal);
        }
        return ret;
    }

    void setLoopInclusion(bool flg) { enableLoopInclusion = flg; }
    char* additionalInfoToPrint() { return additionalInfo; }
};

#endif /* _CacheSimulator_h_ */
