#ifndef _CountAllBlocks_h_
#define _CountAllBlocks_h_

#include <Generate.h>

class CountAllBlocks : public XCoffFileGen {
protected:
    static char*            staticPrintExtension;
    uint32_t                libraryFunctionCount;

    BasicBlock**            blocksToInstrument;
    BasicBlock*             terminationBlock;
    uint32_t                numOfBlocksFiltered;

    /*************************************************/
    /********* Data Section Extension ****************/
    uint64_t                extendedBeginAddr;

    uint64_t                execTOCAddress;
    uint32_t                tocXorUpdateVal;
    uint32_t                tocDistUpdateVal;

    uint64_t                libraryFunctionStubAddrLocs[4];
    uint64_t                libraryFuncStubAddrs[4];

    uint64_t                saveAddressCondReg;
    uint64_t                saveAddressFPCondReg;

    uint64_t                saveAddressLinkReg;
    uint64_t                saveAddressCtrReg;
    uint64_t                saveAddressExpReg;

    uint64_t                saveAddressGprRegs;
    uint64_t                saveAddressFprRegs;

    uint64_t                dataBufferAddress;

    uint64_t                commentAddress;
    char*                   sharedLibraryInfoStr;

    /*************************************************/

public:
    CountAllBlocks(XCoffFile* xcoff,char* extension,uint32_t phaseNo);
    virtual ~CountAllBlocks() { }

    void selectInstrumentationPoints(char* blockFile);

    uint32_t maxByteCountPerInst()        { return 32 * sizeof(uint32_t); }
    uint32_t byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect);
    uint32_t generateCodeForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect,
                                 BaseGen* gen,uint32_t genBufferOffset);
    uint32_t bytesPerBufferEntry() { return sizeof(uint64_t); }
    uint32_t howManyBufferEntry()  { return 1; }
    bool     filterBlockForInst(BasicBlock* bb);
    char*    getSharedLibFuncName(uint32_t i){
        ASSERT(i < getNumOfSharedLibFuncs());
        return "MetaSim_endFuncCall_Freq";
    }
    char* getGenClassName() { return "CountAllBlocks"; }
    uint32_t spaceForBlockInfo()   { return 0; }
    void getSharedLibraryPathAndObj(char** path,char** name,char** obj);

    void initializeReservedData(DataSection* dataSect,BaseGen* gen);
    void reserveDataForInstrumentation();

    int32_t getDataOffsetForInst(uint32_t idx){
        return (int32_t)(dataBufferAddress - execTOCAddress - tocDistUpdateVal);
    }

    uint32_t getNumOfSharedLibFuncs()    { return libraryFunctionCount; }
    uint32_t byteCountForSharedLibFuncWrappers() { return (getNumOfSharedLibFuncs() * 100 * sizeof(uint32_t)); }
    void     setAddrOfSharedLibFuncWrapper(uint32_t i,uint64_t addr) { ASSERT(i<libraryFunctionCount) ; libraryFuncStubAddrs[i] = addr; }

    int32_t getGPRSaveOffset(uint32_t reg){
        return (int32_t)(saveAddressGprRegs + (sizeof(uint64_t)*reg) - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getFPRSaveOffset(uint32_t reg){
        return (int32_t)(saveAddressFprRegs + (sizeof(uint64_t)*reg) - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getCRSaveOffset(){
        return (int32_t)(saveAddressCondReg - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getFPSCRSaveOffset(){
        return (int32_t)(saveAddressFPCondReg - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getEXRSaveOffset(){
        return (int32_t)(saveAddressExpReg - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getCTRSaveOffset(){
        return (int32_t)(saveAddressCtrReg - execTOCAddress - tocDistUpdateVal);
    }
    int32_t getLNKSaveOffset(){
        return (int32_t)(saveAddressLinkReg - execTOCAddress - tocDistUpdateVal);
    }
    BasicBlock** getInstBlocks()            { return blocksToInstrument; }

    void updateTOCRegister(Instruction* insnBuffer,uint32_t* insnIndex);
    void saveCTRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void restCTRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void saveEXRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void restEXRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void saveFPSCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void restFPSCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void saveCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void restCRRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void saveLNKRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void restLNKRegister(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg);
    void generateFuncCallFull(Instruction* insnBuffer,uint32_t* insnIndex,uint32_t reg,
                              uint64_t stubAddr,uint64_t funcAddr);

    uint64_t getSharedLibFuncAddrLocation(uint32_t i){
        ASSERT(i < getNumOfSharedLibFuncs());
        return libraryFunctionStubAddrLocs[i];
    }

    BasicBlock*  findTerminationBlock();
    uint32_t generateSharedLibFuncWrapper(uint32_t libFuncIdx,uint64_t funcCallAddr,
                            uint32_t genBufferOffset,BaseGen* gen);
    int32_t getCommentAddressOffset(){
        return (int32_t)(commentAddress - execTOCAddress - tocDistUpdateVal);
    }
    void verify(uint32_t written) {}
    void printInstrumentationPoints();
};

#endif /* _CountAllBlocks_h_ */
